/*
 *	database.h
 *
 *	This file defines the OPaC Data Base related classes.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_DATABASE_H_
#define	_OPaC_DATABASE_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

/****************************************************************/

#ifndef _OPaC_CONTAINER_H_
#include <OPaC/container.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

/****************************************************************/

/*
 *	A data base is a collection of entities, which may be sorted
 *	using any attribute. If accessed like a container, a default
 *	sorting list will be used.
 */

class OP_PUBLIC OPDataBase : public OPGenericContainer
{

private:
	OPDictRef			dict;				//	dictionary holding sorted dictionaries
	OPHashTableRef		array;				//	array of entities

	Card32				extr_mode;			//	extraction: mode
	OPAtomRef			extr_sort;			//	extraction: sort name
	OPAtomRef			extr_attr_name;		//	extraction: attribute name
	Count				extr_attr_index;	//	extraction: attribute index
	
protected:
	OPDataBase ();
	virtual ~ OPDataBase ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void SpecifyExtractEntity (OPAtomIn sort);
	virtual void SpecifyExtractAttribute (OPAtomIn sort, OPAtomIn name);
	virtual void SpecifyExtractAttribute (OPAtomIn sort, Count index);
	virtual Bool SpecifyExtraction (OPArrayIn x);						/* OVERRIDDEN */
	virtual Count TotalCount () const;									/* OVERRIDDEN */
	virtual Bool Get (Count pos, OPObjectOut object);					/* OVERRIDDEN */
	virtual Bool GetEntity (OPAtomIn sort, Count pos, OPEntityOut entity);
	virtual Bool GetAttribute (OPAtomIn sort, Count pos, OPAtomIn name, OPObjectOut attr);
	virtual Bool GetAttribute (OPAtomIn sort, Count pos, Count index, OPObjectOut attr);
	
	virtual Bool Insert (OPEntityIn entity);
	virtual Bool Remove (OPEntityIn entity);
	virtual void RemoveAll ();
	
	virtual Bool Find (OPAtomIn attr, OPObjectIn match, OPEntityOut entity, Count nth = 0, Bool similar = TRUE);
	
	OP_DEFINE_CLASS (OPDataBase, OPGenericContainer);
};

/****************************************************************/

/*
 *	An entity is a collection of data, which are the entity's
 *	attributes (see OPEntityModel for more details). Here is an
 *	example of an entity using a "BOOK" model :
 *	
 *		+-----------+---------------------+
 *		| Attr.name | Attribute data      |
 *		+===========+=====================+
 *		| Title     | The Gods Themselves |
 *		| Author    | Isaac Asimov        |
 *		| Publisher | Bantam Spectra      |
 *		+-----------+---------------------+
 *	
 *	The attribute data (called just attributes) may be accessed
 *	using the associated name (eg. "Title", "Author"...) or the
 *	index (eg. 1, 2...).
 *
 *	CAUTION : The index starts at 1 (one), not 0 (zero). Zero
 *			  is an invalid index.
 */

class OP_PUBLIC OPEntity : public OPObject
{
	
protected:
	OPEntityModelRef	model;					//	entity model used for this entity
	
	Count				data_num;				//	# of data entries (not really needed)
	OPArrayRef			data_array;				//	array of data entries (may be empty)
	
protected:
	OPEntity ();
	virtual ~ OPEntity ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
public:
	virtual void Initialise (OPEntityModelIn);
	
	virtual Count CountAttributes () const;
	virtual void RemoveAll ();
	
	virtual Bool GetAttribute (Count index, OPObjectOut attr) const;
	virtual Bool GetAttribute (OPAtomIn name, OPObjectOut attr) const;
	
	virtual Bool SetAttribute (Count index, OPObjectIn attr);
	virtual Bool SetAttribute (OPAtomIn name, OPObjectIn attr);
	
	virtual OPEntityModelRet GetEntityModel () const;
	
	OP_DEFINE_CLASS (OPEntity, OPObject);
};

/****************************************************************/

/*
 *	An entity is always defined by a model, specifying which
 *	data the attributes contain. Here is an example, defining
 *	a model for a simple book description :
 *
 *		+--------------+-----------+
 *		| Model Name   | BOOK      |
 *		| Attribute #1 | title     |
 *		| Attribute #2 | author    |
 *		| Attribute #3 | publisher |
 *		+--------------+-----------+
 *
 *	The attribute number is model dependent and can't change
 *	once a model has been defined.
 */

class OP_PUBLIC OPEntityModel : public OPObject
{
	
	friend class OPEntity;
	
protected:
	OPDataBaseRef		containing_db;			//	containing data base
	
	Count				user_num;				//	number of entities using this model
	OPAtomRef			name;					//	name given to this model
	
	Count				name_num;				//	# of attribute names defined
	OPAtomArray			name_array;				//	array of attribute names
	
protected:
	OPEntityModel ();
	virtual ~ OPEntityModel ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
	virtual void AddUser (OPEntity* user);
	virtual void RemoveUser (OPEntity* user);
	
public:
	virtual Bool Initialise (OPDataBaseIn db, OPStringIn name, Count);
	virtual Bool AddAttribute (OPAtomIn name);
	
	virtual Count CountAttributes () const;
	virtual Count GetAttributeIndex (OPAtomIn name) const;
	virtual Bool GetAttributeIndex (OPAtomIn name, CountOut index) const;
	virtual Bool GetAttributeName (Count index, OPAtomOut name) const;
	
	OP_DEFINE_CLASS (OPEntityModel, OPObject);
};

#endif /* _OPaC_DATABASE_H_ */
