/*
 *	grail.h
 *
 *	GRAphical Interface Layer for building graphical interface elements. The
 *	name has been suggested by Erik BRUCHEZ and Daniel MARMIER; "Graal" had
 *	been first used, which is the french translation of "Grail".
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_GRAIL_H_
#define	_OPaC_GRAIL_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

#ifndef _OPaC_WINSTYLE_H_
#include <OPaC/winstyle.h>
#endif

#ifndef _OPaC_TEXTSTYLE_H_
#include <OPaC/textstyle.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

#ifndef _OPaC_STRING_H_
#include <OPaC/string.h>
#endif

#ifndef _OPaC_COLORPIXEL_H_
#include <OPaC/colorpixel.h>
#endif

/****************************************************************/

/*
 *	Internal, private, machine dependent structures :
 */

class GfxSurface;
class GfxFontCache;

struct OPaC_GrWindow;							//	private window description
struct OPaC_GrCursor;							//	private cursor definition
struct OPaC_RemoteAppDesc;						//	private remote app definition

/****************************************************************/

/*
 *	Enumerate the available Grail Line Styles. These are
 *	the standard line styles needed by every graphical
 *	interface.
 *
 *	There exact representation may vary from system to
 *	system.
 */

OP_TYPEDEF (Card32, OPLineStyle);

#define	OP_LINE_SOLID		0					//	________
#define	OP_LINE_DOT			1					//	........
#define	OP_LINE_DASH		2					//	__ __ __

/****************************************************************/

/*
 *	The OPaC Font Manager initialises an internal font data
 *	base describing the available true type fonts.
 */

class OP_PUBLIC OPFontMan : public OPObject
{
	
private:
	static OPFontManRef	current;				//	system-wide default font manager
	
protected:
	OPDataBaseRef		font_families;			//	font families database
	OPEntityModelRef	font_fam_model;			//	font family entity model
	OPEntityModelRef	font_def_model;			//	font definition entity model
	
protected:
	OPFontMan ();
	virtual ~ OPFontMan ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
public:
	virtual void GetFontFamiliesDB (OPDataBaseOut) const;
	
	static const OPFontManRef& Current ()			{ return OPFontMan::current; }
	
	static void InitSystemFontMan (OPFontManIn);
	static void KillSystemFontMan ();
	
	OP_DEFINE_CLASS (OPFontMan, OPObject);
};

/****************************************************************/

/*
 *	The OPaC Graphic Abstraction Layer Font provides an
 *	interface to the fonts used by the underlying font
 *	machinery.
 */

class OPGrailFont : public OPObject
{
	friend class OPGrailPort;
	friend class OPGrailPixmap;
	
private:
	GfxFontCache*		font;					//	font instance
	OPRelCoord			ascender;				//	ascender of a character
	OPRelCoord			descender;				//	descender of a character
	OPRelCoord			linegap;				//	gap between two lines
	
	OPStringRef			name;					//	font name (family name only)
	OPStringRef			style;					//	font style (style name only)
	OPCoord				size;					//	font size in points
	
protected:
	OPGrailFont ();
	virtual ~ OPGrailFont ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
	virtual Bool RealizeFont ();
	
public:
	virtual Bool Initialise (OPStringIn name, OPStringIn style, OPCoord size);
	
	virtual void GetFontName (OPStringOut) const;
	virtual void GetFontStyle (OPStringOut) const;
	virtual OPCoord ReturnFontSize () const;
	
	virtual OPRelCoord ReturnAscender () const;
	virtual OPRelCoord ReturnDescender () const;
	virtual OPCoord ReturnWidth (const LatinChar* text) const;
	virtual OPCoord ReturnWidth (const UnicodeChar* text) const;
	virtual OPCoord ReturnWidth (OPStringIn text) const;
	virtual OPCoord ReturnWidth (OPStringIn text, Count start, Count length) const;
	
	OP_DEFINE_CLASS (OPGrailFont, OPObject);
};

/****************************************************************/

/*
 *	The Grail Port class is the base class used to interface
 *	any graphical operation.
 *	
 *	It can be compared with a GRAFPORT (Apple) or a GDI HDC,
 *	depending on your beliefs :-).
 *	
 *	The Grail Port is used as the base class for graphical
 *	device drivers, such as GrailWindow and GrailPixmap.
 */

class OP_PUBLIC OPGrailPort : public OPObject
{
	friend class OPGrailCursor;
	
protected:
	GfxSurface*			surface;				//	associated main surface
	
	const Card16*		dash_array;				//	array of dashes
	int					dash_num;				//	number of lines and gaps
	
protected:
	OPFontManRef		font_man;				//	associated font manager
	OPAdornerRef		adorner;				//	associated adorner
	
	OPCoord				rect_x1, rect_y1;		//	drawing rectangle top left corner
	OPCoord				rect_x2, rect_y2;		//	drawing rectangle bottom right corner
	
	OPCoord				clip_x1, clip_y1;		//	clipping rectangle top left corner
	OPCoord				clip_x2, clip_y2;		//	clipping rectangle bottom right corner
	
	OPLineStyle			line_style;				//	current line style
	OPColorPixel		fore_color;				//	foreground color
	OPColorPixel		back_color;				//	background color
	OPGrailFontRef		font;					//	current font
	
	Bool				is_ready;				//	TRUE => context initialised
	Bool				is_drawing;				//	TRUE => draw in progress
	
protected:
	OPGrailPort ();
	virtual ~ OPGrailPort ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void InitAdorner (OPAdornerIn adorner);
	
	
	//	The drawing should always be enclosed between a "Begin Draw" and an "End
	//	Draw" invocation. This will set up a local origin at [x1;y1] and clip
	//	everything to the specified rectangle. The rectangle will be marked as
	//	dirty and refreshed ASAP.
	
	virtual void BeginDraw (OPRectIn box, OPRelCoord margin = 0);
	virtual void BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	virtual void EndDraw ();
	virtual void Flush ();
	
	
	//	Clipping functions used to define the clipping rectangle. No other, more
	//	complex, area can be defined using Grail.
	
	virtual void SetClipRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void SetClipRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void ResetClipRect ();
	
	virtual void GetClipRect (OPRectOut box) const;
	virtual void GetClipRect (OPCoordOut x1, OPCoordOut y1,
							  OPCoordOut x2, OPCoordOut y2) const;
	
	
	//	The fundamental graphical operations allow to draw horizontal and vertical
	//	lines very quickly. They also handle rectangles. Support for circles and
	//	more complex graphical shapes might be added in the future, as soon as
	//	the need arises :
	
	virtual void DrawDot (OPCoord x, OPCoord y);
	virtual void DrawDot (OPCoord x, OPCoord y, OPColorPixelIn color);
	
	virtual void DrawLineDX (OPCoord x, OPCoord y, OPCoord dx);
	virtual void DrawLineDY (OPCoord x, OPCoord y, OPCoord dy);
	virtual void DrawLineToX (OPCoord x1, OPCoord y1, OPCoord x2);
	virtual void DrawLineToY (OPCoord x1, OPCoord y1, OPCoord y2);
	
	virtual void DrawRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void DrawRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void FillRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void FillRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	//	The hair lines, like the horizontal and vertical ones, include the end
	//	points. They are only used to draw style-less lines.
	
	virtual void DrawHairLine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy);
	virtual void DrawHairLineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void DrawAALine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy);
	virtual void DrawAALineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	//	It is possible to draw some simple horizontal text for the user interface.
	//	If you need more complex methods, look for ODD :
	
	virtual void ShowText (OPCoord x, OPCoord y,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0,
						   OPCoord dx = 0);
	
	virtual void ShowText (OPRectIn box,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0,
						   OPRelCoord margin = 0);
	
	virtual void ShowText (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0);
	
	
	//	Raster operations are also very useful for the user interface. Here come
	//	the related methods :
	
	virtual void ScrollRect (OPRectIn box, OPRelCoord move_x, OPRelCoord move_y,
							 OPRelCoord margin = 0);
	
	virtual void ScrollRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
							 OPRelCoord move_x, OPRelCoord move_y);
	
	virtual void CopyRect (OPGrailPortIn source,
						   OPCoord source_x, OPCoord source_y,
						   OPRectIn dest_box, OPRelCoord dest_margin = 0);
	
	virtual void CopyRect (OPGrailPortIn source,
						   OPCoord source_x, OPCoord source_y,
						   OPCoord dest_x1, OPCoord dest_y1,
						   OPCoord dest_x2, OPCoord dest_y2);
	
	virtual void SetBitmap (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy,
							const Card8* bits);
	
	virtual	void SetGrayPixels (OPCoord x, OPCoord y, OPCoord dx,
								const Card8* intensity);
	
	virtual	void SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx,
							   const Card8* red,
							   const Card8* green,
							   const Card8* blue);
	
	virtual	void SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, const Card8* rgb);
	
	//	Graphical context manipulation :
	
	virtual void SetForeColor (OPColorPixelIn fore_color);
	virtual void SetBackColor (OPColorPixelIn back_color);
	virtual void SetLineStyle (OPLineStyle line_style);
	virtual void SetFont (OPGrailFontIn font);
	
	virtual OPFontManRet ReturnFontMan () const;
	virtual OPAdornerRet ReturnAdorner ();
	
	//	Compute the pixel co-ordinates of a line (kind of Bresenham algorithm
	//	with optional error information which can be used for anti-aliasing).
	
	virtual void ComputeLineDXDots (Card32 x, Card32 y, Card32 dx, Int32 dy,
									Card32* y_array,
									Card8* y_error = 0,
									Real error = 0.0);
	
	virtual void ComputeLineDYDots (Card32 x, Card32 y, Int32 dx, Card32 dy,
									Card32* x_array,
									Card8* x_error = 0,
									Real error = 0.0);
	
	OP_DEFINE_CLASS (OPGrailPort, OPObject);
};

/****************************************************************/

/*
 *	The Grail Pixmap class provides an easy way to draw off-
 *	screen,	eg. in order to implement double buffering.
 *
 *	The depth is not actually used; internally, the pixmap is
 *	represented with 32-bit RGB + alpha.
 */

class OP_PUBLIC OPGrailPixmap : public OPGrailPort
{
	
protected:
	OPCoord			dx, dy;						//	pixmap width
	Count			depth;						//	pixmap depth (1/2/4/8/16/24)
	
	Count			partition_count;			//	# of partitions (if any)
	OPRect*			partition_array;			//	array of partitions
	
protected:
	OPGrailPixmap ();
	virtual ~ OPGrailPixmap ();
	
	virtual void StoreInstance (OPStorageIn storage);			/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);			/* OVERRIDDEN */
	
public:
	virtual void BeginDraw (OPRectIn box, OPRelCoord = 0);		/* OVERRIDDEN */
	virtual void BeginDraw (OPCoord, OPCoord, OPCoord, OPCoord);/* OVERRIDDEN */
	virtual void EndDraw ();									/* OVERRIDDEN */
	
	virtual void Initialise (OPGrailPixmapIn window, OPCoord dx, OPCoord dy);
	virtual void Initialise (Count depth, OPCoord dx, OPCoord dy);
	
	//	It is possible to partition the drawing space into several smaller
	//	rectangles in order to decrease the memory needed and to increase
	//	the updating speed of very large windows.
	
	virtual Bool AddPartition (OPRectIn box);
	virtual Bool RemovePartition (OPRectIn box);
	virtual Bool RemovePartitions ();
	virtual Bool GetPartition (Count index, OPRectOut box) const;
	virtual Count CountPartitions () const;
	
	void  GetSize (OPCoordOut dx, OPCoordOut dy) const	{ dx = this->dx; dy = this->dy; }
	Count ReturnDepth () const							{ return this->depth; }
	
	virtual Bool GetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* red, Card8* green, Card8* blue);
	virtual OPColorPixel ReturnPixel (OPCoord x, OPCoord y);
	
	OP_DEFINE_CLASS (OPGrailPixmap, OPGrailPort);
};

/****************************************************************/

/*
 *	The Grail Window class provides a mean to create a new
 *	window and its associated pixmap buffer.
 */

class OP_PUBLIC OPGrailWindow : public OPGrailPixmap
{
	
	friend class OPGrailPixmap;					//	OPGrailPixmap may access this->window
	
protected:
	OPaC_GrWindow*	window;						//	machine dependent window info
	
	Bool			is_active;					//	TRUE => active window
	Bool			is_open;					//	TRUE => open window
	Bool			is_reduced;					//	TRUE => reduced window
	Bool			is_drawing;					//	TRUE => currently drawing
	
	OPRect			draw_rect;					//	current drawing rectangle
	OPRect			bounding_box;				//	bounding box	
	OPWindowRef		view;						//	OPaC "view" of the window
	OPWindowStyle	style;						//	window style
	
protected:
	OPGrailWindow ();
	virtual ~ OPGrailWindow ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void BeginDraw (OPRectIn box, OPRelCoord = 0);		/* OVERRIDDEN */
	virtual void BeginDraw (OPCoord, OPCoord, OPCoord, OPCoord);/* OVERRIDDEN */
	virtual void EndDraw ();									/* OVERRIDDEN */
	virtual void Flush ();										/* OVERRIDDEN */
	
	
	//	Initialise a window with a fixed style. The name might not always be
	//	used, but it should always be available :
	
	virtual void Initialise (OPWindowIn view, OPWindowStyle style, OPStringIn name,
							 OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void Initialise (OPWindowIn view, OPWindowStyle style, OPStringIn name,
							 OPRectIn box, OPRelCoord margin = 0);
	
	virtual void ChangeName (OPStringIn name);
	
	virtual void Open ();
	virtual void Reduce ();
	virtual void Close ();
	virtual void ActivateAsFrontWindow (Bool front = TRUE);
	
	virtual void SysChangedActive (Bool);
	virtual void SysChangedOpen (Bool);
	virtual void SysChangedVisible (Bool);
	virtual void SysChangedReduced (Bool);
	
	virtual Bool IsActive () const;
	
	//	The window may be moved explicitely by the program. This should only be
	//	used to reflect dragging...
	
	virtual void SetOrigin (OPCoord x, OPCoord y);
	virtual void GetOrigin (OPCoordOut x, OPCoordOut y) const;
	virtual void SetSize (OPCoord x, OPCoord y);
	virtual void GetSize (OPCoordOut x, OPCoordOut y) const;
	virtual Bool GetDrawRect (OPRectOut box) const;
	
	virtual void UserChangedGeometry ();
	
	virtual void GetFrameSize (OPCoordOut xl, OPCoordOut yt, OPCoordOut xr, OPCoordOut yb);
	
	//	Called by the window manager :
	
	virtual void Redraw ();
	virtual void RedrawIcon ();
	
	//	Interrogation methods :
	
	void GetWindow (OPWindowOut w) const				{ w = this->view; }
	OPaC_GrWindow* ReturnPrivateWindowInfo () const		{ return this->window; }
	
	static Bool FindOwnWindow (OPCoord, OPCoord, OPGrailWindowOut wdo, Count index = 0);
	static Bool FindOPaCWindow (OPCoord, OPCoord, OPaC_RemoteAppDesc* desc, Count index = 0);
	static void GetFrameSize (OPWindowStyle style, OPCoordOut xl, OPCoordOut yt, OPCoordOut xr, OPCoordOut yb);
	static Bool GetDisplayBox (OPRectOut box, Count display = 0);
	static Bool GetDisplayFromPoint (OPCoord x, OPCoord y, CountOut display);
		
	OP_DEFINE_CLASS (OPGrailWindow, OPGrailPixmap);
};

/****************************************************************/

/*
 *	The Grail Cursor can be used to change the graphic
 *	cursor's shape.
 */

class OP_PUBLIC OPGrailCursor : public OPObject
{
	
protected:
	OPaC_GrCursor*		cursor;					//	machine dependent cursor info
	
	OPGrailPixmapRef	foreground;				//	foreground for cursor
	OPGrailPixmapRef	background;				//	background for cursor
	
	Int16				x_foreground;			//	[x] position in foreground pixmap
	Int16				y_foreground;			//	[y] position in foreground pixmap
	
	Int16				x_background;			//	[x] position in background pixmap
	Int16				y_background;			//	[y] position in background pixmap
	
	Int16				dx, dy;					//	size of cursor in pixmaps
	Int16				hot_x, hot_y;			//	hotspot position relative to cursor
	
protected:
	OPGrailCursor ();
	virtual ~ OPGrailCursor ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
public:
	virtual void Initialise (OPCoord hotx, OPCoord hoty,
							 OPCoord dx, OPCoord dy,
							 OPGrailPixmapIn fg, OPCoord xf, OPCoord yf,
							 OPGrailPixmapIn bg, OPCoord xb, OPCoord yb);
	
	virtual void Activate () const;
	virtual void Hide () const;
	virtual void ResetToNormal () const;
	
	OP_DEFINE_CLASS (OPGrailCursor, OPObject);
};

/****************************************************************/

extern OPGrailPixmapRef OP_PUBLIC OPaC_LoadGIF (OPStringIn name, OPMemZone* zone);

#endif /* _OPaC_GRAIL_H_ */
