/*
 *	lookman.h
 *
 *	The look manager is used to specify the look and feel of each
 *	separate OPaC application.
 *
 *	The adornments and some interface details get drawn by the adorner
 *	class, also defined in this file.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_LOOKMAN_H_
#define	_OPaC_LOOKMAN_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

#ifndef _OPaC_SCRSTYLE_H_
#include <OPaC/scrstyle.h>
#endif

#ifndef	_OPaC_ICOSTYLE_H_
#include <OPaC/icostyle.h>
#endif

#ifndef	_OPaC_LOOKCOLOR_H_
#include <OPaC/lookcolor.h>
#endif

#ifndef	_OPaC_LOOKCURSOR_H_
#include <OPaC/lookcursor.h>
#endif

#ifndef	_OPaC_COLORPIXEL_H_
#include <OPaC/colorpixel.h>
#endif

/****************************************************************/

/*
 *	The OPaC adorner is able to draw a frame with only
 *	some edges, specified by a bitset.
 */

OP_TYPEDEF (Card8, OPLookFrame);

#define	OP_FRAME_TOP		((OPLookFrame)(1))	//	draw the top edge
#define	OP_FRAME_BOTTOM		((OPLookFrame)(2))	//	draw the bottom edge
#define	OP_FRAME_LEFT		((OPLookFrame)(4))	//	draw the left edge
#define	OP_FRAME_RIGHT		((OPLookFrame)(8))	//	draw the right edge

/****************************************************************/

enum OPFlushPriority
{
	OP_FLUSH_SYNC			= 0,				//	flush synchronously (immediately)
	OP_FLUSH_FAST			= 1,				//	flush as soon as possible
	OP_FLUSH_NORMAL			= 2,				//	flush with normal priority
	OP_FLUSH_LAZY			= 3,				//	flush when there is nothing else to do
};

OP_TYPEENUM (OPFlushPriority);

/****************************************************************/

#define	OP_LKMAN_MIL		8					//	look manager: max. icon lines in GIF

/****************************************************************/

/*
 *	The OPaC Look Manager exports lots of user interface
 *	related informations (policies, colors, ressources).
 *	
 *	You may want to use the OPaC Adorner for the graphical
 *	operations (selecting the right color, drawing icons,
 *	bevels, etc).
 */

class OP_PUBLIC OPLookMan : public OPObject
{
	
private:
	struct UserIconDef;
	struct OP_PUBLIC AdornerList
	{
		OPAdornerRef	head;					//	head of adorner list
		OPAdornerRef	tail;					//	tail of adorner list
	};
	
private:
	static OPLookManRef	current;				//	current look manager (system's or user's)
	static OPLookManRef	sys_lookman;			//	system's look manager
	static OPLookIcon	icon_conv[OP_ICON_NUM];	//	icon conversion table
	static OPAtomRef	icon_name[OP_ICON_NUM];	//	icon name table
	static OPLookColor	col_conv[OP_COLOR_NUM];	//	color conversion table
	static OPAtomRef	col_name[OP_COLOR_NUM];	//	color name table
	
	static UserIconDef*	user_icon_head;			//	first user icon
	static UserIconDef*	user_icon_tail;			//	last user icon
	static Count		user_icon_num;			//	number of user icons
	
	static AdornerList	adorner;				//	ref. to first/last adorner
	
protected:
	OPGrailPixmapRef	sys_icons;				//	ref. to pixmap containing the system icons
	OPGrailFontRef		sys_font;				//	ref. to the system font
	
	OPGrailCursorRef	cursor_normal;			//	ref. to normal cursor
	OPGrailCursorRef	cursor_busy;			//	ref. to busy cursor
	OPGrailCursorRef	cursor_I_beam;          //  ref. to I beam cursor
	
	OPDicTreeRef		policies;				//	policy dictionary tree
	
	mutable OPLookIcon		 icon_cache;		//	cache for icon access
	mutable int				 icon_cache_index;	//	cache for icon index (if any)
	mutable OPGrailPixmapRef icon_cache_pixmap;	//	cache for icon pixmap
	mutable OPRect			 icon_cache_box;	//	cache for icon position & size in pixmap
	
	OPCoord				icon_dx[OP_ICON_NUM];	//	icon [x] sizes (arrow, radio button, check box...)
	OPCoord				icon_dy[OP_ICON_NUM];	//	icon [y] sizes
	OPCoord				icon_ox[OP_ICON_NUM];	//	icon [x] positions
	OPCoord				icon_oy[OP_ICON_NUM];	//	icon [y] positions
	
	Card8				col_rgb[4][OP_COLOR_NUM];//	color RGB + alpha values
	OPColorPixel		col_pixel[OP_COLOR_NUM]; // color pixel value (alpha + RGB)
	
	OPCoord				flat_3D_effect;			//	width of 3D effect on screen [pix]
	OPCoord				press_3D_effect;		//	offset of 3D effect when pressing widgets [pix]
	OPCoord				widget_frame_size;		//	size of the frame around each widget
	
	Count				auto_start_time;		//	time from click to start of repetition
	Count				auto_repeat_time;		//	time between two auto repetition events
	Count				cursor_off_time;		//	time between cursor off and cursor on
	Count				cursor_on_time;			//	time between cursor on and cursor off
	
	OPScrollArrowPos	scroll_arrow_pos;		//	preferences for scroll arrows
	
	Bool				do_hilite_button;		//	TRUE => hilite button, FALSE => press button
	Bool				do_hilite_icon;			//	TRUE => hilite icon, FALSE => press icon
	Bool				do_hilite_cell;			//	TRUE => hilite cell, FALSE => press cell
	
	Bool				do_show_active;			//	TRUE => show when widget is active
	Bool				do_show_dragging;		//	TRUE => show when widget is the drag source
	Bool				do_show_dropping;		//	TRUE => show when widget is the drop destination
	
protected:
	OPLookMan ();
	virtual ~ OPLookMan ();
	
	virtual void InitFont ();
	virtual void InitIcons ();
	virtual void InitColors ();
	virtual void InitCursors ();
	virtual void InitPolicies ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	static void Insert (UserIconDef*);
	static void Remove (UserIconDef*);
	
	virtual Bool UpdateIconCache (OPLookIcon icon) const;
	virtual void UpdateIconCache (OPLookIcon icon, int index) const;
	virtual void UpdateIconCache (OPLookIcon icon, const UserIconDef* def) const;
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	//	Special access methods :
	
	virtual void SelectCursor (OPLookCursor cursor);
	
	//	General information about the current look :
	
	virtual Bool GetIconSize (OPLookIcon icon, OPCoordOut x, OPCoordOut y) const;
	virtual Bool GetIconPos (OPLookIcon icon, OPCoordOut x, OPCoordOut y) const;
	virtual Bool GetIconName (OPLookIcon icon, OPAtomOut name) const;
	virtual OPLookIcon FindIcon (OPAtomIn name);
	virtual OPLookIcon FindNthIcon (Count index);
	virtual OPLookIcon AddUserIcon (OPAtomIn name, OPGrailPixmapIn src, OPRectIn box);
	virtual void RemoveUserIcon (OPLookIcon icon);
	
	virtual Bool GetColorName (OPLookColor color, OPAtomOut name) const;
	virtual Bool GetColorRGB (OPLookColor color, Card8Out r, Card8Out g, Card8Out b) const;
	virtual Bool GetColorRGB (OPLookColor color, Card8Out r, Card8Out g, Card8Out b, Card8Out t) const;
	virtual Bool GetColorCompactRGB (OPLookColor color, Card32Out pixel) const;
	virtual Bool GetColorPixel (OPLookColor color, OPColorPixelOut pixel) const;
	virtual Bool GetColorPixel (OPAtomIn name, OPColorPixelOut pixel) const;
	
	virtual Bool SetColorRGB (OPLookColor color, Card8 r, Card8 g, Card8 b, Card8 t = 0);
	virtual Bool SetColorCompactRGB (OPLookColor color, Card32 pixel);
	
	virtual OPLookColor FindColor (OPAtomIn name) const;
	virtual OPLookColor FindNthColor (Count index) const;
	
	OPCoord Return3DEffectWidth () const				{ return this->flat_3D_effect; }
	OPCoord Return3DPressOffset () const				{ return this->press_3D_effect; }
	OPCoord ReturnWidgetFrameSize () const				{ return this->widget_frame_size; }
	
	Count ReturnAutoStartTime () const					{ return this->auto_start_time; }
	Count ReturnAutoRepeatTime () const					{ return this->auto_repeat_time; }
	Count ReturnCursorOffTime () const					{ return this->cursor_off_time; }
	Count ReturnCursorOnTime () const					{ return this->cursor_on_time; }
	
	OPScrollArrowPos ReturnScrollArrowPos () const		{ return this->scroll_arrow_pos; }
	const OPGrailFontRef& ReturnSysFont () const		{ return this->sys_font; }
	const OPGrailPixmapRef& ReturnSysIcons () const		{ return this->sys_icons; }
	
	const OPGrailPixmapRef& ReturnIconPixmap (OPLookIcon icon);
	
	//	Miscellanous policies. The returned values describe how different
	//	objects should behave :
	
	Bool IsPolicyHiliteButton () const					{ return this->do_hilite_button; }
	Bool IsPolicyHiliteIcon () const					{ return this->do_hilite_icon; }
	Bool IsPolicyHiliteCell () const					{ return this->do_hilite_cell; }
	Bool IsPolicyShowActive () const					{ return this->do_show_active; }
	Bool IsPolicyShowDragging () const					{ return this->do_show_dragging; }
	Bool IsPolicyShowDropping () const					{ return this->do_show_dropping; }
	
	const OPDicTreeRef& ReturnPolicies () const			{ return this->policies; }
	virtual void ReloadPolicies ();
	virtual void UpdatePolicies ();
	
	//	The look manager will be initialised / freed by the OPaC application
	//	class, you need not worry about that. But you might want to replace
	//	the look manager by another one of your own.
	
	static void Initialise ();
	static void Kill ();
	
	static void Reset ();
	static void Change (OPLookManIn look_man);
	
	static const OPLookManRef& Current () { return OPLookMan::current; }
	
	static void InsertAdorner (OPAdorner* adorner);
	static void RemoveAdorner (OPAdorner* adorner);
	
	OP_DEFINE_CLASS (OPLookMan, OPObject);
};

/****************************************************************/

/*
 *	The Adorner class provides a set of useful drawing methods,
 *	needed to draw buttons, pseudo-3D widgets, etc.
 */

class OP_PUBLIC OPAdorner : public OPObject
{
	
	friend class OPLookMan;
	
private:
	OPGrailPortRef		port;					//	associated grail port
	OPLookManRef		lookman;				//	associated look manager
	
protected:
	OPAdorner ();
	virtual ~ OPAdorner ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */

public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Initialise (OPGrailPortIn grail);
	virtual void Update ();
	
	virtual void SelectSysFont ();
	
	virtual void SetForeColor (OPLookColorIn color);
	virtual void SetBackColor (OPLookColorIn color);
	
	virtual void FillRect (OPRectIn rect, OPLookColor color, OPRelCoord margin = 0);
	virtual void FillRect (OPRectIn rect, OPAtomIn color, OPRelCoord margin = 0);
	
	virtual void DrawRectHigh (OPRectIn rect, OPRelCoord margin = 0);
	virtual void DrawRectLow (OPRectIn rect, OPRelCoord margin = 0);
	virtual void DrawRectHigh (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	virtual void DrawRectLow (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void DrawRectHigh (OPColorPixel color, OPRectIn rect, OPRelCoord margin = 0);
	virtual void DrawRectLow (OPColorPixel color, OPRectIn rect, OPRelCoord margin = 0);
	virtual void DrawRectHigh (OPColorPixel color, OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	virtual void DrawRectLow (OPColorPixel color, OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void DrawRect (OPRectIn rect, OPLookColor color, OPRelCoord margin = 0);
	virtual void DrawRect (OPRectIn rect, OPAtomIn color, OPRelCoord margin = 0);
	
	virtual void DrawFrame (OPRectIn rect, OPRelCoord margin = 0);
	virtual void DrawFrame (OPRectIn rect, OPLookFrame frame, OPRelCoord margin = 0);
	virtual void DrawLineFrame (OPRectIn rect, OPRelCoord margin = 0);
	
	virtual void DrawIcon (OPCoord x, OPCoord y, OPLookIcon icon);
	virtual void DrawIcon (OPCoord x, OPCoord y, OPAtomIn icon);
	
	virtual void Flush (OPFlushPriority priority);
	
	OP_DEFINE_CLASS (OPAdorner, OPObject);
};

/****************************************************************/

#endif /* _OPaC_LOOKMAN_H_ */
