/*
 *	object.h
 *
 *	This file defines the base classes of the OPaC hierarchy. It also
 *	exports some macros for handling the run-time information.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_OBJECT_H_
#define	_OPaC_OBJECT_H_

#ifndef _OPaC_TYPES_H_
#include <OPaC/types.h>
#endif

/****************************************************************/

OP_EXPORT_CLASS (OPArchive);				//	=> OPStorage
OP_EXPORT_CLASS (OPAtom);					//	=> OPText
OP_EXPORT_CLASS (OPMessage);				//	=> OPStorage
OP_EXPORT_CLASS (OPObject);					//	root class, no superclass
OP_EXPORT_CLASS (OPRuntime);				//	=> OPObject
OP_EXPORT_CLASS (OPStorage);				//	=> OPObject
OP_EXPORT_CLASS (OPString);					//	=> OPText
OP_EXPORT_CLASS (OPTarget);					//	=> OPObject
OP_EXPORT_CLASS (OPText);					//	=> OPObject

OP_IMPORT_CLASS (OPApp);					//	=> OPDataSlot
OP_IMPORT_CLASS (OPArray);					//	=> OPContainer
OP_IMPORT_CLASS (OPContainer);				//	=> OPGenericContainer
OP_IMPORT_CLASS (OPDataSlot);				//	=> OPObject
OP_IMPORT_CLASS (OPPhantom);				//	=> OPObject
OP_IMPORT_CLASS (OPTypeInfo);				//	=> OPObject
OP_IMPORT_CLASS (OPValue);					//	=> OPObject
OP_IMPORT_CLASS (OPView);					//	=> OPResponder

OP_TYPEDEF (Card32, OPTargetType);			//	from <OPaC/targetmacro.h>

/****************************************************************/

class OPMemZone;							//	memory zone
class OPClassMan;							//	class manager
class OPaC_Target;							//	internal target publishing class
class OPaC_PubData;							//	internal public data access class
class OPaC_AtomMan;							//	internal atomic string manager
class OPaC_DebugHook;						//	internal debugging hook

struct OPaC_ParamDesc;						//	from <OPaC/targetmacro.h>
struct OPaC_PubDataDef;						//	from <OPaC/pubmacro.h>
struct OPaC_BitsetDataDef;					//	from <OPaC/pubmacro.h>

/****************************************************************/

#include <OPaC/operator_new.h>

/****************************************************************/

typedef void (*OPaC_ConstrFunc) (void*, Card32);
typedef const OPRuntimeRef& (*OPaC_ReturnRuntimeFunc)();

/****************************************************************/

/*
 *	This macro is provided to access the run-time informations
 *	about the named class type.
 */

#define	OP_RUNTIME(c)											\
	(c##Ref::ReturnRuntime ())

/****************************************************************/

/*
 *	The call frame stores information about the current execution
 *	point in the source (thanks to the OP_METHOD macros).
 */

class OP_PUBLIC OPaC_CallFrame
{
	
public:
	enum Type
	{
		NONE,								//	no stack frame
		METHOD,								//	stack frame for method (+ this)
		MEMBER,								//	stack frame for static member
		OPERATOR,							//	stack frame for operator
		FUNCTION							//	stack frame for function
	};
	
	class OP_PUBLIC StaticInfo
	{
	public:
		Type				type;			//	frame type (method/member/function)
		const LatinChar*	func_name;		//	function name
		const LatinChar*	file_name;		//	file containing function
		Count				file_line;		//	line in the file
		Count				total_calls;	//	total # of calls
	public:
		StaticInfo ();
		StaticInfo (Type t, const LatinChar* func, const LatinChar* file, Count line);
		~ StaticInfo ();
	};
	
private:
	static OPaC_CallFrame*	head;			//	base of the stack frame list
	static Count			total;			//	total # of calls
	static StaticInfo		empty;			//	empty stack frame definition (unkown function)
	static OPaC_DebugHook*	debug_hook;		//	internal debugging hook
	
private:
	OPaC_CallFrame*			original;		//	original (self if not a copy)
	OPaC_CallFrame*			next;			//	next stack frame
	StaticInfo*				static_info;	//	information about method/function/...
	const void*				method_this;	//	"this" pointer for methods
	Count					call_id;		//	call ID (regeneration number)
	
public:
	OPaC_CallFrame ();
	OPaC_CallFrame (StaticInfo* info, const void* that = 0);
	~OPaC_CallFrame ();
	
	//	Accessor methods :
	
	const StaticInfo* ReturnStaticInfo () const	{ return this->static_info; }
	
	Type ReturnType () const					{ return this->static_info->type; }
	const LatinChar* ReturnFuncName () const	{ return this->static_info->func_name; }
	const LatinChar* ReturnFileName () const	{ return this->static_info->file_name; }
	Count ReturnFileLine () const				{ return this->static_info->file_line; }
	Count ReturnNthCallNumber () const			{ return this->static_info->total_calls; }
	OPaC_CallFrame* ReturnNextFrame () const	{ return this->next; }
	const void* ReturnThis () const				{ return this->method_this; }
	Count ReturnCallID () const					{ return this->call_id; }
	
	static OPaC_CallFrame* ReturnCallFrame ()	{ return OPaC_CallFrame::head; }
	static OPaC_CallFrame* ReturnCallerFrame ()	{ return OPaC_CallFrame::head->next; }
	static const LatinChar* ReturnTypeName (Type type);
};


#define	OP_METHOD(m,n)													\
	OP_CALLFRAME (OPaC_CallFrame::METHOD, m, n, this)

#define	OP_MEMBER(m,n)													\
	OP_CALLFRAME (OPaC_CallFrame::MEMBER, m, n, 0)

#define	OP_OPERATOR(o,n)												\
	OP_CALLFRAME (OPaC_CallFrame::OPERATOR, o, n, 0)

#define	OP_FUNCTION(f,n)												\
	OP_CALLFRAME (OPaC_CallFrame::FUNCTION, f, n, 0)

#define	OP_ARG(n,T,a)													\
	OP_USE(a)	// suppress silly warnings on some compilers

#if defined(_DEBUG)
#define	OP_CALLFRAME(t,m,n,a)											\
	static OPaC_CallFrame::StaticInfo __static_call_frame_info (t, #m, __OPaC_src_file, __LINE__); \
	OPaC_CallFrame __current_call_frame (&__static_call_frame_info, a);	\
	OP_USE (__current_call_frame)
#else
#define	OP_CALLFRAME(t,m,n,a)											\
	/* ... */
#endif

/****************************************************************/

#define	OP_DEFINE_SLOT(T,n,s,g)
#define	OP_DEFINE_SLOT_READ(T,n,g)
#define	OP_DEFINE_SLOT_WRITE(T,n,s)
#define	OP_DEFINE_EVENT(E)		void E (Count id = OP_REPORT_AUTO)
#define	OP_DEFINE_ACTOR(A)		virtual void A ()
#define	OP_DEFINE_OBSERVER(O)	virtual void O (OPObjectIn object, Card32 id)

/****************************************************************/

/*
 *	The following macros expand in different ways depending
 *	what kind of project is being compiled.
 */

#define	OP_EXPORT(c,s)				OP_EXPORT_CLASS(c);  OP_DEFINE_REF(c,s)
#define	OP_DEFINE_REF(c,s)			OP_DEFINE_X_REF(c,s,OP_PUBLIC)
#define	OP_DEFINE_REF_BEGIN(c,s)	OP_DEFINE_X_REF_BEGIN(c,s,OP_PUBLIC)
#define	OP_DEFINE_REF_END(c,s)		OP_DEFINE_X_REF_END(c,s,OP_PUBLIC)
#define	OP_DEFINE_SPECIAL_REF_BEGIN(c,s) OP_DEFINE_X_SPECIAL_REF_BEGIN(c,s,OP_PUBLIC)
#define	OP_DEFINE_SPECIAL_REF_END(c,s)	 OP_DEFINE_X_REF_END(c,s,OP_PUBLIC)
#define	OP_DEFINE_ARRAY(c,s)		OP_DEFINE_X_ARRAY(c,s,OP_PUBLIC)

#define	APP_EXPORT(c,s)				APP_EXPORT_CLASS(c); APP_DEFINE_REF(c,s)
#define	APP_DEFINE_REF(c,s)			OP_DEFINE_X_REF(c,s,APP_PUBLIC)
#define	APP_DEFINE_REF_BEGIN(c,s)	OP_DEFINE_X_REF_BEGIN(c,s,APP_PUBLIC)
#define	APP_DEFINE_REF_END(c,s)		OP_DEFINE_X_REF_END(c,s,APP_PUBLIC)
#define	APP_DEFINE_ARRAY(c,s)		OP_DEFINE_X_ARRAY(c,s,APP_PUBLIC)

/****************************************************************/

#define OP_DEFINE_X_REF_BEGIN(c,s,x)							\
	OP_DEFINE_X_REF_BEGIN_0(c,s,x);								\
	OP_DEFINE_X_REF_BEGIN_1(c,s,x);								\
	OP_DEFINE_X_REF_BEGIN_2(c,s,x)

#define OP_DEFINE_X_SPECIAL_REF_BEGIN(c,s,x)					\
	OP_DEFINE_X_REF_BEGIN_0(c,s,x);								\
	OP_DEFINE_X_REF_BEGIN_1(c,s,x)

#define OP_DEFINE_X_REF_BEGIN_0(c,s,x)							\
																\
extern x c##Ref OPaC_ValueTo##c (OPValueIn val)

#define OP_DEFINE_X_REF_BEGIN_1(c,s,x)							\
																\
class x c##Ref : public s##Ref									\
{																\
	friend class OPObject;										\
	friend class c;												\
	friend x OPStorageIn operator >> (OPStorageIn, c##Out);		\
																\
	c* ReturnRawObject () const;								\
																\
protected:														\
	c##Ref (OPRuntimeIn rt, const OPaC_CallFrame* frame);		\
																\
public:															\
	c##Ref ();													\
	c##Ref (c* ptr);											\
	c##Ref (c##In ref);											\
	c##Ref (OPObjectIn obj);									\
	virtual ~c##Ref ();											\
																\
	c* CheckAndResolve () const;								\
	c* Resolve () const;										\
	c* operator -> () const;									\
																\
	static const OPRuntimeRef& ReturnRuntime ();				\
																\
	c##Ref& operator = (c* ptr);								\
	c##Ref& operator = (c##In ref);								\
	c##Ref& operator = (OPObjectIn obj);						\
																\
	virtual Card32 SizeOf () const;								\
																\
	Bool operator == (const c##In ref) const;					\
	Bool operator != (const c##In ref) const

#define OP_DEFINE_X_REF_BEGIN_2(c,s,x)							\
	c##Ref (OPValueIn);											\
	c##Ref& operator = (OPValueIn val)

#define	OP_DEFINE_X_REF_END(c,s,x)								\
																\
};																\
																\
extern x const c##Ref c##Null;									\
extern x OPaC_TypeDef c##_type;									\
extern x OPTypeDef    c##TypeDef

#define	OP_DEFINE_X_REF(c,s,x)									\
	OP_DEFINE_X_REF_BEGIN(c,s,x);								\
	OP_DEFINE_X_REF_END(c,s,x)

#define OP_DEFINE_X_ARRAY(c,s,x)								\
																\
																\
extern x OPaC_TypeDef c##Array_type;							\
extern x OPTypeDef    c##ArrayTypeDef;							\
																\
class x c##Array : public s##Array								\
{																\
public:															\
	c##Ref& operator [] (Count p) const							\
	{ return (c##Ref&)(((OPArrayRef*)(this))[0][p]); }			\
																\
	virtual ~ c##Array ();										\
																\
																\
	c##Array& operator = (OPArray* a)							\
	{ OPObjectRef::_Set ((OPObject*)a); return *this; }			\
																\
	c##Array& operator = (OPArrayIn a)							\
	{ OPObjectRef::_Set (a); return *this; }					\
}

/****************************************************************/

#define APP_DEF_CLASS_BEGIN(c,s)	OP_DEF_CLASS_BEGIN(c,s)
#define	APP_DEF_CLASS_END(c,s)		OP_DEF_CLASS_END(c,s)
#define	APP_DEFINE_CLASS(c,s)		OP_DEFINE_CLASS(c,s)

#define	APP_IMPLEMENT_CLASS(c)		OP_IMPLEMENT_CLASS(c)
#define	APP_ABSTRACT_CLASS(c)		OP_ABSTRACT_CLASS(c)
#define	APP_SPECIAL_CLASS(c)		OP_SPECIAL_CLASS(c)

/****************************************************************/

#define OP_DEF_CLASS_BEGIN(c,s)									\
																\
	typedef s super;											\
																\
	friend void c##CollectData (OPRuntime*);					\
																\
private:														\
	c (const c&);												\
	void operator = (const c&);									\
																\
public:															\
	static const OPRuntimeRef& ReturnSuperRuntime ()			\
		{ return OP_RUNTIME (s); }								\
																\
	static OPRuntimeRef runtime;								\
																\
public:															\
	virtual OPRuntimeRet ReturnRuntime () const;				\
	virtual Bool IsAbstractClass () const;						\
	virtual Bool IsSpecialClass () const;						\
	virtual void _Destroy ();									\
	static void Construct (void* ptr, Card32 size);				\
	static void InitConstruct (void* ptr, Card32 size);			\
																\
protected:														\
	c (void* ptr, Card32 size) : s (ptr, size) { }				\
																\
private:														\
	friend class c##Ref;										\
	friend class c##Array

//	End of class definition. This will generate some inlined functions
//	which are needed by the reference class.

#define	OP_DEF_CLASS_END(c,s)									\
	OP_DEF_CLASS_END_1(c,s)										\
	OP_DEF_CLASS_END_2(c,s)
																\
#define	OP_DEF_CLASS_END_1(c,s)									\
																\
inline															\
c##Ref::c##Ref (OPObjectIn obj) : s##Ref ()						\
{ OPObjectRef::_Set (obj, c##Ref::ReturnRuntime (),				\
					 OPaC_CallFrame::ReturnCallFrame ()); }		\
																\
inline															\
c##Ref::c##Ref (OPValueIn val) : s##Ref ()						\
{ OPObjectRef::_Set (val, c##Ref::ReturnRuntime (),				\
					 OPaC_CallFrame::ReturnCallFrame ()); }		\
																\
inline c##Ref&													\
c##Ref::operator = (OPObjectIn obj)								\
{ OPObjectRef::_Set (obj, c##Ref::ReturnRuntime ()); return *this; } \
																\
inline c##Ref&													\
c##Ref::operator = (OPValueIn val)								\
{ OPObjectRef::_Set (val, c##Ref::ReturnRuntime ()); return *this; } \
																\
inline c* c##Ref::CheckAndResolve () const						\
{ return (c*)((this->object) ? this->object->CheckAndResolve ()	\
	: this->ReturnRuntime ()->ResolveNull ()); }				\
																\
inline c* c##Ref::Resolve () const								\
{ return (c*)((this->object) ? this->object->Resolve () : 0); } \
																\
inline c* c##Ref::operator -> () const							\
{ return (c*)(this->CheckAndResolve ()); }						\
																\
inline															\
c##Ref::c##Ref (c* ptr) : s##Ref (ptr) { }						\
																\
inline															\
c##Ref::c##Ref (c##In ref) : s##Ref (ref) { }					\
																\
inline c##Ref&													\
c##Ref::operator = (c* ptr)										\
{ OPObjectRef::_Set ((OPObject*)ptr); return *this; }			\
																\
inline c##Ref&													\
c##Ref::operator = (c##In ref)									\
{ OPObjectRef::_Set (ref); return *this; }

#define	OP_DEF_CLASS_END_2(c,s)									\
																\
inline															\
c##Ref::c##Ref (OPRuntimeIn rt, const OPaC_CallFrame* frame)	\
 : s##Ref (rt, frame) { }										\
																\
inline															\
c##Ref::c##Ref ()												\
 : s##Ref (c##Ref::ReturnRuntime (), OPaC_CallFrame::ReturnCallFrame ()) { } \
																\
struct c##s##dummy { int a


#define	OP_DEFINE_CLASS(c,s)									\
	OP_DEF_CLASS_BEGIN(c,s);									\
};																\
	OP_DEF_CLASS_END(c,s)

#define	OP_DEFINE_CLASS_SPECIAL_REF(c,s)						\
	OP_DEF_CLASS_BEGIN(c,s);									\
};																\
	OP_DEF_CLASS_END_2(c,s)

/****************************************************************/

#define	OP_IMPLEMENT_COMMON(c)									\
																\
c##Ref OPaC_ValueTo##c (OPValueIn val)							\
{ return (c##Ref)(val); }										\
																\
static OPRuntimeRef local_opac_runtime (0);						\
																\
const OPRuntimeRef& c::ReturnRuntime () const					\
{ return local_opac_runtime; }									\
																\
const OPRuntimeRef& c##Ref::ReturnRuntime ()					\
{ return local_opac_runtime; }									\
																\
Card32 c##Ref::SizeOf () const { return sizeof (c); }			\
																\
c* c##Ref::ReturnRawObject () const								\
{ return (c*)(OPObjectRef::ReturnRawObject ()); }				\
																\
void c::_Destroy ()												\
{ this->~c (); }												\
																\
static const LatinChar class_##c##_name[] = #c;					\
																\
Bool															\
c##Ref::operator == (const c##In ref) const						\
{ return (ref.Resolve () == this->Resolve ())					\
   /* || ((*this)->IsEqual (ref)) */							\
		 ? TRUE : FALSE; }										\
																\
Bool															\
c##Ref::operator != (const c##In ref) const						\
{ return (ref.Resolve () == this->Resolve ())					\
   /* || ((*this)->IsEqual (ref)) */							\
		 ? FALSE : TRUE; }										\
																\
c##Ref::~c##Ref () { }											\
c##Array::~c##Array () { }										\
																\
const c##Ref c##Null;											\
OPaC_TypeDef c##_type = { #c, 0, OP_TYPE_OBJECT, &(c##Ref::ReturnRuntime ()), 0 };	\
OPTypeDef	 c##TypeDef = & c##_type;							\
OPaC_TypeInfo c##_reg_type (&c##_type);							\
OPaC_TypeDef c##Array_type = { #c "Array", 0, OP_TYPE_ARRAY, &(OPArrayRef::ReturnRuntime ()), 0 };	\
OPTypeDef    c##ArrayTypeDef = & c##Array_type;					\
OPaC_TypeInfo c##Array_reg_type (&c##Array_type)

#define	OP_IMPLEMENT_CLASS(c)									\
																\
Bool c::IsAbstractClass () const								\
{ return FALSE; }												\
																\
Bool c::IsSpecialClass () const									\
{ return FALSE; }												\
																\
OP_IMPLEMENT_COMMON(c);											\
																\
void c::Construct (void* ptr, Card32 size)						\
{ (void)(new (local_opac_runtime, ptr, size, __LINE__, __OPaC_src_file) c); }	\
																\
void c::InitConstruct (void* ptr, Card32 size)					\
{ (void)(new (local_opac_runtime, ptr, size, __LINE__, __OPaC_src_file)\
			  c (ptr, size)); }									\
																\
OPStorageIn operator >> (OPStorageIn storage, c##Out ref)		\
{ storage->RestoreObject (ref, OP_RUNTIME (c));					\
  return storage; }												\
																\
static OPaC_RuntimeInfo g_runtime_info							\
	( local_opac_runtime, c::ReturnSuperRuntime,				\
	  class_##c##_name, __OPaC_src_file, sizeof (c),			\
	  c::Construct, c::InitConstruct,							\
	  FALSE, FALSE )

/*
 *	This macro replaces the OP_IMPLEMENT_CLASS macro when
 *	the class is an abstract one.
 */

#define	OP_ABSTRACT_CLASS(c)									\
																\
Bool c::IsAbstractClass () const								\
{ return TRUE; }												\
																\
Bool c::IsSpecialClass () const									\
{ return FALSE; }												\
																\
OP_IMPLEMENT_COMMON(c);											\
																\
void c::Construct (void* ptr, Card32 size)						\
{ OP_ABORT ( "Cannot instanciate " #c ); }						\
																\
void c::InitConstruct (void* ptr, Card32 size)					\
{ OP_ABORT ( "Cannot instanciate " #c ); }						\
																\
OPStorageIn operator >> (OPStorageIn storage, c##Out ref)		\
{ storage->RestoreObject (ref, OP_RUNTIME (c));					\
  return storage; }												\
																\
static OPaC_RuntimeInfo g_runtime_info							\
	( local_opac_runtime, c::ReturnSuperRuntime,				\
	  class_##c##_name, __OPaC_src_file, sizeof (c),			\
	  c::Construct, c::InitConstruct,							\
	  TRUE, FALSE )

/*
 *	This macro replaces the OP_IMPLEMENT_CLASS macro when
 *	the class is a special one.
 *
 *	There will be no allocation done automatically by the
 *	class manager. The restoring of data will not be done
 *	automatically...
 *
 *	This is useful for the OPAtom class !
 */

#define	OP_SPECIAL_CLASS(c)										\
																\
OP_IMPLEMENT_COMMON(c);											\
																\
Bool c::IsAbstractClass () const								\
{ return FALSE; }												\
																\
Bool c::IsSpecialClass () const									\
{ return TRUE; }												\
																\
void c::Construct (void* ptr, Card32 size)						\
{ (void)(new (local_opac_runtime, ptr, size, __LINE__, __OPaC_src_file) c); }	\
																\
void c::InitConstruct (void* ptr, Card32 size)					\
{ (void)(new (local_opac_runtime, ptr, size, __LINE__, __OPaC_src_file)\
			  c (ptr, size)); }									\
																\
static OPaC_RuntimeInfo g_runtime_info							\
	( local_opac_runtime, c::ReturnSuperRuntime,				\
	  class_##c##_name, __OPaC_src_file, sizeof (c),			\
	  c::Construct, c::InitConstruct,							\
	  FALSE, TRUE )

/****************************************************************/

#define	OP_MORE_IMPLEMENTATION(c)								\
																\
static const LatinChar class_##c##_name[] = #c

/****************************************************************/

/*
 *	The OPaC Runtime Information class is just used during
 *	start-up to record data exported by the implementation
 *	macros.
 *
 *	The data contained in this class is then use by the
 *	class manager in order to set up the real run-time
 *	information instances.
 */

class OP_PUBLIC OPaC_RuntimeInfo
{
	friend class OPRuntime;
	friend class OPClassMan;
	
private:
	static OPaC_RuntimeInfo* last;				//	last recorded run-time information
	
protected:
	OPaC_RuntimeInfo*	prev;					//	previously recorded run-time information
    
    OPRuntimeRef*		runtime;				//	pointer to original run-time info in class
    
	OPaC_ReturnRuntimeFunc runtime_func;		//	function returning super-class's OPRuntimeRef
	OPaC_ConstrFunc		constr_func;			//	function constructing an instance of this class
	OPaC_ConstrFunc		init_constr_func;		//	function constructing a dummy instance (booting)
	
	const LatinChar*	class_name;				//	class name
	const LatinChar*	file_name;				//	implementation file name
	
	Card32				instance_size;			//	size of an instance of this class
	
	Bool				is_abstract;			//	TRUE => may not allocate instances of this class
	Bool				is_special;				//	TRUE => allocation done by reference class only
	
public:
	OPaC_RuntimeInfo (OPRuntimeInOut runtime,
					  OPaC_ReturnRuntimeFunc runtime_func,
					  const LatinChar* class_name,
					  const LatinChar* file_name,
					  Card32 instance_size,
					  OPaC_ConstrFunc constr_func,
					  OPaC_ConstrFunc iconstr_func,
					  Bool is_abstract,
					  Bool is_special);
};

/****************************************************************/

extern OP_PUBLIC const OPObjectRef OPObjectNull;
extern OP_PUBLIC OPaC_TypeDef OPObject_type;
extern OP_PUBLIC OPTypeDef    OPObjectTypeDef;

extern OP_PUBLIC OPObjectRef OPaC_ValueToOPObject (OPValueIn val);

/****************************************************************/

/*
 *	The Object reference class encapsulates a pointer to
 *	some instance, allowing OPaC to track object usage.
 */

class OP_PUBLIC OPObjectRef
{
	friend class OPObject;
	friend class OPStorage;
	friend OP_PUBLIC OPStorageIn operator >> (OPStorageIn, OPObjectOut);
	friend class OPClassMan;
	friend class OPRuntime;
	friend class OPMessage;
	friend class OPMemZone;
	
	//	The ObjectRef class needs also some "runtime" static field, but it cannot
	//	be defined here, since the OPRuntimeRef class isn't defined yet. That's
	//	why "OPObjectRef", "operator =" and "ReturnRuntime" methods aren't inlined
	//	neither.
	
private:
	static Bool report_creation;				//	TRUE => notify Class Manager on creation
	
	OPObjectRef*		next;					//	next reference to same object
	OPObjectRef*		prev;					//	previous reference to same object
	const LatinChar*	c_name;					//	pointer to class name (if object is valid)
	OPObject*			embedded_in_object;		//	object in which this reference is embedded
	const OPaC_CallFrame* frame;				//	frame where reference resides (if any)
	
#ifdef _DEBUG
public:
	static OPObjectRef*	g_head;					//	head of global reference list
	static OPObjectRef*	g_tail;					//	tail of global reference list
	OPObjectRef*		g_next;					//	next element in global reference list
	OPObjectRef*		g_prev;					//	previous element in global reference list
	OPaC_CallFrame		frame_copy;
#endif
	
protected:
	OPObject*			object;					//	Ref classes are just smart pointer wrappers
	
	void _Set (OPObjectIn obj, const OPaC_CallFrame* frame = 0);
	void _Set (OPObjectIn obj, OPRuntimeIn runtime, const OPaC_CallFrame* frame = 0);
	void _Set (OPValueIn obj, const OPaC_CallFrame* frame = 0);
	void _Set (OPValueIn obj, OPRuntimeIn runtime, const OPaC_CallFrame* frame = 0);
	
	OPObjectRef (OPRuntimeIn rt, const OPaC_CallFrame* frame);
	
	static void ReportCreation (OPObjectRef* ref, OPRuntimeIn rt);
	static void RequireDeletion (OPObject*& obj);
	
public:
	
	//	Some methods get implemented automatically by OP_DEFINE_CLASS_END and
	//	will be inlined in the header file. Since OPObjectRef has to be defined
	//	manually, the methods have to be put in here :
	
	OPObjectRef ();
	OPObjectRef (OPObject* ptr);
	OPObjectRef (OPObjectIn ref);
	virtual ~ OPObjectRef ();
	
	static const OPRuntimeRef& ReturnRuntime ();
	static void RealiseDeletions ();
	
	OPObject* CheckAndResolve () const;
	OPObject* Resolve () const;
	OPObject* operator -> () const;
	
	//	For debugging, etc.
	
	Address ReturnRawAddress () const;
	OPObject* ReturnRawObject () const;
	const LatinChar* ReturnClassName () const;
	OPObject* ReturnEmbeddingObject () const;
	const OPaC_CallFrame* ReturnEmbeddingFrame () const;
	
	OPObjectRef& operator = (OPObject* ptr)		{ OPObjectRef::_Set (ptr); return *this; }
	OPObjectRef& operator = (OPObjectIn obj);
	OPObjectRef& operator = (OPValueIn obj);
	
	Bool IsValid () const;
	Bool IsInvalid () const;
	Bool IsPhantom () const;
	
	Bool operator == (OPObjectIn ref) const;
	Bool operator != (OPObjectIn ref) const;
	
	//	You may send messages to objects using some special, OPaC-specific,
	//	construct : the "ref <- msg" notation. That's why we need the following
	//	operator.
	
	Bool operator < (OPMessageIn msg) const;
	
	virtual Card32 SizeOf () const;
};

/****************************************************************/

OP_EXPORT (OPRuntime, OPObject);
OP_EXPORT (OPTarget, OPObject);
OP_EXPORT (OPText, OPObject);

/****************************************************************/

OP_TYPECLASS (OPStorageSync);

class OP_PUBLIC OPStorageSync
{
	friend OPStorageRef;
	
protected:
	const LatinChar*	internal;
	mutable OPObjectRef	name;
	mutable Bool		is_name_ok;
	
	void UpdateInternalName () const;
	
public:
	OPStorageSync (const LatinChar* name);
	~OPStorageSync ();
	
	static const OPStorageSync mark;
	static const OPStorageSync begin;
	static const OPStorageSync repeat;
	static const OPStorageSync end;
	static const OPStorageSync end_repeat;
	static const OPStorageSync type;
};

/****************************************************************/

OP_DEFINE_REF_BEGIN (OPStorage, OPObject);
	
	OPStorageIn operator << (OPObjectIn object) const;
	OPStorageIn operator << (OPAtomIn atom) const;
	OPStorageIn operator << (OPArrayIn object) const;
	
	OPStorageIn operator << (Card8 x) const;
	OPStorageIn operator << (Card16 x) const;
	OPStorageIn operator << (Card32 x) const;
	OPStorageIn operator << (Card64 x) const;
	OPStorageIn operator << (Int8 x) const;
	OPStorageIn operator << (Int16 x) const;
	OPStorageIn operator << (Int32 x) const;
	OPStorageIn operator << (Int64 x) const;
	OPStorageIn operator << (Real x) const;
	OPStorageIn operator << (OPStorageSyncIn x) const;
	OPStorageIn operator << (OPColorPixelIn x) const;
	
	OPStorageIn operator >> (Card8& x) const;
	OPStorageIn operator >> (Card16& x) const;
	OPStorageIn operator >> (Card32& x) const;
	OPStorageIn operator >> (Card64& x) const;
	OPStorageIn operator >> (Int8& x) const;
	OPStorageIn operator >> (Int16& x) const;
	OPStorageIn operator >> (Int32& x) const;
	OPStorageIn operator >> (Int64& x) const;
	OPStorageIn operator >> (Real& x) const;
	OPStorageIn operator >> (OPStorageSyncIn x) const;
	OPStorageIn operator >> (OPColorPixelOut x) const;
	
OP_DEFINE_REF_END (OPStorage, OPObject);

/****************************************************************/

/*
 *	When reporting a change, the ID can either be a valid one
 *	inherited from the caller or one of the two following two:
 *
 *	- OP_REPORT_AUTO, the change is reported automatically and
 *	  should not lead to a new unique ID : use the latest valid
 *	  ID instead.
 *
 *	- OP_REPORT_NEW, the change is a new one and should lead
 *	  to a new unique ID.
 */

#define	OP_REPORT_AUTO		((Count)(0))
#define	OP_REPORT_NEW		((Count)(1))

#define	OP_REPORT_MIN_ID	((Count)(2))
#define	OP_REPORT_MAX_ID	((Count)(~0))

/****************************************************************/

/*
 *	The OPaC Object class allows to attach targets (delegates,
 *	observers) and will also provide some functionality needed
 *	by the container classes.
 */

class OP_PUBLIC OPObject
{
	friend class OPStorage;						//	used to access Store/RestoreInstance
	friend class OPArchive;
	friend class OPClassMan;
	friend class OPObjectRef;
	friend class OPAtom;
	friend class OPMemZone;
	
public:
	struct MoreInfo;
	
public:
	static Count			global_report_id;
	static OPaC_DebugHook*	debug_hook;
	
private:
	MoreInfo*			more;					//	more (private) info, if needed
	OPMemZone*			mem_zone;				//	associated memory zone
	Count				current_report_id;		//	current report ID used by "Report"
	
	Count				ref_count;				//	# of references
	OPObjectRef*		ref_head;				//	pointer to first reference
	OPObjectRef*		ref_tail;				//	pointer to last reference
	
private:
	OPObject (const OPObject&);					//	no implementation
	void operator = (const OPObject&);			//	no implementation
	
protected:
	OPObject ();
	OPObject (OPMemZone*);
	OPObject (void* ptr, Card32 size);
	virtual ~ OPObject ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
	OPMemZone* ReturnMemZone () const			{ return this->mem_zone; }
	Bool IsObjectConnected () const				{ return this->more != 0; }
	
	Bool QuickReport (OPAtomIn event, Count id = OP_REPORT_AUTO)
		{ return (this->more == 0) ? FALSE : this->Report (event, id); }
	
	Bool QuickReport (const LatinChar* event, Count id = OP_REPORT_AUTO)
		{ return (this->more == 0) ? FALSE : this->Report (event, id); }
	
	Bool QuickReportChange (Count id = OP_REPORT_AUTO)
		{ return (this->more == 0) ? FALSE : this->ReportChange (id); }
	
	Bool QuickReportEmbeddedChange (OPObjectRef& ref, Count id = OP_REPORT_AUTO)
		{ return (this->more == 0) ? FALSE : this->ReportEmbeddedChange (ref, id); }
	
	Bool GetUnknownData (OPObjectOut unknown) const;
	Bool SetUnknownData (OPObjectIn unknown);
	
public:
	
	//	The following methods are usually added automatically by the definition
	//	macro :
	
	virtual OPRuntimeRet ReturnRuntime () const;
	virtual Bool IsAbstractClass () const;
	virtual Bool IsSpecialClass () const;
	virtual void _Destroy ();
	virtual OPObject* Resolve ();
	virtual OPObject* CheckAndResolve ();
	
	static void Construct (void* ptr, Card32 size);
	static void InitConstruct (void* ptr, Card32 size);
	
	//	OPObject's methods :
	
	virtual void Awake ();
	virtual void AwakeFromApp (OPAppIn app);
	virtual void AwakeFromObject (OPObjectIn object);
	
	virtual OPObjectRef ReturnSelf ();
	
	virtual Bool IsKindOf (OPRuntimeIn) const;
	virtual Bool IsA (OPRuntimeIn) const;
	
	//	Delete does neither free nor destroy the object, it just removes the
	//	connections with the other objects, so that its ref count will reach
	//	zero...
	//
	//	Each class should provide its own deletion method, so that the base
	//	class won't have too many ref's to cope with.
	
	virtual void Delete ();
	virtual void MarkAsDead ();
	virtual void MarkAsDeadButOne (OPObjectIn keep);
	
	//	The following methods are used to sort and store data into container
	//	classes. They should always be overridden by the subclass !
	
	virtual Count ComputeHashValue (Count num) const;
	virtual int   CompareWith (OPObjectIn object) const;
	virtual Bool  IsEqual (OPObjectIn object) const;
	
	virtual OPObjectRef Clone () const;
	virtual OPObjectRef Copy (OPObjectIn copy);
	
	//	Report methods.
	
	virtual Bool Report (OPAtomIn event, Count id = OP_REPORT_AUTO);
	virtual Bool Report (const LatinChar* event, Count id = OP_REPORT_AUTO);
	virtual Bool ReportChange (Count id = OP_REPORT_AUTO);
	virtual Bool ReportEmbeddedChange (OPObjectRef& ref, Count id = OP_REPORT_AUTO);
	
	virtual Count GetCurrentReportID () const;
	virtual void SetCurrentReportID (Count id);
	static Count NewReportID ();
	
	//	When observers have been attached to an object, they may be activated
	//	through the reporter methods. The attached observers can be listed by
	//	using the "CountAttached" and "GetAttached" methods, whereas observers
	//	published by the object can be listed with "Count/GetObservers".
	
	virtual void AttachObserver (OPObjectIn obs, OPAtomIn event, Card32 id = 0);
	virtual void AttachObserver (OPObjectIn obs, const LatinChar* event, Card32 id = 0);
	virtual void AttachObserver (OPObjectIn obs, OPAtomIn event, OPAtomIn actor, Card32 id = 0);
	virtual void AttachObserver (OPObjectIn obs, const LatinChar* event, const LatinChar* actor, Card32 id = 0);
	virtual void DetachObserver (OPObjectIn obs, OPAtomIn event, Card32 id = 0);
	virtual void DetachObserver (OPObjectIn obs, const LatinChar* event, Card32 id = 0);
	virtual void DetachObserver (OPObjectIn obs, OPAtomIn event, OPAtomIn actor, Card32 id = 0);
	virtual void DetachObserver (OPObjectIn obs, const LatinChar* event, const LatinChar* actor, Card32 id = 0);
	
	static void AttachObserver (OPObjectIn to, OPObjectIn obs, OPAtomIn event, OPAtomIn actor, Card32 id);
	
	virtual void DetachObserver (OPObjectIn obs);
	virtual void DetachSelfFromAll ();
	
	//	Other objects may get informations at run time about the available
	//	services an object provides, using some simple minded enumerators...
	
	virtual Count CountObservers () const;
	virtual Count CountActors () const;
	virtual Count CountMethods () const;
	virtual Count CountPublicData () const;
	virtual Count CountAttached () const;
	virtual Count CountAttributes () const;
	
	virtual Bool GetObserver (Count index, OPAtomOut name) const;
	virtual Bool GetActor (Count index, OPAtomOut name) const;
	virtual Bool GetMethod (Count index, OPAtomOut name) const;
	virtual Bool GetMethod (Count index, OPAtomOut name, Card32Out param_num, const OPaC_ParamDesc* param_array) const;
	virtual Bool GetPublicData (Count index, OPAtomOut name) const;
	virtual Bool GetPublicData (Count index, OPAtomOut name, OPTypeInfoOut type) const;
	virtual Bool GetPublicData (Count index, OPAtomOut name, OPStringOut desc, OPTypeInfoOut type) const;
	virtual Bool GetPublicDataNotifier (OPAtomIn name, Count nth_notifier, OPAtomOut notifier) const;
	virtual Bool GetAttached (Count index, OPAtomOut event, OPObjectOut obs, Card32Out id) const;
	virtual Bool GetAttached (Count index, OPAtomOut event, OPAtomOut actor, OPObjectOut obs, Card32Out id) const;
	
	virtual Bool SetAttribute (const LatinChar* key);
	virtual Bool SetAttribute (OPObjectIn key);
	virtual Bool SetAttribute (const LatinChar* key, OPObjectIn value);
	virtual Bool SetAttribute (OPObjectIn key, OPObjectIn value);
	virtual Bool RemoveAttribute (const LatinChar* key, Count nth = 0);
	virtual Bool RemoveAttribute (OPObjectIn key, Count nth = 0);
	virtual Bool RemoveAttribute (Count index);
	virtual Bool GetAttribute (const LatinChar* key, Count nth = 0) const;
	virtual Bool GetAttribute (OPObjectIn key, Count nth = 0) const;
	virtual Bool GetAttribute (const LatinChar* key, OPObjectOut value, Count nth = 0) const;
	virtual Bool GetAttribute (OPObjectIn key, OPObjectOut value, Count nth = 0) const;
	virtual Bool GetAttribute (Count index, OPObjectOut key, OPObjectOut value) const;
	
	//	Information which can be retrieved by iterations can also be collected
	//	automatically into a container using following methods :
	
	virtual Bool GetObserver (OPContainerOut info) const;
	virtual Bool GetActor (OPContainerOut info) const;
	virtual Bool GetMethod (OPContainerOut info) const;
	virtual Bool GetPublicData (OPContainerOut info) const;
	virtual Bool GetAttached (OPContainerOut info) const;
	
	//	Perform some action. This is used to provide some simple and efficient
	//	dynamic method calling, without any parameters. More elaborate message
	//	sending is done by the OPMessage class, using the "<-" construct...

	virtual Bool Action (OPAtomIn event);
	virtual Bool Action (const LatinChar* event);
	
	//	The Data Report class registers with its targets. This guarantees that
	//	these connections are saved if originating from the "outer world".
	
	virtual Bool RegisterAsTargetOf (OPObjectIn initiator);
	virtual Bool UnregisterAsTargetOf (OPObjectIn initiator);
	
	//	Objects may export and/or import some data -- using data fields -- which
	//	can be accessed using the "QueryData" and "DefineData" methods, depending
	//	on the data type. These have to be overridden by the subclasses if special
	//	functionality is needed.
	
	virtual Bool QueryData (OPAtomIn name, OPObjectOut data) const;
	virtual Bool QueryData (OPAtomIn name, OPValueOut data) const;
	virtual Bool QueryData (Count index, OPObjectOut data) const;
	virtual Bool QueryData (Count index, OPValueOut data) const;
	
	virtual Bool DefineData (OPAtomIn name, OPObjectIn data);
	virtual Bool DefineData (OPAtomIn name, OPValueIn data);
	virtual Bool DefineData (Count index, OPObjectIn data);
	virtual Bool DefineData (Count index, OPValueIn data);
	
	//	It is possible to query a tree (such as o.a.b.c, where "o" is the root
	//	specified as an object and "a.b.c" the branch described by nodes "a",
	//	"b" and leaf "c"). The elements of the path are delimited by a dot "."
	
	static Bool Action (OPObjectIn o, OPStringIn path);
	static Bool QueryData (OPObjectIn o, OPStringIn path, OPObjectOut data);
	static Bool QueryData (OPObjectIn o, OPStringIn path, OPValueOut data);
	static Bool DefineData (OPObjectIn o, OPStringIn path, OPObjectIn data);
	static Bool DefineData (OPObjectIn o, OPStringIn path, OPValueIn data);
};

/****************************************************************/

/*
 *	The run-time class information is stored in the OPaC
 *	Runtime class. Use the member functions to access the
 *	data.
 */

class OP_PUBLIC OPRuntime : public OPObject
{
	friend class OPClassMan;
	friend class OPObjectRef;
	
private:
	struct TargetCache
	{
		OPaC_Target*	target;					//	corresponding target (nth defined by index)
		Count			index;					//	index into target list
		OPRuntime*		runtime;				//	runtime info about target type
	};

	struct NotifyInfo;
	
private:
	OPRuntimeRef		self;					//	reference to self
	OPRuntimeRef		super_class;			//	super class run-time information
	
	OPaC_RuntimeInfo*	info;					//	pointer to lower-level data
	
	OPaC_Target*		target_list;			//	list of run-time defined targets
	Count				target_num;				//	# targets defined at run-time
	TargetCache			target_cache;
	
	OPaC_PubData*		data_list;				//	list of published data
	OPaC_PubData**		data_array;				//	array of published data
	Count				data_num;				//	# published data
	
	NotifyInfo*			notify_list;			//	list of registered notification events
	Count				notify_num;				//	# registered notification events

	//	Fields used by the class manager :
	
	OPRuntime*			next;					//	next class
	
protected:
	OPRuntime ();
	~OPRuntime ();
	
	void Initialise (OPaC_RuntimeInfo* info);
	void LinkSuper ();
	void CollectTargets ();
	void CompileWhileInit ();
	void CompileAfterInit ();
	
public:
	virtual void Delete ();						/* OVERRIDDEN */
	virtual OPObject* ResolveNull ();
	
	//	Don't use following methods :
	
	virtual void PublishData (OPaC_PubDataDef& def);
	virtual void PublishDataEx (OPaC_PubDataDef&, const OPaC_BitsetDataDef*, Card32 size);
	Bool ConstructObject (void* ptr, Card32 size) const;
	
	//	The following methods are publicly available; they allow accessing
	//	run-time information and manipulating instances :
	
	virtual Bool IsKindOf (OPRuntimeIn) const;
	virtual Bool IsA (OPRuntimeIn) const;
	
	const OPRuntimeRef& ReturnClass () const	{ return this->self; }
	const OPRuntimeRef& ReturnSuper () const	{ return this->super_class; }
	
	const LatinChar* ReturnClassName () const	{ return this->info->class_name; }
	const LatinChar* ReturnFileName () const	{ return this->info->file_name; }
	Card32 ReturnInstanceSize () const			{ return this->info->instance_size; }
	
	Count CountTargets () const					{ return this->target_num; }
	Count CountPubData () const					{ return this->data_num; }
	Count CountMethods (OPAtomIn) const;
	Count CountNotify () const					{ return this->notify_num; }
	Count CountNotify (OPAtomIn field) const;
	
	Bool GetTarget (Count, OPAtomOut, OPTargetTypeOut, const OPaC_Target* & );
	Bool GetPubData (Count, const OPaC_PubData* & );
	Bool GetNotify (Count, OPAtomOut);
	Bool GetNotify (OPAtomIn field, Count, OPAtomOut);
	
	const OPaC_Target* FindMethod (OPAtomIn, Count nth = 0) const;
	const OPaC_Target* FindObserver (OPAtomIn) const;
	const OPaC_Target* FindActor (OPAtomIn) const;
	const OPaC_PubData* FindPubData (OPAtomIn, Count nth = 0) const;
	
	Bool IsAbstract () const					{ return this->info->is_abstract; }
	Bool IsSpecial () const						{ return this->info->is_special; }
	
	OP_DEFINE_CLASS (OPRuntime, OPObject);
};

/****************************************************************/

/*
 *	The storage class is a generic class used to store
 *	and retrieve data from some sequential record file.
 *
 *	An archive and a message are both typical examples.
 */

class OP_PUBLIC OPStorage : public OPObject
{
	
protected:
	enum StorageMode
	{
		STORE_BIN_ONLY = 0,						//	do not accept objects
		STORE_SHALLOW  = 1,						//	store simple references to objects
		STORE_DEEP     = 2						//	store complete references to objects
	};
	
private:
	struct RefBase;
	struct NameBase;
	struct StorageInfos;
	
private:
	StorageMode			storage_mode;			//	current storage mode
	StorageInfos*		storage_infos;			//	additional storage informations
	
	RefBase*			r_base;					//	reference data base (both for store/restore)
	NameBase*			n_base;					//	name data base (both for store/restore)
	
	Count				num_pending_store;		//	current # of pending stores
	Count				num_recursion;			//	current # of recursions
	
	Byte				push_buffer[64];		//	push back buffer
	Card32				push_size;				//	size of data pushed back
	
	Card32				id_base;				//	base for store/restore of IDs
	
private:
	Card32 LookForRef (OPObjectIn obj, BoolOut is_new);
	Card32 LookForName (const LatinChar* name, BoolOut is_new);
	
	void ResolveRef (Card32 id, OPObjectOut obj);
	void ResolveName (Card32 id, const LatinChar*& name);
	
	Bool FindNameFromSlot (OPObjectIn object, OPAtomOut name);
	Bool FindObjectFromSlot (OPAtomIn name, OPObjectOut object);
	
	void PopAndRestoreData (void* data, Card32 size);
	void PushBackData (void* data, Card32 size);
	void StoreID (Card32 id);
	Card32 RestoreID ();
	
protected:
	OPStorage ();
	virtual ~ OPStorage ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	//	Internal accessor methods :
	
	void SetStorageMode (StorageMode mode);
	StorageMode ReturnStorageMode () const;
	
	virtual void RestoreToMark (Card8 mark);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */

	//	The data slots can be used to attach the restored objects dynamically to
	//	other ones...
	
	virtual void SetDataSlot (OPDataSlotIn slot);
	virtual void RemoveDataSlot (OPDataSlotIn slot);
	virtual OPDataSlotRet ReturnDataSlot (Count index = 0) const;
	
	//	The storage class provides the base mechanisms used to store objects and
	//	miscellanous simple data types. They will all be converted into binary
	//	data, which will have to be handled by the child classes.
	
	virtual void StoreObject (OPObjectIn object);
	virtual void StoreArray (OPArrayIn array);
	virtual void StoreAtom (OPAtomIn atom);
	virtual void StoreAtomArray (OPAtomArrayIn array);
	virtual void StoreString (OPStringIn text);
	virtual void StorePhantom (OPPhantomIn phantom);
	
	virtual void RestoreObject (OPObjectOut object, OPRuntimeIn runtime);
	virtual void RestoreArray (OPArrayOut array, OPRuntimeIn runtime);
	virtual void RestoreAtom (OPAtomOut atom);
	virtual void RestoreAtomArray (OPAtomArrayOut array);
	virtual void RestoreString (OPStringOut text);
	virtual void RestorePhantom (OPPhantomOut phantom);
	
	virtual void FinallyAwake ();
	virtual void FinallyAwakeFromApp (OPAppIn app);
	virtual void FinallyAwakeFromObject (OPObjectIn object);
	
	//	Don't be afraid of automatic type conversion, this will work even with
	//	some "storage << 10", because every integer value is always saved with
	//	full 64-bit width (in a compressed way).
	
	virtual void StoreBool (Bool x);
	virtual void StoreInt (Int64 x);
	virtual void StoreCard (Card64 x);
	virtual void StoreReal (Real x);
	virtual void StoreSyncMark ();
	virtual void StoreBlockBegin ();
	virtual void StoreBlockEnd ();
	virtual void StoreBinData (const void* data, Card32 size);
	
	virtual Bool   RestoreBool ();
	virtual Int64  RestoreInt ();
	virtual Card64 RestoreCard ();
	virtual Real   RestoreReal ();
	virtual void   RestoreToSyncMark ();
	virtual void   RestoreBlockBegin ();
	virtual void   RestoreToBlockEnd ();
	virtual void   RestoreBinData (void* data, Card32 size);
	
	virtual Bool CheckForMark ();
	virtual Bool CheckForBegin ();
	virtual Bool CheckForEnd ();
	
	//	Some child class has to provide the binary input/output methods :
	
	virtual void StoreData (const void* data, Card32 size) = 0;
	virtual void RestoreData (void* data, Card32 size) = 0;
	
	OP_DEFINE_CLASS (OPStorage, OPObject);
};

/****************************************************************/

inline OPStorageIn OPStorageRef::operator << (OPObjectIn object) const
{ (*this)->StoreObject (object); return (*this); }

inline OPStorageIn OPStorageRef::operator << (OPArrayIn array) const
{ (*this)->StoreArray (array); return (*this); }

inline OPStorageIn OPStorageRef::operator << (OPAtomIn atom) const
{ (*this)->StoreAtom (atom); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Card8 x) const
{ (*this)->StoreCard (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Card16 x) const
{ (*this)->StoreCard (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Card32 x) const
{ (*this)->StoreCard (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Card64 x) const
{ (*this)->StoreCard (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Int8 x) const
{ (*this)->StoreInt (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Int16 x) const
{ (*this)->StoreInt (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Int32 x) const
{ (*this)->StoreInt (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Int64 x) const
{ (*this)->StoreInt (x); return (*this); }

inline OPStorageIn OPStorageRef::operator << (Real x) const
{ (*this)->StoreReal (x); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Card8& x) const
{ x = (Card8)(*this)->RestoreCard (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Card16& x) const
{ x = (Card16)(*this)->RestoreCard (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Card32& x) const
{ x = (Card32)(*this)->RestoreCard (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Card64& x) const
{ x = (Card64)(*this)->RestoreCard (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Int8& x) const
{ x = (Int8)(*this)->RestoreInt (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Int16& x) const
{ x = (Int16)(*this)->RestoreInt (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Int32& x) const
{ x = (Int32)(*this)->RestoreInt (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Int64& x) const
{ x = (Int64)(*this)->RestoreInt (); return (*this); }

inline OPStorageIn OPStorageRef::operator >> (Real& x) const
{ x = (*this)->RestoreReal (); return (*this); }

/****************************************************************/

/*
 *	The Text class is just an abstract class which provides
 *	a canvas for the Atom and String classes.
 */

class OP_PUBLIC OPText : public OPObject
{
	
protected:
	OPText ();
	virtual ~ OPText ();
	
public:
	virtual OPStringRef ReturnString () const = 0;
	virtual Bool SetString (OPStringIn text) = 0;
	
	virtual const LatinChar* ReturnQuickLatin () const = 0;
	virtual const UnicodeChar* ReturnQuickUnicode () const = 0;
	
	OP_DEFINE_CLASS (OPText, OPObject);
};

/****************************************************************/

#endif /* _OPaC_OBJECT_H_ */
