/*
 *	objectpane.h
 *
 *	The Object Pane class represents objects in a grid, in a visual
 *	way (like files in a browser).
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_OBJECTPANE_H_
#define	_OPaC_OBJECTPANE_H_

#ifndef	_OPaC_SCROLLGRID_H_
#include <OPaC/scrgrid.h>
#endif

#ifndef	_OPaC_ICOSTYLE_H_
#include <OPaC/icostyle.h>
#endif

/****************************************************************/

/*
 *	The objects can be represented in different ways and
 *	sorted vertically, horizontally, etc.
 */

OP_TYPEDEF (Card32, OPObjectPaneMode);			//	it is a bitset, therefore not an enum

#define	OP_PANE_MODE_MASK		0x000F
#define	OP_PANE_MODE_ICON_LARGE	0x0000			//	large icon with name underneath
#define	OP_PANE_MODE_ICON_SMALL	0x0001			//	small icon with name on the right
#define	OP_PANE_MODE_NAME_ONLY	0x0002			//	just the name

#define	OP_PANE_LAYOUT_MASK		0x00F0
#define	OP_PANE_LAYOUT_LINES	0x0000			//	display layout by lines
#define	OP_PANE_LAYOUT_COLUMNS	0x0010			//	display layout by columns

/****************************************************************/

/*
 *	The Object Pane helps representing objects visually.
 */

class OP_PUBLIC OPObjectPane : public OPScrollGrid
{

protected:
	OPObjectPaneMode	mode;					//	display mode for the objects
	OPGenericProviderRef provider;				//	object provider
	
	Count				num_objects;			//	total number of objects
	Count				num_columns;			//	number of columns in grid
	Count				num_lines;				//	number of lines in grid
	
	OPCoord				cell_dx;
	OPCoord				cell_dy;
	
	OPCoord				name_dx;
	OPCoord				name_dy;
	
	OPCoord				large_icon_dx;
	OPCoord				large_icon_dy;
	OPCoord				small_icon_dx;
	OPCoord				small_icon_dy;
	
protected:
	OPObjectPane ();
	virtual ~ OPObjectPane ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void CleanUp ();
	virtual void ComputeCellSize ();
	virtual void ComputeLayout ();
	virtual Bool GetCellLayout (Count index, CountOut x, CountOut y) const;
	virtual Bool CreateCell (Count index, OPControlOut cell);

public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Initialise (OPCellGridIn grid, OPRectIn box); /* OVERRIDDEN */
	virtual void Initialise (OPRectIn box, OPGenericProviderIn provider);
	virtual void Initialise (OPRectIn box, OPGenericProviderIn provider, OPAtomIn model);
	
	virtual Bool SetMode (OPObjectPaneMode mode);
	virtual Bool GetMode (OPObjectPaneModeOut mode) const;
	virtual OPObjectPaneMode ReturnMode () const;
	
	virtual Bool SetNameDX (OPCoord dx);
	virtual Bool SetNameDY (OPCoord dy);
	virtual OPCoord ReturnNameDX () const;
	virtual OPCoord ReturnNameDY () const;
	
	virtual Bool SetProvider (OPGenericProviderIn provider);
	virtual Bool GetProvider (OPGenericProviderOut provider) const;
	virtual OPGenericProviderRet ReturnProvider () const;
	
	virtual Bool GetRankFromPos (Count x, Count y, CountOut rank) const;
	virtual Bool GetPosFromRank (Count rank, CountOut x, CountOut y) const;
	
	virtual Bool Update ();
	virtual void GridChangedSize ();					/* OVERRIDDEN */
	
	OP_DEFINE_SLOT (OPGenericProvider, provider, SetProvider, ReturnProvider);
	OP_DEFINE_CLASS (OPObjectPane, OPScrollGrid);
};

/****************************************************************/

/*
 *	The generic provider is a base class which must be
 *	overridden in order to return the miscellaneous
 *	informations needed by the object pane.
 */

class OP_PUBLIC OPGenericProvider : public OPObject
{
protected:
	OPGenericProvider ();
	virtual ~ OPGenericProvider ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual Count ReturnNumObjects () = 0;
	
	virtual Bool GetObject (Count index, OPObjectOut object) = 0;
	virtual Bool GetText (Count index, OPStringOut text) = 0;
	virtual Bool GetLargeIcon (Count index, OPLookIconOut icon) = 0;
	virtual Bool GetSmallIcon (Count index, OPLookIconOut icon) = 0;
	virtual Bool GetMarkIcon (Count index, OPLookIconOut icon) = 0;
	
	virtual Bool GetCompleteView (Count index, OPCoord dx, OPCoord dy, OPControlOut cell);
	
	//	If a node gives access to other nodes (kind of subtree), we need a generalised
	//	browser. The best current browser is a provider.
	
	virtual Bool EnterSubtree (Count index, OPGenericProviderOut provider);
	virtual Bool ReleaseSubtree (Count index);
	
	OP_DEFINE_CLASS (OPGenericProvider, OPObject);
};

/****************************************************************/

/*
 *	Return informations about an object. This can be
 *	used to build an object browser.
 */

class OP_PUBLIC OPObjectProvider : public OPGenericProvider
{
protected:
	OPGenericContainerRef container;			//	container with the objects

protected:
	OPObjectProvider ();
	virtual ~ OPObjectProvider ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Initialise (OPGenericContainerIn container);
	
	virtual Bool GetContainer (OPGenericContainerOut container) const;
	virtual Count ReturnNumObjects ();
	
	virtual Bool GetObject (Count index, OPObjectOut object);
	virtual Bool GetText (Count index, OPStringOut text);
	virtual Bool GetLargeIcon (Count index, OPLookIconOut icon);
	virtual Bool GetSmallIcon (Count index, OPLookIconOut icon);
	virtual Bool GetMarkIcon (Count index, OPLookIconOut icon);
	
	OP_DEFINE_CLASS (OPObjectProvider, OPGenericProvider);
};

/****************************************************************/

/*
 *	Return informations taken from a data base.
 */

class OP_PUBLIC OPDBProvider : public OPObjectProvider
{
	
protected:
	OPAtomRef			db_sort_name;			//	db key used for sorting
	OPAtomRef			db_data_name;			//	db key containing the object (real data)
	OPAtomRef			db_text_name;			//	db key of the object's textual name
	OPAtomRef			db_large_icon_name;		//	db key of the large icon representation
	OPAtomRef			db_small_icon_name;		//	db key of the small icon representation
	OPAtomRef			db_mark_icon_name;		//	db key of the mark (usually also an icon)

protected:
	OPDBProvider ();
	virtual ~ OPDBProvider ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual OPLookIcon MapObjectToIcon (OPObjectIn object) const;
	
public:
	virtual Bool SetDBNameForSorting (OPAtomIn name);
	virtual Bool SetDBNameForObject (OPAtomIn name);
	virtual Bool SetDBNameForText (OPAtomIn name);
	virtual Bool SetDBNameForLargeIcon (OPAtomIn name);
	virtual Bool SetDBNameForSmallIcon (OPAtomIn name);
	virtual Bool SetDBNameForMarkIcon (OPAtomIn name);

	virtual Bool GetObject (Count index, OPObjectOut object);
	virtual Bool GetText (Count index, OPStringOut text);
	virtual Bool GetLargeIcon (Count index, OPLookIconOut icon);
	virtual Bool GetSmallIcon (Count index, OPLookIconOut icon);
	virtual Bool GetMarkIcon (Count index, OPLookIconOut icon);
	
	OP_DEFINE_CLASS (OPDBProvider, OPObjectProvider);
};

/****************************************************************/

#endif /* _OPaC_OBJECTPANE_H_ */
