/*
 *	view.h
 *
 *	Base View class definition. The view class is the base class for every
 *	visual interface element (widget).
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_VIEW_H_
#define	_OPaC_VIEW_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

#ifndef	_OPaC_LOOKCURSOR_H_
#include <OPaC/lookcursor.h>
#endif

/****************************************************************/

/*
 *	Each view has some capabilities and has some internal
 *	state. It is represented by the following bitset :
 */

OP_TYPEDEF (Card32, OPViewState);				//	it is a bitset, therefore not an enum

#define	OP_VIEW_IS_ENABLED	OP_BIT (0)			//	TRUE => view may get events
#define	OP_VIEW_IS_ALIVE	OP_BIT (1)			//	TRUE => view may become active
#define	OP_VIEW_IS_FROZEN	OP_BIT (2)			//	TRUE => view is frozen -- NIY
#define	OP_VIEW_IS_HOT		OP_BIT (3)			//	TRUE => view is hot -- NIY
#define	OP_VIEW_IS_ACTIVE	OP_BIT (4)			//	TRUE => view has keyboard focus
#define	OP_VIEW_IS_DRAGGING	OP_BIT (5)			//	TRUE => view is the current drag source
#define	OP_VIEW_IS_DROPPING	OP_BIT (6)			//	TRUE => view is the current drop destination
#define	OP_VIEW_IS_DISABLED	OP_BIT (7)			//	TRUE => view has been globally disabled
#define	OP_VIEW_IS_CONNECT	OP_BIT (8)			//	TRUE => view is the current connection destination
#define	OP_VIEW_IS_HIDDEN	OP_BIT (9)			//	TRUE => view is hidden

#define	OP_VIEW_OK_DRAG		OP_BIT (16)			//	TRUE => view accepts dragging
#define	OP_VIEW_OK_DROP		OP_BIT (17)			//	TRUE => view accepts dropping
#define	OP_VIEW_OK_DETECT	OP_BIT (18)			//	TRUE => view accepts detection
#define	OP_VIEW_OK_MAIN		OP_BIT (19)			//	TRUE => view accepts to become main responder
#define	OP_VIEW_OK_UI_EDIT	OP_BIT (20)			//	TRUE => view accepts user interface editing
#define	OP_VIEW_OK_UI_MOVE	OP_BIT (21)			//	TRUE => view accepts user interface moving
#define	OP_VIEW_OK_DRAW		OP_BIT (22)			//	TRUE => view uses standard Draw method

/****************************************************************/

OP_TYPEDEF (Card32, OPDrawOp);					//	it is a bitset

#define	OP_DRAW_NONE		((OPDrawOp)(0x00))	//	draw nothing
#define	OP_DRAW_OBJECT		((OPDrawOp)(0x01))	//	draw the object (foreground)
#define	OP_DRAW_BACKGROUND	((OPDrawOp)(0x02))	//	draw the background
#define	OP_DRAW_ALL			((OPDrawOp)(0x03))	//	draw everything (background + object)

/****************************************************************/

/*
 *	The drag & drop protocol defines what kind of operation is
 *	going on. The following are supported :
 *
 *	- User is dragging an object in order to drop it (real drag
 *	  and drop).
 *
 *	- User is dragging a connection point.
 */

#define	DEF_ENUM_BEGIN(T)	enum T {
#define	DEF_ENUM(T,n)		T = n
#define	DEF_ENUM_END(T)		};

#include <OPaC/enum/dragdrop.h>

OP_TYPEENUM (OPDragDropProtocol);

/****************************************************************/

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

/****************************************************************/

/*
 *	Base class for every responder object; a View object is
 *	basically a Responder... This means that the user interface
 *	is crowded with responders :-)
 */

class OP_PUBLIC OPResponder : public OPObject
{
	
private:
	OPResponderRef		next_responder;			//	link with next responder
	
protected:
	OPResponder ();
	virtual ~ OPResponder ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void SetNextResponder (OPResponderIn next);
	virtual Bool GetNextResponder (OPResponderOut next) const;
	
	virtual Bool GetView (OPViewOut view);
	
	//	Any responder may become the Main Responder. The Main Responder
	//	will get the key events first and will usually be the active
	//	responder in a window.
	
	virtual Bool WouldAcceptMain () const;
	virtual void BecomeMain (Bool clicked);
	virtual void ResignMain ();
	
	//	The HandleEvent should not be overridden, since it does the
	//	dispatching into the more specific event handling methods.
	//	Only OPWindow is supposed to override this method !
	
	virtual void DispatchEvent (OPEventIn event);
	
	//	The following methods will make the event handling easier :
	
	virtual void KeyDown (OPEventIn event);
	virtual void KeyChange (OPEventIn event);
	virtual void KeyUp (OPEventIn event);
	
	virtual void ButtonPress (OPEventIn event);
	virtual void ButtonRelease (OPEventIn event);
	
	virtual void MotionNotify (OPEventIn event);
	virtual void DragNotify (OPEventIn event);
	
	virtual void Tick ();
	virtual void Miscellanous (OPEventIn event);
	
	//	The following method cannot be called by the dispatching
	//	method. It will be called by the window or application.
	
	virtual void EnterNotify (OPEventIn event);
	virtual void LeaveNotify (OPEventIn event);
	
	OP_DEFINE_CLASS (OPResponder, OPObject);
};

/****************************************************************/

/*
 *	The OPaC View class describes a graphical element used
 *	in the user interface. It specifies a hierarchy and a
 *	geometry.
 */

class OP_PUBLIC OPView : public OPResponder
{
		
	struct OP_PUBLIC ViewChildren
	{
		OPViewRef		head;					//	first child
		OPViewRef		tail;					//	last child
	};
	
	struct OP_PUBLIC ViewSiblings
	{
		OPViewRef		next;					//	next sibling
		OPViewRef		prev;					//	previous sibling
	};
	
protected:
	static OPAtomRef	atom_look_man_changed;	//	atom "LookManChanged" (used by Broadcast)
	
	OPWindowRef			window;					//	link to containing window
	OPViewRef			parent;					//	link to parent view
	OPSpringBoardRef	spring;					//	link to associated spring curve

	ViewSiblings		sibling;				//	link to sibling views
	ViewChildren		child;					//	link to children views (subviews)
	
	OPListRef			active_list;			//	activation list of children
	
	OPRect				box;					//	window-relative bounding box
	OPViewState			state;					//	current state of the view (bitset)
	Int32				tag;					//	tag (can be used freely by programmer)
	OPAtomRef			name;					//	user friendly name
	
	Count				num_overlays;			//	# of overlays registered
	
	OPLookCursor		look_cursor;			//	associated look cursor
	OPAtomRef			named_cursor;			//	assocaited named cursor (if any)
	
	Bool				no_draw_back;			//	don't call OPView's drawing method
	
protected:
	OPView ();
	virtual ~ OPView ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void ChangeGeometry (OPRectIn box);
	virtual void SetEnableState (Bool ok, Bool broadcast = TRUE);
	
	virtual Bool RemoveActive (OPViewIn child);
	virtual OPListRet ReturnAndCreateActiveList ();
	virtual void NotifyActiveListChange ();
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	virtual Bool GetView (OPViewOut view);				/* OVERRIDDEN */
	virtual Bool WouldAcceptMain () const;				/* OVERRIDDEN */
	virtual void BecomeMain (Bool clicked);				/* OVERRIDDEN */
	virtual void ResignMain ();							/* OVERRIDDEN */

	virtual void MakeMainResponder ();
	
	//	The view has some complex relationship with other views, an associated
	//	geometry and lots of attributes. Here are the accessors :
	
	virtual Bool SetWindow (OPWindowIn window);
	virtual Bool SetParent (OPViewIn parent);
	virtual Bool SetSpringBoard (OPSpringBoardIn spring);

	virtual Bool WouldAcceptChild (OPViewIn child) const;
	virtual void NotifyInsertChild (OPViewIn child);
	virtual void NotifyRemoveChild (OPViewIn child);
	
	virtual void ReorderFirst ();
	virtual void ReorderLast ();
	virtual void ReorderBefore (OPViewIn which_view);
	virtual void ReorderAfter (OPViewIn which_view);
	
	virtual Bool SetBox (OPRectIn box);
	virtual void SetOrigin (OPCoord x, OPCoord y);
	virtual void SetSize (OPCoord dx, OPCoord dy);
	virtual void ChangeOrigin (OPRelCoord dx, OPRelCoord dy);
	virtual void ChangeSize (OPRelCoord dx, OPRelCoord dy);
	
	virtual Bool GetWindow (OPWindowOut window) const;
	virtual Bool GetParent (OPViewOut view) const;
	virtual Bool GetSpringBoard (OPSpringBoardOut spring) const;
	
	virtual OPWindowRet ReturnWindow () const;
	virtual OPViewRet ReturnParent () const;
	virtual OPSpringBoardRet ReturnSpringBoard () const;
	
	virtual Bool GetNextSibling (OPViewOut view) const;
	virtual Bool GetPrevSibling (OPViewOut view) const;
	virtual Bool GetFirstChild (OPViewOut view) const;
	virtual Bool GetLastChild (OPViewOut view) const;
	
	virtual void GetOrigin (OPCoordOut x, OPCoordOut y) const;
	virtual void GetSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetMinSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetMaxSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual OPRectRet ReturnBox () const;
	virtual Bool GetBox (OPRectOut box) const;
	
	//	Activation list manipulation.
	
	virtual Bool SetFirstActive ();
	virtual Bool SetLastActive ();
	virtual Bool SetActiveBefore (OPViewIn sibling);
	virtual Bool SetActiveAfter (OPViewIn sibling);
	
	virtual Bool RemoveActive ();
	
	virtual OPViewRef ReturnFirstActive () const;
	virtual OPViewRef ReturnLastActive () const;
	virtual OPViewRef ReturnNextActive () const;
	virtual OPViewRef ReturnPrevActive () const;
	
	virtual Bool FindNextActive (OPViewOut view) const;
	virtual Bool FindPrevActive (OPViewOut view) const;
	
	virtual Bool SetActiveRank (Int32 pos);
	virtual Int32 ReturnActiveRank () const;
	
	virtual OPListRet ReturnActiveList () const;
	
	//	User tag and name manipulation.
	
	virtual Bool SetUserTag (Int32 tag);
	virtual Bool SetUserName (OPTextIn name);
	virtual Int32 ReturnUserTag () const;
	virtual OPAtomRet ReturnUserName () const;
	
	//	Access to the view's state. Only some flags are publicly modifiable.
	//	The others will have to be modified with some specialized method.
	
	virtual void SetFlagOkDetect (Bool ok);
	virtual void SetFlagOkUIEdit (Bool ok);
	virtual void SetFlagOkUIMove (Bool ok);
	virtual void SetFlagOkMain (Bool ok);
	virtual void SetFlagOkDrag (Bool ok);
	virtual void SetFlagOkDrop (Bool ok);
	
	virtual void SetEnable (Bool yes);
	virtual void SetVisible (Bool yes);
	
	virtual Bool ReturnEnable () const;
	virtual Bool ReturnVisible () const;
	
	virtual void SetGlobalEnable (Bool mode);
	virtual Bool ReturnGlobalEnable () const;
	
	virtual void ChangeState (OPViewState state, Bool set);
	virtual Bool SetState (OPViewState state);
	virtual OPViewState ReturnState () const;
	
	//	Miscellanous methods :
	
	virtual Bool FindUserTag (Int32 tag, OPViewOut view);
	virtual Bool FindUserName (OPAtomIn name, OPViewOut view);
	virtual Bool Detect (OPCoord x, OPCoord y, OPViewOut view);
	virtual Bool Detect (OPRectIn box);
	
	virtual Bool IsMainResponder () const;
	virtual void ParentResizing (OPCoord dx, OPCoord dy);
	virtual void ChildChangedSize (OPViewIn child);
	virtual void Broadcast (OPAtomIn message, OPObjectIn value);
	
	//	Draw the entire view or just part of it. This will always draw into
	//	a buffer before being dumped to the screen (this results in slower
	//	initial drawing but gets rid of some ugly glitches and speeds up the
	//	redrawing operations significantly).
	
	virtual void Draw (OPGrailPortIn port, OPDrawOp what = OP_DRAW_ALL);
	virtual void Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what = OP_DRAW_ALL);
	virtual void DrawOverlay (OPGrailPortIn port, OPRectIn rect);
	virtual void Display ();
	virtual void Display (OPRectIn rect);
	
	virtual void EnterNotify (OPEventIn event);				/* OVERRIDDEN */
	virtual void LeaveNotify (OPEventIn event);				/* OVERRIDDEN */
	
	//	Special accessor methods for frequently used semi-global variables :
	
	virtual OPGrailPortRet ReturnGrailPort () const;
	virtual OPLookManRet ReturnLookMan () const;
	virtual OPAdornerRet ReturnAdorner () const;
	
	//	Drag & drop protocol handling :
	
	virtual Bool DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp);
	
	OP_DEFINE_SLOT (Card32, state,    SetState,   ReturnState);
	OP_DEFINE_SLOT (Card32, user_tag, SetUserTag, ReturnUserTag);
	OP_DEFINE_SLOT (Int32, active_rank, SetActiveRank, ReturnActiveRank);
	
	OP_DEFINE_CLASS (OPView, OPResponder);
};

/****************************************************************/

#endif /* _OPaC_VIEW_H_ */
