/*
 *	sd_file_io.cxx
 *	
 *	File input/output functions. Use POSIX standard.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 09.02.98

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <OPaC/string.h>
#include <OPaC/sd/file_io.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;

/****************************************************************/

// Open file in read mode.

OPaC_FileDesc
OPaC_OpenFileRead (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileRead, 1);
	OP_ARG (1, OPStringIn, name);
	
	int    filedesc;
	int    mode = O_RDONLY;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	filedesc = open (buffer, mode);
	
	if (filedesc == -1)
		return 0;
	
	OP_ASSERT (filedesc);
	
	return (OPaC_FileDesc)(filedesc);
}

// Open file in write mode

OPaC_FileDesc
OPaC_OpenFileWrite (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileWrite, 1);
	OP_ARG (1, OPStringIn, name);
	
	int    filedesc;
	int    mode = O_RDWR | O_CREAT;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	filedesc = open (buffer, mode, S_IRUSR | S_IWUSR | S_IRGRP);
	
	if (filedesc == -1)
		return 0;
	
	OP_ASSERT (filedesc);
	
	return (OPaC_FileDesc)(filedesc);
}

// Open file in append mode

OPaC_FileDesc
OPaC_OpenFileAppend (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileAppend, 1);
	OP_ARG (1, OPStringIn, name);

	int filedesc;
	int mode = O_WRONLY | O_APPEND;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	filedesc = open (buffer, mode);
	
	if (filedesc == -1)
		return 0;
	
	OP_ASSERT (filedesc);
	
	return (OPaC_FileDesc)(filedesc);
 
}

// Read bytes from a file

Size
OPaC_ReadFile (OPaC_FileDesc file, void* void_buffer, Size size)
{
	OP_FUNCTION (OPaC_ReadFile, 3);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, void*, void_buffer);
	OP_ARG (3, Size, size);
	
	int filedesc = (int)file;
	OP_ASSERT (filedesc);
	
	Size         total_read = 0;
	unsigned int chunk_size = 0;
	LatinChar* buffer     = (LatinChar*)(void_buffer);
	
	while (size > 10000) {
		
		chunk_size = read (filedesc, buffer, 10000);
		total_read = total_read + chunk_size;
		buffer     = buffer + chunk_size;
		size       = size - 10000;
		
		if (chunk_size != 10000) return total_read;
	}
	
	chunk_size = read (filedesc, buffer, (unsigned int)(size));
	total_read = total_read + chunk_size;
	
	return total_read;
}

// Wrtie bytes to a file

Size
OPaC_WriteFile (OPaC_FileDesc file, const void* void_buffer, Size size)
{
	OP_FUNCTION (OPaC_WriteFile, 3);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, const void*, void_buffer);
	OP_ARG (3, Size, size);
	
	int filedesc = (int)(file);
	OP_ASSERT (filedesc);
	
	Size             total_write = 0;
	unsigned int     chunk_size  = 0;
	const LatinChar* buffer      = (const LatinChar*)(void_buffer);
	
	while (size > 10000) {
		
		chunk_size  = write (filedesc, buffer, 10000);
		total_write = total_write + chunk_size;
		buffer      = buffer + chunk_size;
		size        = size - 10000;
		
		if (chunk_size != 10000) return total_write;
	}
	
	chunk_size  = write (filedesc, buffer, (unsigned int)(size));
	total_write = total_write + chunk_size;
	
	return total_write;
}

// Move to offset in file

Bool
OPaC_SeekFileAbs (OPaC_FileDesc file, Int32 offset)
{
	OP_FUNCTION (OPaC_SeekFileAbs, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Int32, offset);
	
	int filedesc = (int)(file);
	int mode  = (offset < 0) ? SEEK_END : SEEK_SET;
	off_t pos   = (offset < 0) ? (off_t)(-offset - 1) : (off_t)offset;
	
	OP_ASSERT (filedesc);
	
	return ( lseek (filedesc, pos, mode) == -1) ? FALSE : TRUE;
}

// Move offset bytes from current position

Bool
OPaC_SeekFileRel (OPaC_FileDesc file, Int32 offset)
{
	OP_FUNCTION (OPaC_SeekFileRel, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Int32, offset);
	
	int filedesc = (int)(file);
	int mode  = SEEK_CUR;
	off_t pos   = offset;
	
	OP_ASSERT (filedesc);
	
	return ( lseek (filedesc, pos, mode) == -1) ? FALSE : TRUE;
}

// Return the current position

Size
OPaC_TellFilePos (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TellFilePos, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	int   filedesc = (int)(file);
	int   mode  = SEEK_CUR;
	off_t pos   = 0;

	off_t offset = lseek (filedesc, pos, mode);
	
	OP_ASSERT (offset >= 0);
	
	return (Size)(offset);
}

// Return the file size

Size
OPaC_TellFileSize (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TellFileSize, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	int   filedesc = (int)(file);
	int   mode  = SEEK_END;
	off_t pos   = 0;

	off_t offset = llseek (filedesc, pos, mode);
	
	OP_ASSERT (offset >= 0);
	
	return (Size)(offset);
}

// Synchronise the file in memory and the disk

Bool
OPaC_FlushFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_FlushFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);

	return fsync( (int)file ) == -1 ? FALSE : TRUE;
}

// Truncate the file at the current position

Bool
OPaC_TruncFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TruncFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	int   filedesc = (int)(file);
	int   mode  = SEEK_SET;
	off_t pos   = 0;
	
	off_t size = lseek (filedesc, pos, mode);
	
	if (size < 0) {
		return FALSE;
	}
	
	return (ftruncate ((int)file, (size_t)size)) >= 0 ? TRUE : FALSE;
}

// Resize the file to size

Bool
OPaC_ResizeFile (OPaC_FileDesc file, Size size)
{
	OP_FUNCTION (OPaC_ResizeFile, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Size, size);
	OP_ASSERT (file);
	
	int filedesc = (int)(file);
	
	return (ftruncate (filedesc, (size_t)size)) >= 0 ? TRUE : FALSE;
}

// Close the file

Bool
OPaC_CloseFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_CloseFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	int filedesc = (int)(file);
	return (close (filedesc) == -1) ? FALSE : TRUE;
}
