/*
 *	sd_grailcur.cxx
 *
 *	CAUTION: For 32-bit Windows only !
 *
 *	GRAphical Interface Layer for Cursor shape. Implementation for X11.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 08.01.98

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"
#include "../common/sd_grafix.h"

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPGrailCursor);

/****************************************************************/


/*
 *	Create an empty graphical cursor.
 */

OPGrailCursor::OPGrailCursor ()
{
	OP_METHOD (OPGrailCursor, 0);
	
	this->cursor = OP_ALLOC_FROM_ZONE (this->ReturnMemZone (), OPaC_GrCursor);
	this->cursor->xcursor = 0;
	
	this->foreground = 0;
	this->background = 0;
	
	this->x_foreground = 0;
	this->y_foreground = 0;
	
	this->x_background = 0;
	this->y_background = 0;
	
	this->dx = 0;
	this->dy = 0;
	
	this->hot_x = 0;
	this->hot_y = 0;
}



/*
 *	Free & destroy a graphical cursor.
 */

OPGrailCursor::~OPGrailCursor ()
{
	OP_METHOD (~OPGrailCursor, 0);
		
	OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->cursor);
}



/*
 *	Save the cursor. This cannot be done, since this would
 *	lead to machine dependent archives.
 */

void
OPGrailCursor::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to save Grail Cursor");
}



/*
 *	Restore the cursor. This cannot be done.
 */

void
OPGrailCursor::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore Grail Cursor");
}



/*
 *	Initialize a cursor from the specified pixel maps. The
 *	background will be opaque wherever it is set to black.
 *	The foreground will be painted wherever a black back-
 *	ground has been drawn.
 *
 *	The cursor definition is the same as on the Smaky
 *	operating system and has been chosen because it is
 *	quite easy to draw the masks manually.
 */

void
OPGrailCursor::Initialise (OPCoord hotx, OPCoord hoty,
						   OPCoord dx, OPCoord dy,
						   OPGrailPixmapIn fg, OPCoord xf, OPCoord yf,
						   OPGrailPixmapIn bg, OPCoord xb, OPCoord yb)
{
	OP_METHOD (Initialise, 10);
	OP_ARG (1, OPCoord, hotx);
	OP_ARG (2, OPCoord, hoty);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	OP_ARG (5, OPGrailPixmapIn, fg);
	OP_ARG (6, OPCoord, xf);
	OP_ARG (7, OPCoord, yf);
	OP_ARG (8, OPGrailPixmapIn, bg);
	OP_ARG (9, OPCoord, xb);
	OP_ARG (10, OPCoord, yb);
	
	//	When the cursor size is invalid, don't use any cursor at all. This
	//	will force the use of a default (system) arrow cursor.
	
	if ( (dx * dy) == 0) {
		
		this->cursor->xcursor = 0;
		
		this->foreground = 0;
		this->background = 0;
		
		this->x_foreground = 0;
		this->y_foreground = 0;
		
		this->x_background = 0;
		this->y_background = 0;
		
		this->dx = 0;
		this->dy = 0;
		
		this->hot_x = 0;
		this->hot_y = 0;
		
		return;
	}
	
	
	//	The grail pixmap reference and both memory chunks have to be recorded
	//	as an automatic variable so that these will be properly destroyed on
	//	exception.
	
	OPCoord ddx, ddy;
	
	Pixmap cbg = XCreatePixmap (xdpy, DefaultRootWindow(xdpy),
							    (int)dx, (int)dy, 1);
	
	Pixmap cfg = XCreatePixmap (xdpy, DefaultRootWindow(xdpy),
								(int)dx, (int)dy, 1);
	
	GC gc = XCreateGC (xdpy, cbg, 0, 0);
	
	//  Build the two monochrome images used to represent the foreground
	//  and the background (source and mask).
	
	//  NIY
	
	XColor xblack;
	XColor xwhite;
	
	xblack.pixel = XBlackPixel (xdpy, DefaultScreen (xdpy));
	xwhite.pixel = XWhitePixel (xdpy, DefaultScreen (xdpy));
	
	XQueryColor (xdpy, XDefaultColormap (xdpy, DefaultScreen(xdpy)), &xblack);
	XQueryColor (xdpy, XDefaultColormap (xdpy, DefaultScreen(xdpy)), &xwhite);
							 
 	this->cursor->xcursor = XCreatePixmapCursor (xdpy,
												 cfg,
												 cbg,
												 &xblack,         // Foreground
												 &xwhite,         // Background
												 (unsigned int) hot_x,      // Hotspot
												 (unsigned int) hot_y );   // 
															

	XFreeGC(xdpy, gc);
	XFreePixmap(xdpy, cbg);
	XFreePixmap(xdpy, cfg);

	//
	//	NB :
	//
	//	The monochrome Grail Pixmap will be deleted automatically when its use
	//	counter will reach zero.
	//
	
	this->foreground   = fg;
	this->background   = bg;
	this->x_foreground = (Int16)(xf);
	this->y_foreground = (Int16)(yf);
	this->x_background = (Int16)(xb);
	this->y_background = (Int16)(yb);
	this->dx = (Int16)(dx);
	this->dy = (Int16)(dy);
	this->hot_x = (Int16)(hotx);
	this->hot_y = (Int16)(hoty);
}



/*
 *	Make the cursor visible on screen.
 */

void
OPGrailCursor::Activate () const
{
	OP_METHOD (Activate, 0);
	
	Window xwindow;
	int revertWindow;

	XGetInputFocus(xdpy, &xwindow, &revertWindow);

	XDefineCursor(xdpy, xwindow, cursor->xcursor);
}



/*
 *	Switch back to the default cursor.
 */

void
OPGrailCursor::ResetToNormal () const
{
	OP_METHOD (ResetToNormal, 0);

	Window* xwindow;

	XGetInputFocus(xdpy, xwindow, NULL);

	XUndefineCursor(xdpy, *xwindow);	
}



/*
 *	Remove any cursor from the screen.
 */

void
OPGrailCursor::Hide () const
{
	OP_METHOD (Hide, 0);

	Window* xwindow;

	XGetInputFocus(xdpy, xwindow, NULL);

	XUndefineCursor(xdpy, *xwindow);	
}
