/*
 *	filedir.cxx
 *
 *	The File Dir class provides access to directories.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#include <OPaC/container.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/time.h>

#include <OPaC/sd/filedir.h>

OP_IMPLEMENTATION_FILE;

/****************************************************************/

class OPaC_PlatformFileDir : public OPaC_FileDir
{
private:
	HANDLE				handle;
	WIN32_FIND_DATA		find;
	Bool				is_first;
	
protected:
	virtual ~ OPaC_PlatformFileDir ();
	
public:
	OPaC_PlatformFileDir ();
	
	virtual Bool Open (OPStringIn path, OPStringIn pattern);
	virtual Bool Close ();
	virtual Bool FindNext (OPaC_FileInfoOut info);
	virtual Bool FindNext (OPaC_FileInfoOut info, OPDictOut attr);
};

/****************************************************************/

OPaC_FileDir*
OPaC_FileDir::New ()
{
	return NEW (OPaC_PlatformFileDir);
}

/****************************************************************/

/*
 *	Convert a name with forward slashes into a legal Win32
 *	name.
 */

static void
ConvertFileName (LatinChar* name)
{
	while (*name) {
		if (*name == '/') *name = '\\';
		name++;
	}
}

/*
 *	Convert the NT file time (which is coded as Universal
 *	Time - UTC) into OPaC's absolute (local) time.
 */

static Int64
ConvertFileTimeToAbsTime (const FILETIME& time)
{
	FILETIME   filetime;
	SYSTEMTIME systime = { 0 };
	
	FileTimeToLocalFileTime (&time, &filetime);
	FileTimeToSystemTime (&filetime, &systime);
	
	return OPTime::DateTimeToAbsTime (systime.wDay, systime.wMonth, systime.wYear,
									  systime.wHour, systime.wMinute, systime.wSecond,
									  systime.wMilliseconds);
}

/*
 *	Convert the Windows find data about a file into an
 *	OPaC file information structure.
 */

static void
ConvertFindDataToInfo (WIN32_FIND_DATA& find, OPaC_FileInfoOut info)
{
	info.name = find.cFileName;
	info.size = ((Card64)(find.nFileSizeHigh) << 32) | find.nFileSizeLow;
	
	//	File times are stored in UTC time on the file system. This will be translated
	//	to local time for further processing by the caller...
	
	info.creation_time    = ConvertFileTimeToAbsTime (find.ftCreationTime);
	info.last_write_time  = ConvertFileTimeToAbsTime (find.ftLastWriteTime);
	info.last_access_time = ConvertFileTimeToAbsTime (find.ftLastAccessTime);
	
	info.is_directory = (find.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? TRUE : FALSE;
	info.is_link      = FALSE;
	info.is_hidden    = (find.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) ? TRUE : FALSE;
	info.is_read_ok   = TRUE;
	info.is_write_ok  = (find.dwFileAttributes & FILE_ATTRIBUTE_READONLY) ? FALSE : TRUE;
	
}

/****************************************************************/

OPaC_PlatformFileDir::OPaC_PlatformFileDir ()
{
	this->handle   = INVALID_HANDLE_VALUE;
	this->is_first = FALSE;
}

OPaC_PlatformFileDir::~OPaC_PlatformFileDir ()
{
	if (this->is_open) {
		this->Close ();
	}
}

Bool
OPaC_PlatformFileDir::Open (OPStringIn path, OPStringIn pattern)
{
	LatinChar buf_path[MAX_PATH+MAX_PATH];
	LatinChar buf_pattern[MAX_PATH];
	
	strcpy (buf_path, ".");
	strcpy (buf_pattern, "*");
	
	if ( (path.IsValid ())
	  && (path->CountCharacters ()) ) {
		path->GetStringCopy (buf_path, MAX_PATH);
	}
	if ( (pattern.IsValid ())
	  && (pattern->CountCharacters ()) )  {
		pattern->GetStringCopy (buf_pattern, MAX_PATH);
	}
	
	ConvertFileName (buf_path);
	
	LatinChar* name = buf_path + strlen (buf_path);
	
	//	Add the terminating back slash if it is absent from the user's path.
	
	if (name[-1] != '\\') {
		name[0] = '\\';
		name[1] = 0;
	}
	
	name = buf_path;
	strcat (name, buf_pattern);
	
	this->Close ();
	this->handle = FindFirstFile (name, & this->find);
	
	if (this->handle == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	
	//	We have acquired the first "sample" file. Return it when the user first
	//	calls the FindNext method.
	
	this->is_first = TRUE;
	this->is_open  = TRUE;
	
	return TRUE;
}

Bool
OPaC_PlatformFileDir::Close ()
{
	this->is_open = FALSE;
	
	if (this->handle != INVALID_HANDLE_VALUE) {
		FindClose (this->handle);
		this->handle = INVALID_HANDLE_VALUE;
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPaC_PlatformFileDir::FindNext (OPaC_FileInfoOut info)
{
	if (! this->is_first) {
		if (FindNextFile (this->handle, & this->find) == FALSE) {
			return FALSE;
		}
	} else {
		this->is_first = FALSE;
	}
	
	ConvertFindDataToInfo (this->find, info);
	
	return TRUE;
}

Bool
OPaC_PlatformFileDir::FindNext (OPaC_FileInfoOut info, OPDictOut attr)
{
	if (this->FindNext (info)) {
		attr = OPDictNull;
		return TRUE;
	}
	
	return FALSE;
}

/****************************************************************/

/*
 *	Return the properties of a specified file (with complete
 *	path).
 */

Bool
OPaC_FileDir::Properties (OPStringIn name, OPaC_FileInfoOut info)
{
	OP_MEMBER (Properties, 2);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, OPaC_FileInfoOut, info);
	
	if ( (name.IsInvalid ())
	  || (name->CountCharacters () == 0) ) {
		return FALSE;
	}
	
	WIN32_FIND_DATA find;
	HANDLE          handle;
	
	LatinChar buffer[MAX_PATH];
	
	name->GetStringCopy (buffer, MAX_PATH);
	ConvertFileName (buffer);
	
	handle = FindFirstFile (buffer, & find);
	
	if (handle == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	
	ConvertFindDataToInfo (find, info);
	FindClose (handle);
	
	return TRUE;
}

/****************************************************************/
