/*
 *	colorpixel.cxx
 *
 *	Defines the "color pixel" representation, which is an automatically
 *	converting pixel representation format.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/colorpixel.h>
#include <OPaC/lookcolor.h>
#include <OPaC/lookman.h>
#include <OPaC/opexport.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_TYPECLASS_VALUE (OPColorPixel, OP_TYPE_COLOR);

/****************************************************************/

const OPColorPixel OPColorPixel::none (OP_COLOR_NONE);

/****************************************************************/

OPColorPixel::OPColorPixel ()
{
	this->pixel = 0;
	this->color = OP_COLOR_NONE;
}

OPColorPixel::~OPColorPixel ()
{
}

/*
 *	Store a color definition.
 */

void
OPColorPixel::StoreInstance (OPStorageIn storage) const
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	OPAtomRef name;
	OPLookMan::Current ()->GetColorName (this->color, name);
	
	static OPStorageSync info ("<OPColorPixel pixel name>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->pixel << name;
	storage << OPStorageSync::end;
}

/*
 *	Restoring an application is not possible.
 */

void
OPColorPixel::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	OPAtomRef name;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->pixel >> name;
	storage >> OPStorageSync::end;
	
	this->color = OPLookMan::Current ()->FindColor (name);
}

void
OPColorPixel::UpdatePixel () const
{
	if (this->color == OP_COLOR_ANY) return;
	if (this->color == OP_COLOR_NONE) return;
	
	OPLookMan::Current ()->GetColorCompactRGB (this->color, this->pixel);
}
	
/****************************************************************/
	
OPColorPixel::OPColorPixel (Card32 pixel)
{
	this->pixel = pixel;
	this->color = OP_COLOR_ANY;
}

OPColorPixel::OPColorPixel (OPLookColor color)
{
	this->pixel = 0;
	this->color = color;
}

OPColorPixel::OPColorPixel (OPColorPixelIn pixel)
{
	this->pixel = pixel.pixel;
	this->color = pixel.color;
}

OPColorPixel::OPColorPixel (OPAtomIn name)
{
	this->pixel = 0;
	this->color = OPLookMan::Current ()->FindColor (name);
}

OPColorPixel::OPColorPixel (OPValueIn value)
{
	const OPColorPixel& color = value->ReturnColor ();
	this->pixel = color.pixel;
	this->color = color.color;
}

static inline Card8 real_to_card (Real x)
{
	if (x <= 0) return 0;
	if (x >= 1) return 255;
	return (Card8)(x*256);
}

OPColorPixel::OPColorPixel (Real rr, Real rg, Real rb, Real rt)
{
	Card8 t = real_to_card (rt);
	Card8 r = real_to_card (rr);
	Card8 g = real_to_card (rg);
	Card8 b = real_to_card (rb);
	
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

OPColorPixel::OPColorPixel (Card8 r, Card8 g, Card8 b, Card8 t)
{
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

OPColorPixel&
OPColorPixel::operator = (Card32 pixel)
{
	this->pixel = pixel;
	this->color = OP_COLOR_ANY;
	return (*this);
}

OPColorPixel&
OPColorPixel::operator = (OPLookColor color)
{
	this->pixel = 0;
	this->color = color;
	return (*this);
}

OPColorPixel&
OPColorPixel::operator = (OPColorPixelIn pixel)
{
	this->pixel = pixel.pixel;
	this->color = pixel.color;
	return (*this);
}

OPColorPixel&
OPColorPixel::operator = (OPAtomIn name)
{
	this->pixel = 0;
	this->color = OPLookMan::Current ()->FindColor (name);
	return (*this);
}

OPColorPixel&
OPColorPixel::operator = (OPValueIn value)
{
	const OPColorPixel& color = value->ReturnColor ();
	this->pixel = color.pixel;
	this->color = color.color;
	return (*this);
}

OPColorPixel::operator Card32 () const
{
	this->UpdatePixel ();
	
	return this->pixel;
}

OPColorPixel::operator OPLookColor () const
{
	return this->color;
}

OPColorPixel::operator OPAtomRef () const
{
	OPAtomRef name;
	OPLookMan::Current ()->GetColorName (this->color, name);
	return name;
}

/****************************************************************/

void
OPColorPixel::SetRed (Card8 r)
{
	Card8 g = this->Green ();
	Card8 b = this->Blue ();
	Card8 t = this->Transparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetGreen (Card8 g)
{
	Card8 r = this->Red ();
	Card8 b = this->Blue ();
	Card8 t = this->Transparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetBlue (Card8 b)
{
	Card8 r = this->Red ();
	Card8 g = this->Green ();
	Card8 t = this->Transparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetTransparency (Card8 t)
{
	Card8 r = this->Red ();
	Card8 g = this->Green ();
	Card8 b = this->Blue ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetLuminance (Card8 l)
{
	Card8 lum = this->Luminance ();
	if (lum == l) return;
	
	this->SetLuminance (l / 255.0);
}

void
OPColorPixel::SetRed (Real r)
{
	Real g = this->RealGreen ();
	Real b = this->RealBlue ();
	Real t = this->RealTransparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetGreen (Real g)
{
	Real r = this->RealRed ();
	Real b = this->RealBlue ();
	Real t = this->RealTransparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetBlue (Real b)
{
	Real r = this->RealRed ();
	Real g = this->RealGreen ();
	Real t = this->RealTransparency ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetTransparency (Real t)
{
	Real r = this->RealRed ();
	Real g = this->RealGreen ();
	Real b = this->RealBlue ();
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
	this->color = OP_COLOR_ANY;
}

void
OPColorPixel::SetLuminance (Real l)
{
	Real lum = this->RealLuminance ();
	Real t   = this->RealTransparency ();
	
	if (lum == l) {
		return;
	}
	
	this->color = OP_COLOR_ANY;
	
	if (l < 0.0) {
		this->pixel = OPColorPixel::MakeCompactRGB (0.0, 0.0, 0.0, t);
		return;
	}
	if (l > 1.0) {
		this->pixel = OPColorPixel::MakeCompactRGB (1.0, 1.0, 1.0, t);
		return;
	}
	if (lum == 0.0) {
		this->pixel = OPColorPixel::MakeCompactRGB (l, l, l, t);
		return;
	}
	
	lum = l / lum;
	
	Real r = this->RealRed ()   * lum;
	Real g = this->RealGreen () * lum;
	Real b = this->RealBlue ()  * lum;
	
	for (int i = 0; i < 4; i++) {
		
		//	Iterative distribution of the excess luminosity (if any).
		
		Real delta = 0.0;
		Real carry = 1024.0;
		
		if (r > 1.0) {
			delta += (r - 1.0) * 307.0;
			carry -= 307.0;
			r = 1.0;
		}
		if (g > 1.0) {
			delta += (g - 1.0) * 614.0;
			carry -= 614.0;
			g = 1.0;
		}
		if (b > 1.0) {
			delta += (b - 1.0) * 103.0;
			carry -= 103.0;
			b = 1.0;
		}
		
		if (delta == 0.0) break;
		
		delta /= carry;
		
		if (r < 1.0) r += delta;
		if (g < 1.0) g += delta;
		if (b < 1.0) b += delta;
	}	
	
	this->pixel = OPColorPixel::MakeCompactRGB (r, g, b, t);
}

	
/****************************************************************/

Card8
OPColorPixel::Red () const
{
	this->UpdatePixel ();
	return (Card8)(this->pixel >> 16);
}

Card8
OPColorPixel::Green () const
{
	this->UpdatePixel ();
	return (Card8)(this->pixel >> 8);
}

Card8
OPColorPixel::Blue () const
{
	this->UpdatePixel ();
	return (Card8)(this->pixel >> 0);
}

Card8
OPColorPixel::Transparency () const
{
	this->UpdatePixel ();
	return (Card8)(this->pixel >> 24);
}

Card8
OPColorPixel::Luminance () const
{
	this->UpdatePixel ();
	
	Card32 r = (this->pixel >> 16) & 0xFF;
	Card32 g = (this->pixel >>  8) & 0xFF;
	Card32 b = (this->pixel >>  0) & 0xFF;
	
	return (Card8)((r*307+g*614+b*103) >> 10);
}

Real
OPColorPixel::RealRed () const
{
	return this->Red () / 255.0;
}

Real
OPColorPixel::RealGreen () const
{
	return this->Green () / 255.0;
}

Real
OPColorPixel::RealBlue () const
{
	return this->Blue () / 255.0;
}

Real
OPColorPixel::RealTransparency () const
{
	return this->Transparency () / 255.0;
}

Real
OPColorPixel::RealLuminance () const
{
	return this->Luminance () / 255.0;
}

/****************************************************************/

Bool
OPColorPixel::operator == (Card32 x) const
{
	this->UpdatePixel ();
	return (x == this->pixel) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator == (OPLookColor x) const
{
	return (x == this->color) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator == (OPColorPixelIn x) const
{
	if (this->color == OP_COLOR_ANY) {
		if (x.color == OP_COLOR_ANY) {
			return (this->pixel == x.pixel) ? TRUE : FALSE;
		} else {
			return FALSE;
		}
	}
	return (this->color == x.color) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator == (OPAtomIn x) const
{
	OPAtomRef name;
	OPLookMan::Current ()->GetColorName (this->color, name);
	return x == name;
}
	
Bool
operator == (Card32 a, OPColorPixelIn b)
{
	return b == a;
}

Bool
operator == (OPLookColor a, OPColorPixelIn b)
{
	return b == a;
}

Bool
operator == (OPAtomIn a, OPColorPixelIn b)
{
	return b == a;
}

/****************************************************************/

Bool
OPColorPixel::operator != (Card32 x) const
{
	this->UpdatePixel ();
	return (x != this->pixel) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator != (OPLookColor x) const
{
	return (x != this->color) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator != (OPColorPixelIn x) const
{
	if (this->color == OP_COLOR_ANY) {
		if (x.color == OP_COLOR_ANY) {
			return (this->pixel == x.pixel) ? FALSE : TRUE;
		} else {
			return TRUE;
		}
	}
	return (this->color != x.color) ? TRUE : FALSE;
}

Bool
OPColorPixel::operator != (OPAtomIn x) const
{
	OPAtomRef name;
	OPLookMan::Current ()->GetColorName (this->color, name);
	return x != name;
}
	
Bool
operator != (Card32 a, OPColorPixelIn b)
{
	return b != a;
}

Bool
operator != (OPLookColor a, OPColorPixelIn b)
{
	return b != a;
}

Bool
operator != (OPAtomIn a, OPColorPixelIn b)
{
	return b != a;
}

/****************************************************************/

Card32
OPColorPixel::MakeCompactRGB (Card8 r, Card8 g, Card8 b, Card8 t)
{
	return (Card32)((t << 24) | (r << 16) | (g << 8) | (b));
}

Card32
OPColorPixel::MakeCompactRGB (Real rr, Real rg, Real rb, Real rt)
{
	if (rr < 0.0) rr = 0.0; else if (rr > 1.0) rr = 1.0;
	if (rg < 0.0) rg = 0.0; else if (rg > 1.0) rg = 1.0;
	if (rb < 0.0) rb = 0.0; else if (rb > 1.0) rb = 1.0;
	if (rt < 0.0) rt = 0.0; else if (rt > 1.0) rt = 1.0;
	Card8 r = (Card8)(rr * 255.0);
	Card8 g = (Card8)(rg * 255.0);
	Card8 b = (Card8)(rb * 255.0);
	Card8 t = (Card8)(rt * 255.0);
	return (Card32)((t << 24) | (r << 16) | (g << 8) | (b));
}

