/*
 *	dockview.cxx
 *
 *	The docking view can be torn off its parent view and moved
 *	around in its own window. It provides the basic support for
 *	palettes and main menus.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/colorpixel.h>
#include <OPaC/dockview.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/window.h>

#include <OPaC/targetmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPDockView);

/****************************************************************/

OP_IMPLEMENT_TYPEDEF_VAL (Card32, OPDockStyle, OP_TYPE_CARD, Card32);

/****************************************************************/

static const OPaC_BitsetDataDef local_style_info_1[] = {
	{ OP_DOCK_HANDLE_MASK,	"Docking handle position",	OP_BITSET_DATA_MASK },
	{ OP_DOCK_HANDLE_NONE,	"no handle",				OP_BITSET_DATA_VALUE },
	{ OP_DOCK_HANDLE_LEFT,	"left handle, vertical",	OP_BITSET_DATA_VALUE },
	{ OP_DOCK_HANDLE_TOP,	"top handle, horizontal",	OP_BITSET_DATA_VALUE }
};

static const OPaC_BitsetDataDef local_style_info_2[] = {
	{ OP_DOCK_STRIPE_MASK,	"Docking handle shape",		OP_BITSET_DATA_MASK },
	{ OP_DOCK_STRIPE_SINGLE,"single stripe",			OP_BITSET_DATA_VALUE },
	{ OP_DOCK_STRIPE_DOUBLE,"double stripe",			OP_BITSET_DATA_VALUE },
	{ OP_DOCK_STRIPE_THIN,	"hair thin stripe",			OP_BITSET_DATA_VALUE },
	{ OP_DOCK_STRIPE_BUMP,	"bump-like",				OP_BITSET_DATA_VALUE }
};

static const OPaC_BitsetDataDef local_style_info_3[] = {
	{ OP_DOCK_SEL_MASK,		"To select, click on",		OP_BITSET_DATA_MASK },
	{ OP_DOCK_SEL_HANDLE,	"handle",					OP_BITSET_DATA_BIT	},
	{ OP_DOCK_SEL_BORDER,	"border",					OP_BITSET_DATA_BIT	},
	{ OP_DOCK_SEL_BODY,		"body",						OP_BITSET_DATA_BIT	}
};

static const OPaC_BitsetDataDef local_style_info_4[] = {
	{ OP_DOCK_DRAW_MASK,	"Drawing gadgets",			OP_BITSET_DATA_MASK },
	{ OP_DOCK_DRAW_GLOW,	"glow on hover",			OP_BITSET_DATA_BIT	},
	{ OP_DOCK_DRAW_SHADE,	"shade while docking",		OP_BITSET_DATA_BIT	},
	{ OP_DOCK_DRAW_FRAME,	"frame and background",		OP_BITSET_DATA_BIT	}
};


/****************************************************************/

OP_TARGET_ACTOR (OPDockView, StartDragging);

OP_PUBLISH_SLOT (OPDockView, style, ReturnStyle, SetStyle, "Docking style", OPDockStyle);

OP_PUBLISH_BITSET_INFO (OPDockView, style, local_style_info_1);
OP_PUBLISH_BITSET_INFO (OPDockView, style, local_style_info_2);
OP_PUBLISH_BITSET_INFO (OPDockView, style, local_style_info_3);
OP_PUBLISH_BITSET_INFO (OPDockView, style, local_style_info_4);

/****************************************************************/

OPDockView::OPDockView ()
{
	OP_METHOD (OPDockView, 0);
	
	this->is_dragging = FALSE;
	this->hot_x  = 0;
	this->hot_y  = 0;
	this->style  = OP_DOCK_HANDLE_LEFT | OP_DOCK_STRIPE_SINGLE | OP_DOCK_SEL_HANDLE | OP_DOCK_DRAW_FRAME;
	this->border = 2;
	this->UpdateHandleWidth ();
}

OPDockView::~OPDockView ()
{
	OP_METHOD (~OPDockView, 0);
}

void
OPDockView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPDockView dock_window style border width>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->dock_window << this->style << this->border << this->width;
	storage << OPStorageSync::end;
}

void
OPDockView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->dock_window >> this->style >> this->border >> this->width;
	storage >> OPStorageSync::end;
	
	this->UpdateHandleWidth ();
}


/*
 *	Delete the docking view.
 */

void
OPDockView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->dock_window.IsValid ()) {
		this->SetParent (OPViewNull);
		this->dock_window->Delete ();
		this->dock_window = OPWindowNull;
	}
	
	super::Delete ();
}


/*
 *	Find out about the width of the handle, depending on its
 *	striping.
 */

void
OPDockView::UpdateHandleWidth ()
{
	OP_METHOD (UpdateHandleWidth, 0);
	
	switch (this->style & OP_DOCK_STRIPE_MASK) {
		
		case OP_DOCK_STRIPE_SINGLE:
			this->width = 9;
			break;
		
		case OP_DOCK_STRIPE_DOUBLE:
			this->width = 11;
			break;
		
		case OP_DOCK_STRIPE_THIN:
			this->width = 7;
			break;
		
		case OP_DOCK_STRIPE_BUMP:
			this->width = 12;
			break;
		
		default:
			break;
	}
}

/*
 *	Define the docking view's style.
 */

Bool
OPDockView::SetStyle (OPDockStyleIn style)
{
	OP_METHOD (SetStyle, 1);
	OP_ARG (1, OPDockStyleIn, style);
	
	this->style = style;
	this->UpdateHandleWidth ();
	
	return TRUE;
}


OPDockStyle
OPDockView::ReturnStyle () const
{
	OP_METHOD (ReturnStyle, 0);
	
	return this->style;
}


Bool
OPDockView::StartDragging ()
{
	OP_METHOD (StartDragging, 0);
	
	if (this->is_dragging) {
		return FALSE;
	}
	
	const OPEventRef& event = OPEvent::CurrentEvent ();
	
	if (this->is_dragging) {
		return FALSE;
	}
	
	this->hot_x = event->ReturnX ();
	this->hot_y = event->ReturnY ();
	
	if (this->dock_window.IsInvalid ()) {
		
		OPCoord   wx = 0;
		OPCoord   wy = 0;
		OPViewRef parent = this->ReturnParent ();
		
		this->window->GetOrigin (wx, wy);
		
		wx += this->box.X1 ();
		wy += this->box.Y1 ();
		
		this->hot_x -= this->box.X1 ();
		this->hot_y -= this->box.Y1 ();
		
		this->dock_window = OP_NEW (OPWindow);
		this->dock_window->SetStyle (OP_WIN_NAKED | OP_WIN_FLOAT_TOOL | OP_WIN_TOP_LEVEL);
		this->dock_window->SetOrigin (wx, wy);
		this->dock_window->SetSize (this->box.DX (), this->box.DY ());
		
		this->SetParent (OPViewNull);
		this->SetOrigin (0, 0);
		this->SetParent (this->dock_window);
		this->dock_window->ShowNormal ();
		
		parent->Display ();
	}
	
	OPEvent::GrabWindowEvents (this->dock_window, TRUE);
	OPEvent::HandOverToWindow (this->dock_window);
	
	this->dock_window->SetDraggingView (this);
	
	this->is_dragging = TRUE;
	
	return TRUE;
}

/*
 *	When the user presses a mouse button, we might start a
 *	dragging operation...
 */

void
OPDockView::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPCoord x = event->ReturnX ();
	OPCoord y = event->ReturnY ();
	
	if (this->style & OP_DOCK_SEL_BODY) {
		this->StartDragging ();
		return;
	}
	
	if (this->style & OP_DOCK_SEL_BORDER) {
		if ( (x < this->box.X1 ()+this->border)
		  || (x > this->box.X2 ()-this->border)
		  || (y < this->box.Y1 ()+this->border)
		  || (y > this->box.Y2 ()-this->border) ) {
			this->StartDragging ();
			return;
		}
	}
	
	if (this->style & OP_DOCK_SEL_HANDLE) {
		
		switch (this->style & OP_DOCK_HANDLE_MASK) {
			
			case OP_DOCK_HANDLE_LEFT:
				if (x < this->box.X1 ()+this->width) {
					this->StartDragging ();
					return;
				}
				break;
			
			case OP_DOCK_HANDLE_TOP:
				if (y < this->box.Y1 ()+this->width) {
					this->StartDragging ();
					return;
				}
				break;
			
			default:
				break;
		}
	}
}


/*
 *	The user released the mouse button : dock the view in the
 *	window (if any) or leave the view floating around.
 */

void
OPDockView::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->is_dragging == FALSE) {
		return;
	}
	
	OPEvent::GrabWindowEvents (this->dock_window, FALSE);
	
	this->DragNotify (event);
	this->dock_window->SetDraggingView (OPViewNull);
	this->is_dragging = FALSE;
}


/*
 *	When the mouse cursor is being moved around, we have to
 *	move the docking view.
 */

void
OPDockView::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->is_dragging == FALSE) {
		return;
	}
	
	OPCoord wx = 0;
	OPCoord wy = 0;
	
	this->dock_window->GetOrigin (wx, wy);
	
	wx += event->ReturnX () - this->hot_x;
	wy += event->ReturnY () - this->hot_y;
	
	this->dock_window->SetOrigin (wx, wy);
}



static void
draw_stripe (OPGrailPortIn port, OPCoord x, OPCoord y, OPCoord dx, OPCoord dy,
			 int ix, int iy, int n)
{
	for (int num = n; num--; x += 4*ix, y += 4*iy) {
		
		port->SetForeColor (OP_COLOR_WHITE);
		port->DrawLineDX (x, y,   dx-1);
		port->DrawLineDY (x, y+1, dy-2);
		
		port->SetForeColor (OP_COLOR_DGRAY);
		port->DrawLineDX (x, y+dy-1, dx);
		port->DrawLineDY (x+dx-1, y, dy-1);
	}
}

static OPColorPixel
mix_colors (OPColorPixelIn a, OPColorPixelIn b, Real ratio)
{
	Real ra = 1.0 - ratio;
	Real rb = ratio;
	return OPColorPixel (a.RealRed () * ra + b.RealRed () * rb,
						 a.RealGreen () * ra + b.RealGreen () * rb,
						 a.RealBlue () * ra + b.RealBlue () * rb,
						 a.RealTransparency () * ra + b.RealTransparency () * rb);
}

static void
draw_bump (OPGrailPortIn port, OPCoord x, OPCoord y, OPCoord dx, OPCoord dy,
		   int ix, int iy, int n)
{
	int num;
	
	if (ix) {
		for (num = n; num; num--) {
			port->SetForeColor (mix_colors (OP_COLOR_WHITE, OP_COLOR_LGRAY, (Real)(num-1)/n));
			port->DrawLineDY (x++, y, dy);
		}
		for (num = n; num; num--) {
			port->SetForeColor (mix_colors (OP_COLOR_DGRAY, OP_COLOR_WHITE, (Real)(num-1)/n));
			port->DrawLineDY (x++, y, dy);
		}
	} else {
		for (num = n; num; num--) {
			port->SetForeColor (mix_colors (OP_COLOR_WHITE, OP_COLOR_LGRAY, (Real)(num-1)/n));
			port->DrawLineDX (x, y++, dx);
		}
		for (num = n; num; num--) {
			port->SetForeColor (mix_colors (OP_COLOR_DGRAY, OP_COLOR_WHITE, (Real)(num-1)/n));
			port->DrawLineDX (x, y++, dx);
		}
	}
}


/*
 *	Draw the docking view adorners.
 */

void
OPDockView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	if (what & OP_DRAW_BACKGROUND) {
		
		if (this->style & OP_DOCK_DRAW_FRAME) {
			adorner->FillRect (this->box, OP_COLOR_LGRAY);
			adorner->DrawFrame (this->box);
		}
	}
		
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		OPCoord ox = this->box.X1 ();
		OPCoord oy = this->box.Y1 ();
		OPCoord dx = this->box.DX ();
		OPCoord dy = this->box.DY ();
		
		switch (this->style & OP_DOCK_STRIPE_MASK) {
			
			case OP_DOCK_STRIPE_SINGLE:
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_LEFT) {
					draw_stripe (port, ox+4, oy+4, 3, dy-8, 1, 0, 1);
				}
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_TOP) {
					draw_stripe (port, ox+4, oy+4, dx-8, 3, 0, 1, 1);
				}
				break;
			
			case OP_DOCK_STRIPE_DOUBLE:
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_LEFT) {
					draw_stripe (port, ox+3, oy+3, 3, dy-6, 1, 0, 2);
				}
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_TOP) {
					draw_stripe (port, ox+3, oy+3, dx-6, 3, 0, 1, 2);
				}
				break;
			
			case OP_DOCK_STRIPE_THIN:
				break;
			
			case OP_DOCK_STRIPE_BUMP:
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_LEFT) {
					draw_bump (port, ox+3, oy+3, 3, dy-6, 1, 0, 3);
				}
				if ((this->style & OP_DOCK_HANDLE_MASK) == OP_DOCK_HANDLE_TOP) {
					draw_bump (port, ox+3, oy+3, dx-6, 3, 0, 1, 3);
				}
				break;
			
			default:
				break;
		}
	}
}

