/*
 *	iterator.cxx
 *
 *	This file exports the generic iterator class which can be
 *	used to iterate over containers, data bases and any other
 *	class derived from the generic container class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/iterator.h>
#include <OPaC/container.h>
#include <OPaC/performer.h>

OP_IMPLEMENTATION_FILE;
OP_ABSTRACT_CLASS (OPIterator);

/****************************************************************/

OPIterator::OPIterator ()
{
	OP_METHOD (OPIterator, 0);
	this->did_change = FALSE;
	this->is_valid   = FALSE;
}

OPIterator::~OPIterator ()
{
	OP_METHOD (~OPIterator, 0);
}

void
OPIterator::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPIterator>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << OPStorageSync::end;
}

void
OPIterator::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::end;
}


/*
 *	The source calls the iterator so that it knows whom it should
 *	notify on deletion...
 */

void
OPIterator::SetSource (OPGenericContainerIn source)
{
	OP_METHOD (SetSource, 1);
	OP_ARG (1, OPGenericContainerIn, source);
	
	this->source = source;
}

void
OPIterator::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->source.IsValid ()) {
		this->source->RemoveIterator (this);
		this->source = 0;
	}
	
	super::Delete ();
}

Bool
OPIterator::GetSource (OPGenericContainerOut source) const
{
	OP_METHOD (GetSource, 1);
	OP_ARG (1, OPGenericContainerOut, source);
	
	source = this->source;
	return source.IsValid ();
}

const OPGenericContainerRef&
OPIterator::ReturnSource () const
{
	OP_METHOD (ReturnSource, 0);
	return this->source;
}

Bool
OPIterator::DidChange ()
{
	OP_METHOD (DidChange, 0);
	Bool changed = this->did_change;
	this->did_change = FALSE;
	return changed;
}


OPObjectRef
OPIterator::Get (Count index)
{
	OP_METHOD (Get, 1);
	OP_ARG (1, Count, index);
	
	OPObjectRef object;
	
	if (this->Get (object, index)) {
		return object;
	}
	
	return OPObjectNull;
}

Bool
OPIterator::ForEach (OPPerformerIn performer, Count index)
{
	OP_METHOD (ForEach, 2);
	OP_ARG (1, OPPerformerIn, performer);
	OP_ARG (2, Count, index);
	
	OPObjectRef object;
	
	if (this->First ()) {
		do {
			if ( (this->Get (object, index))
			  && (performer->Perform (object) == FALSE) ) {
				return FALSE;
			}
		} while (this->Next ());
	}

	return TRUE;
}

