/*
 *	lookman.cxx
 *
 *	The look manager is used to define the user interface look
 *	depending on the color & gray models currently available.
 *	
 *	There is only one look manager for each app, but there will
 *	be exactly one adorner associated to each grail port.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPLookMan);

/****************************************************************/

OPLookManRef			OPLookMan::current;
OPLookManRef			OPLookMan::sys_lookman;
OPLookMan::AdornerList	OPLookMan::adorner;

/****************************************************************/

OP_IMPLEMENT_TYPEENUM (OPLookCursor)
{
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_DEFAULT),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_NORMAL),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_BUSY),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_I_BEAM),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_NONE),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_CURSOR_USER),
	OP_IMPLEMENT_TYPEENUM_END ()
};

OP_IMPLEMENT_TYPEENUM (OPFlushPriority)
{
	OP_IMPLEMENT_TYPEENUM_DEF (OP_FLUSH_SYNC),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_FLUSH_FAST),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_FLUSH_NORMAL),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_FLUSH_LAZY),
	OP_IMPLEMENT_TYPEENUM_END ()
};

/****************************************************************/

/*
 *	Initialise the standard look manager.
 */

OPLookMan::OPLookMan ()
{
	OP_METHOD (OPLookMan, 0);
	
	this->InitFont ();
	this->InitIcons ();
	this->InitColors ();
	this->InitCursors ();
	this->InitPolicies ();
}

void
OPLookMan::InitFont ()
{
	OP_METHOD (InitFont, 0);
	
	this->sys_font = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPGrailFont);
	this->sys_font->Initialise ("Arial", "", 12);
}

void
OPLookMan::InitCursors ()
{
	OP_METHOD (InitCursors, 0);
	
	this->cursor_normal = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPGrailCursor);
	this->cursor_busy   = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPGrailCursor);
	this->cursor_I_beam = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPGrailCursor);
	
	this->cursor_busy->Initialise (8, 8, 16, 16, this->sys_icons, 35, 19, this->sys_icons, 52, 19);
	this->cursor_I_beam->Initialise (8, 8, 16, 16, this->sys_icons, 1, 19, this->sys_icons, 18, 19);
}


/*
 *	Free the resource which have been allocated by the
 *	look manager.
 */

OPLookMan::~OPLookMan ()
{
	OP_METHOD (~OPLookMan, 0);
	OP_ASSERT (OPLookMan::adorner.head.IsValid () == FALSE);
}



/*
 *	You cannot save the look manager. This will produce
 *	a run-time error...
 */

void
OPLookMan::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store Look Manager");
}



/*
 *	You cannot restore a look manager. This will produce
 *	a run-time error...
 */

void
OPLookMan::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore Look Manager");
}



/*
 *	Prepare the look manager for its destruction.
 */

void
OPLookMan::Delete ()
{
	OP_METHOD (Delete, 0);
	super::Delete ();
}



/*
 *	Select the specified cursor. This will change the
 *	cursor view.
 */

void
OPLookMan::SelectCursor (OPLookCursor cursor)
{
	OP_METHOD (SelectCursor, 1);
	OP_ARG (1, OPLookCursor, cursor);
	
	switch (cursor) {
		
		case OP_CURSOR_NORMAL:
			this->cursor_normal->Activate ();
			break;
		
		case OP_CURSOR_BUSY:
			this->cursor_busy->Activate ();
			break;
		
		case OP_CURSOR_I_BEAM:
			this->cursor_I_beam->Activate ();
			break;
		
		case OP_CURSOR_NONE:
			this->cursor_normal->Hide ();
			break;
		
		default:
			break;
	}
}




void
OPLookMan::Initialise ()
{
	OP_MEMBER (Initialise, 0);
	OPLookMan::sys_lookman  = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPLookMan);
	OPLookMan::current      = OPLookMan::sys_lookman;
	OPLookMan::adorner.head = 0;
	OPLookMan::adorner.tail = 0;
}



void
OPLookMan::Kill ()
{
	OP_MEMBER (Kill, 0);
	OPLookMan::current     = 0;
	OPLookMan::sys_lookman = 0;
}



void
OPLookMan::Reset ()
{
	OP_MEMBER (Reset, 0);
	OPLookMan::current = OPLookMan::sys_lookman;
}



void
OPLookMan::Change (OPLookManIn look_man)
{
	OP_MEMBER (Change, 1);
	OP_ARG (1, OPLookManIn, look_man);
	OPLookMan::current = look_man;
}



void
OPLookMan::InsertAdorner (OPAdorner* adorner)
{
	OP_MEMBER (InsertAdorner, 1);
	OP_ARG (1, OPAdorner*, adorner);
/*
	adorner->sibling.next = 0;
	adorner->sibling.prev = OPLookMan::adorner.tail;
	
	(OPLookMan::adorner.tail.IsValid ()
	 ? OPLookMan::adorner.tail->sibling.next
	 : OPLookMan::adorner.head) = adorner;
	
	(OPLookMan::adorner.tail) = adorner;
*/
}



void
OPLookMan::RemoveAdorner (OPAdorner* adorner)
{
	OP_MEMBER (RemoveAdorner, 1);
	OP_ARG (1, OPAdorner*, adorner);
/*
	(adorner->sibling.next.IsValid ()
	 ? adorner->sibling.next->sibling.prev
	 : OPLookMan::adorner.tail) = adorner->sibling.prev;
	(adorner->sibling.prev.IsValid ()
	 ? adorner->sibling.prev->sibling.next
	 : OPLookMan::adorner.head) = adorner->sibling.next;
*/
}

/****************************************************************/
