/*
 *	monitorana.cxx
 *
 *	Analogic data monitor. This is a kind of gauge.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/


#include <OPaC/convert.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/monitorana.h>
#include <OPaC/new.h>
#include <OPaC/values.h>

#include <OPaC/pubmacro.h>

#include <math.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPMonitorAnalogic);

OP_PUBLISH_SLOT (OPMonitorAnalogic, max_value, ReturnMaximum, SetMaximum, "Maximum", Real);
OP_PUBLISH_SLOT (OPMonitorAnalogic, is_symmetric, IsSymmetric, SetSymmetric, "Is symmetric", Bool);
OP_PUBLISH_SLOT (OPMonitorAnalogic, is_rectified, IsRectified, SetRectified, "Is rectified", Bool);
OP_PUBLISH_SLOT (OPMonitorAnalogic, is_dynamic_max, IsDynamicMax, SetDynamicMax, "Is dynamic maxium", Bool);

/****************************************************************/

	
OPMonitorAnalogic::OPMonitorAnalogic ()
{
	this->shape |= OP_CTRL_RELIEF;
	
	this->dis_value = 0.0;
	this->max_value = 1.0;
	
	this->is_symmetric   = FALSE;
	this->is_rectified   = FALSE;
	this->negative_flag  = FALSE;
	this->is_dynamic_max = FALSE;
}

OPMonitorAnalogic::~OPMonitorAnalogic ()
{
}


/*
 *	Save the monitor's parameters.
 */

void
OPMonitorAnalogic::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPMonitorAnalogic max_value is_symmetric is_rectified is_dynamic_max>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->max_value;
	storage << this->is_symmetric << this->is_rectified << this->is_dynamic_max;
	storage << OPStorageSync::end;
}



/*
 *	Restore the monitor's parameters.
 */

void
OPMonitorAnalogic::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->max_value;
	storage >> this->is_symmetric >> this->is_rectified >> this->is_dynamic_max;
	storage >> OPStorageSync::end;
}



void
OPMonitorAnalogic::DoUpdate ()
{
	OPValueRef v = this->values[0];
	Real x = v;
	Bool negative = (x < 0) ? TRUE : FALSE;
	Real abs_x    = (x < 0) ? -x : x;
	
	if ( (!is_symmetric)
	  && (!is_rectified)
	  && (negative) ) {
		x = 0.0;
	}
	
	if (is_rectified) {
		x = abs_x;
	}
	
	if (abs_x > this->max_value) {
		
		Real sign = (x < 0) ? -1 : 1;
		
		if (this->is_dynamic_max) {
			int  num_digits = (int) log10 (abs_x);
			Real power_10   = pow (10.0, num_digits);
			
			if ( (abs_x / power_10) < 5.0) {
				this->max_value = power_10 * 5.0;
			} else {
				this->max_value = power_10 * 10.0;
			}
		} else {
			x = sign * this->max_value;
		}
	}
	
	this->dis_value = x;
	this->negative_flag = negative;
}



Bool
OPMonitorAnalogic::SetMaximum (Real max)
{
	if (max < 0) max = -max;
	this->max_value = max;
	return TRUE;
}

Bool
OPMonitorAnalogic::SetSymmetric (Bool yes)
{
	this->is_symmetric = yes;
	return TRUE;
}

Bool
OPMonitorAnalogic::SetRectified (Bool yes)
{
	this->is_rectified = yes;
	return TRUE;
}

Bool
OPMonitorAnalogic::SetDynamicMax (Bool yes)
{
	this->is_dynamic_max = yes;
	return TRUE;
}


/*
 *	Draw an analogic gauge, based on a thin needle. A graduation
 *	is provided in the background.
 */

void
OPMonitorAnalogic::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	OPCoord margin = (this->shape & OP_CTRL_NOFRAME) ? 0 : lookman->ReturnWidgetFrameSize ();
	OPCoord effect = (this->shape & OP_CTRL_RELIEF) ? lookman->Return3DEffectWidth () : 0;
	OPCoord delta  = margin + effect;
	
	if (what & OP_DRAW_BACKGROUND) {
		adorner->FillRect (this->box, OP_COLOR_WHITE, delta);
	}
	
	if (what & OP_DRAW_OBJECT) {
		
		int w = (int)(this->box.DX () - 2*delta) / 2;
		int h = (int)(this->box.DY () - 2*delta);
		int r = h - 2;
		
		if ( (r < 5) || (w < 10) ) return;
		
		if (w > r) w = r;
		
		adorner->SetForeColor (OP_COLOR_BLACK);
		
		Real alpha = asin ((Real)(w) / (Real)(r));
		Real aleft = -alpha;
		Real azero = 0.0;
		Real ratio = this->dis_value / this->max_value;
		Real max   = this->max_value;
		
		if (! this->is_symmetric) {
			azero  = -alpha;
			alpha *= 2;
			aleft  = 0.0;
		}
		
		Real angle = ratio * alpha + azero;
		
		Real x = this->box.X1 () + delta + w;
		Real y = this->box.Y1 () + delta + h - 1;
		
		OPStringRef text;
		int  num_digits = (int) log10 (max);
		Real power_10   = pow (10.0, num_digits);
		Card32 int_max  = (int)(max / power_10);
		
		if (this->is_symmetric) {
			
			for (int d = -5; d <= 5; d++) {
				Real x1 = x + sin (d * alpha / 5 + azero) * r * 0.9;
				Real y1 = y - cos (d * alpha / 5 + azero) * r * 0.9;
				Real x2 = x + sin (d * alpha / 5 + azero) * r;
				Real y2 = y - cos (d * alpha / 5 + azero) * r;
				port->DrawAALineTo (x1, y1, x2, y2);
			}
			
			OPConvert::CardToString (int_max, "NUM:-%;", text);
			port->ShowText (x + sin (-alpha + azero) * r * 0.8,
							y - cos (-alpha + azero) * r * 0.8,
							text, 0, 0,
							OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER);
			
			port->ShowText (x + 1 * r * 0.8,
							y - 0 * r * 0.8,
							"0", 0, 0,
							OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER);
			
			OPConvert::CardToString (int_max, "NUM:+%;", text);
			port->ShowText (x + sin (alpha + azero) * r * 0.8,
							y - cos (alpha + azero) * r * 0.8,
							text, 0, 0,
							OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER);
			
		} else {
			
			for (int d = 0; d <= 10; d++) {
				Real x1 = x + sin (d * alpha / 10 + azero) * r * 0.9;
				Real y1 = y - cos (d * alpha / 10 + azero) * r * 0.9;
				Real x2 = x + sin (d * alpha / 10 + azero) * r;
				Real y2 = y - cos (d * alpha / 10 + azero) * r;
				port->DrawAALineTo (x1, y1, x2, y2);
			}
			
			port->ShowText (x + sin (azero) * r * 0.8,
							y - cos (azero) * r * 0.8,
							"0", 0, 0,
							OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER);
			
			OPConvert::CardToString (int_max, "NUM:%;", text);
			port->ShowText (x + sin (alpha + azero) * r * 0.8,
							y - cos (alpha + azero) * r * 0.8,
							text, 0, 0,
							OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER);
		}
		
		OPConvert::CardToString ((int)(power_10), "NUM:x %;", text);
		
		port->ShowText (x + sin (alpha + azero) * r * 0.9,
						y - cos (alpha + azero) * r * 0.7,
						text, 0, 0,
						OP_TEXT_HPOS_RIGHT | OP_TEXT_VPOS_TOP);
		
		port->ShowText (x + sin (aleft + azero) * r * 0.9,
						y - cos (aleft + azero) * r * 0.7,
						(this->negative_flag) ? "-" : "+", 0, 0,
						OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_TOP);
		
		Real dx = + sin (angle)*r;
		Real dy = - cos (angle)*r;
		
		port->DrawAALine (x, y, dx, dy);
	}
}
