/*
 *	new.cxx
 *
 *	This file implements the C++ memory allocation operators and the
 *	associated debugging functions.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/memory.h>
#include <OPaC/string.h>
#include <OPaC/new.h>
#include <OPaC/except.h>

OP_IMPLEMENTATION_FILE;

/****************************************************************/

#ifdef _DEBUG
#include <stdio.h>
#endif /* _DEBUG */

/****************************************************************/

#define	NEW_HEADER_ID		0xFEA55AEF
#define	NUM_HEAD_IN_ARRAY   127
#define	MIN_BUF_SIZE		16384

/****************************************************************/

OPMemZone* OPaC_data_zone  = 0;
OPMemZone* OPaC_app_zone   = 0;
OPMemZone* OPaC_temp_zone  = 0;
OPMemZone* OPaC_iface_zone = 0;
OPMemZone* OPaC_sys_zone   = 0;

OPMemZone* OPMemZone::zone_head = 0;
OPMemZone* OPMemZone::zone_tail = 0;
Card32     OPMemZone::vector_offset = 0;

/****************************************************************/

static void* last_delete_ptr = 0;

/****************************************************************/

/*
 *	Each memory chunk allocated through operator new will get
 *	a header containing some debug information.
 */

struct NewHeader
{
	Card32				size;					//	memory size (without header)
	Count				source_line;			//	source line # in file
	const LatinChar*	source_file;			//	source file name
	const LatinChar*	source_type;			//	source data type
	
	Card32				header_id;				//	id to check against underflow
};

/****************************************************************/

/*
 *	The operator "new" has been prohibited. You *may not*
 *	use it in the OPaC applications. There are lots of
 *	other variants defined as macros, so use them !
 */

void*
opac_operator_new (size_t size)
{
	OP_OPERATOR (new, 1);
	OP_ARG (1, size_t, size);
	OP_ABORT ("No <operator new> allowed in application.");
	return 0;
}



/*
 *	A special variant of the "new" operator exists in order
 *	to call *just* the constructor using a pre-allocated
 *	buffer. There is an equivalent destructor definition.
 */

void*
opac_operator_new (size_t size, OPRuntimeIn runtime,
				   void* buffer_ptr, Card32 buffer_len,
				   Count line, const LatinChar* file)
{
	OP_OPERATOR (new, 5);
	OP_ARG (1, OPRuntimeIn, runtime);
	OP_ARG (2, void*, buffer_ptr);
	OP_ARG (3, Card32, buffer_len);
	OP_ARG (4, Count, line);
	OP_ARG (5, const LatinChar*, file);
	
	OP_ASSERT (size <= buffer_len);
	OP_ASSERT (buffer_len);
	OP_ASSERT (buffer_ptr);
	OP_ASSERT (runtime.IsValid ());
	
	return buffer_ptr;
}

void*
opac_operator_new (size_t size, void* buffer_ptr,
				   Count line, const LatinChar* file)
{
	OP_OPERATOR (new, 4);
	OP_ARG (1, size_t, size);
	OP_ARG (2, void*, buffer_ptr);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	
	OP_ASSERT (size);
	OP_ASSERT (buffer_ptr);
	
	return buffer_ptr;
}



/*
 *	A special operator "delete" won't free any memory, but
 *	it will just call the destructor. You should never have
 *	to call it yourself.
 */

void
opac_operator_delete (void* ptr)
{
	OP_OPERATOR (delete, 1);
	OP_ARG (1, void*, ptr);
	last_delete_ptr = ptr;
}



/*
 *	The NEW_xxx macros will have to use this special "new"
 *	operator, which will allocate memory and also call the
 *	constructor.
 *
 *	Throws :
 *
 *	  - OPExcMemory (FULL) if no more memory is available.
 */

void*
opac_operator_new (size_t size, OPMemZone* zone,
				   const LatinChar* type,
				   Count line, const LatinChar* file)
{
	OP_OPERATOR (new, 5);
	OP_ARG (1, size_t, size);
	OP_ARG (2, OPMemZone*, zone);
	OP_ARG (3, const LatinChar*, type);
	OP_ARG (4, Count, line);
	OP_ARG (5, const LatinChar*, file);
	OP_ASSERT (zone);
	
	return zone->Alloc (size, type, line, file);
}

/****************************************************************/

struct DummyObject
{
	Card32	x;
	DummyObject ();
	virtual ~DummyObject ();
};

DummyObject::DummyObject () { x = 0; }
DummyObject::~DummyObject () { }

/****************************************************************/

/*
 *	Create the memory zone. It still needs initialization
 *	in order to be ready for use.
 */

OPMemZone::OPMemZone ()
{
	OP_METHOD (OPMemZone, 0);
	
	static Bool offset_ok = FALSE;
	
	if (offset_ok == FALSE) {
		
		//	Try to find out how much memory is really allocated when an
		//	array is being built with operator new. Most compilers add
		//	a counter at the beginning of the memory zone and return a
		//	pointer to the first element, not the beginning of the memory.
		
		char buffer[sizeof (DummyObject)*2+64];
		
		DummyObject* dummy = NEW_USING (buffer) DummyObject[2];
		delete[] dummy;
		
		//	Find out where the memory really starts. The offset has to be
		//	recorded so we can adjust the pointer when freeing vectors of
		//	objects.
		
		Address address = (Address)(dummy);
		Address freed = (Address)(last_delete_ptr);
		
		OPMemZone::vector_offset = (Card32)(address - freed);
		
		offset_ok = TRUE;
	}
	
	this->prev = this->zone_tail;
	this->next = 0;
	
	(this->next ? this->next->prev : this->zone_tail) = this;
	(this->prev ? this->prev->next : this->zone_head) = this;
	
	this->is_zone_ok = FALSE;
	this->sys_zone   = 0;
	
	this->buf_head = 0;
	this->buf_tail = 0;
	
	this->array_head = 0;
	this->array_tail = 0;
	
	this->buf_allocated = 0;
	this->buf_remaining = 0;
	
	this->latest_allocation = 0;
#ifdef _DEBUG
	this->fill_on_new    = TRUE;
	this->fill_on_delete = TRUE;
	this->data_on_new    = 0xEEEEEEEE;		//	empty
	this->data_on_delete = 0xDDDDDDDD;		//	deleted
#else
	this->fill_on_new    = FALSE;
	this->fill_on_delete = FALSE;
	this->data_on_new    = 0x00000000;
	this->data_on_delete = 0x00000000;
#endif
	
	this->zone_name = "<NO NAME>";
	
	for (int i = 0; i < MEM_STAT_SMALL; i++) {
		this->stat_alloc[i] = 0;
	}
	
	this->stat_alloc_large = 0;
}



/*
 *	Destroy the memory zone and free the associated
 *	buffers.
 */

OPMemZone::~OPMemZone ()
{
	OP_METHOD (~OPMemZone, 0);
	
#ifdef _DEBUG
	if (this->buf_allocated != this->buf_remaining) {
		
		const char* what[] = { "BIN", "OBJ", "ARR" };
		
		//	Hmmm. Somebody forgot to free memory. Write a list to a debugging
		//	file so that we can find out later where we didn't free the memory.
		
		BufferHeader* buffer = this->buf_head;
		FILE*         file   = fopen ("leak-mem.log", "at"); 
		
		fprintf (file, "----------------------------------------\n");
		fprintf (file, "  Zone %s\n", this->zone_name);
		fprintf (file, "----------------------------------------\n");
		
		for (int i = 1; i < MEM_STAT_SMALL; i++) {
			fprintf (file, "%d bytes:\t%d\n", i << SYS_ALIGN_SHIFT, this->stat_alloc[i]);
		}
		fprintf (file, "More bytes:\t%d\n", this->stat_alloc_large);
		fprintf (file, "----------------------------------------\n");
		
		while (buffer) {
			ChunkHeader* chunk = buffer->head;
			while (chunk) {
				if (chunk->mem_type != OP_MTYPE_FREE) {
					fprintf (file, "%s:%d (%s) - %s, %d",
							 (const char*)(chunk->source_file),
							 (int)(chunk->source_line),
							 (const char*)(chunk->source_type),
							 what[chunk->mem_type],
							 (int)(chunk->call_id));
					
					if (chunk->mem_type == OP_MTYPE_OBJECT) {
						void* ptr = chunk->ptr;
#ifdef _DEBUG
						ptr = (void*)(((Address)(ptr)) + sizeof (void*));
#endif
						OPObject* obj = (OPObject*)(ptr);
						fprintf (file, " (ref %ld)", obj->ref_count);
					}
					
					fprintf (file, "\n");
				}
				chunk = chunk->next;
			}
			buffer = buffer->next;
		}
		
		fprintf (file, "\n");
		fclose (file);
	}
#endif /* _DEBUG */
	
	if (this->is_zone_ok) {
		this->FreeAndDestroy ();
	}
	
	(this->next ? this->next->prev : this->zone_tail) = this->prev;
	(this->prev ? this->prev->next : this->zone_head) = this->next;
}



/*
 *	Free the memory zone. This will be called by the
 *	destructor and by Kill for the system memory zone.
 *	This does not remove the zone from the zone list,
 *	just frees the memory... and remove the system's
 *	zone.
 */

void
OPMemZone::FreeAndDestroy ()
{
	OP_METHOD (FreeAndDestroy, 0);
	
	if (this->is_zone_ok) {
		
		this->KillRefFromZone ();
		this->KillRefToZone ();
		
		//	Free the memory used by the buffers.
		
		BufferHeader* buffer  = this->buf_head;
		BufferHeader* current = 0;
		
		while (buffer) {
			
			current = buffer;
			buffer  = buffer->next;
			
			SysMemFree (this->sys_zone,
						current->buffer, current->size, current->mem_id);
		}
		
		//	Now free memory used by the buffer and chunk headers.
		
		HeaderArray* array = this->array_head;
		HeaderArray* next  = 0;
		
		while (array) {
			next = array->next;
			SysMemFree (this->sys_zone, array->buffer, array->size, array->mem_id);
			array = next;
		}
		
		SysMemCloseZone (this->sys_zone);
	}
	
	this->is_zone_ok = FALSE;
	this->sys_zone   = 0;
	
	this->buf_head = 0;
	this->buf_tail = 0;
	
	this->array_head = 0;
	this->array_tail = 0;
	
	this->buf_allocated = 0;
	this->buf_remaining = 0;

	this->latest_allocation = 0;
}



/*
 *	Initialise the zone and give it a name. This name will
 *	be passed to the system and is only needed as a debug
 *	information.
 */

Bool
OPMemZone::Initialise (const LatinChar* name)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, const LatinChar*, name);
	
	if (this->is_zone_ok) {
		return TRUE;
	}
	
	this->is_zone_ok = SysMemOpenZone (name, this->sys_zone);
	this->zone_name  = name;
	
	return this->is_zone_ok;
}



/*
 *	Allocate a header (either for a new buffer or for a
 *	new chunk).
 */

OPMemZone::Header*
OPMemZone::NewHeader ()
{
	OP_METHOD (NewHeader, 0);
	OP_ASSERT (this->is_zone_ok);
	
	HeaderArray* array = this->array_head;
	
	while (array) {
		if (array->num_free) {
			Header* free = array->free;
			array->free  = free->next;
			array->num_free--;
			
			return free;
		}
		array = array->next;
	}
	
	//	There are no more free headers in the header array(s). We will have
	//	to add a header array !
	
	BytePtr ptr       = 0;
	Card32  allocated = 0;
	Card32  mem_id    = 0;
	
	if (SysMemAlloc (this->sys_zone, sizeof (HeaderArray),
					 ptr, allocated, mem_id) == FALSE) {
		return 0;
	}
	
	array = (HeaderArray*)(ptr);
	
	array->next = this->array_head;
	array->prev = 0;
	
	(array->prev ? array->prev->next : this->array_head) = array;
	(array->next ? array->next->prev : this->array_tail) = array;
	
	array->size   = allocated;
	array->buffer = ptr;
	array->mem_id = mem_id;
	
	array->num_free = NUM_HEAD_IN_ARRAY - 1;
	array->free     = & array->array[1];
	array->padding  = 0;
	
	//	Create the linked list of free headers in the array.
	
	for (int i = 1; i < NUM_HEAD_IN_ARRAY - 1; i++) {
		array->array[i].next = & array->array[i+1];
	}
	
	array->array[NUM_HEAD_IN_ARRAY-1].next = 0;
	
	return & array->array[0];
}



/*
 *	Delete a header and put it back into the array of free
 *	headers.
 */

void
OPMemZone::DeleteHeader (Header* header)
{
	OP_METHOD (DeleteHeader, 1);
	OP_ARG (1, Header*, header);
	
	if (this->latest_allocation == (ChunkHeader*)(header)) {
		this->latest_allocation = 0;
	}
	
	HeaderArray* array = this->array_head;
	Address      addr  = (Address)(header);
	
	while (array) {
		
		if ( ((Address)(& array->array[0]) <= (addr))
		  && ((Address)(& array->array[NUM_HEAD_IN_ARRAY-1]) >= (addr)) ) {
			header->next = array->free;
			array->free = header;
			array->num_free++;
			return;
		}
		
		array = array->next;
	}
	
	OP_ABORT ("Not found matching header array.");
}



/*
 *	Find the biggest free hole in the buffer.
 */

void
OPMemZone::UpdateFreeHole (BufferHeader* buffer)
{
	OP_METHOD (UpdateFreeHole, 1);
	OP_ARG (1, BufferHeader*, buffer);
	
	Card32       size = 0;
	ChunkHeader* iter = buffer->head;
	
	while (iter) {
		
		if ( (iter->mem_type == OP_MTYPE_FREE)
		  && (iter->size > size) ) {
			size = iter->size;
		}
		
		iter = iter->next;
	}
	
	buffer->free_hole = size;
}



/*
 *	Merge the chunk and the following one. This can be done
 *	for any kind of chunk.
 */

void
OPMemZone::MergeChunks (ChunkHeader* chunk, BufferHeader* buffer)
{
	OP_METHOD (MergeChunks, 2);
	OP_ARG (1, ChunkHeader*, chunk);
	OP_ARG (2, BufferHeader*, buffer);
	
	ChunkHeader* next = chunk->next;
	
	chunk->size += next->size;
	chunk->next  = next->next;
	
	(chunk->next ? chunk->next->prev : buffer->tail) = chunk;
	
	Header* header = (Header*)(next);
	this->DeleteHeader (header);
}



/*
 *	Allocate memory for the specified caller. Remember where
 *	the allocation has been done to facilitate the debugging.
 *
 *	Throws :
 *
 *	  - OPExcMemory (FULL) if no more memory is available.
 */

void*
OPMemZone::Alloc (Card32 size, const LatinChar* type,
                  Count line, const LatinChar* file,
				  Card8 mem_type)
{
	OP_METHOD (Alloc, 5);
	OP_ARG (1, Card32, size);
	OP_ARG (2, const LatinChar*, type);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	OP_ARG (5, Card8, mem_type);
	
#ifdef _DEBUG
	size += sizeof (void*);
#endif /* _DEBUG */
	size += SYS_ALIGN_ROUND;
	size &= SYS_ALIGN_MASK;
	
	if (size == 0) size = SYS_ALIGN_SIZE;
	
	Count stat_index = size >> SYS_ALIGN_SHIFT;
	
	if (stat_index < MEM_STAT_SMALL) {
		this->stat_alloc[stat_index]++;
	} else {
		this->stat_alloc_large++;
	}
	
	Count affected_size = size;
	
	//	First allocate the chunk header, then try to find some memory in one of
	//	the memory buffers.
	
	Header*       header = this->NewHeader ();
	BufferHeader* buffer = this->buf_head;
	
	BufferHeader* the_buf   = 0;
	ChunkHeader*  the_chunk = 0;
	Card32        the_size  = 0xFFFFFFFFL;
	
	while (buffer) {
		
		//	Scan every buffer which might satisfy the request.
		
		if (buffer->free_hole >= size) {
			
			//	We might find some matching hole in this buffer. Just scan every
			//	available chunk and select the smallest one, which satisfies the
			//	request.
			
			ChunkHeader* chunk = buffer->head;
			
			while (chunk) {
				
				if ( (chunk->mem_type == OP_MTYPE_FREE)
				  && (chunk->size >= size)
				  && (chunk->size < the_size) ) {
					
					the_buf   = buffer;
					the_chunk = chunk;
					the_size  = chunk->size;
					
					if (the_size == size) {
						goto matching_size_found;
					}
				}
				
				chunk = chunk->next;
			}
		}
		
		buffer = buffer->next;
	}
	
matching_size_found:
	
	if (the_buf == 0) {
		
		//	There is no buffer available, so just allocate one which would be
		//	large enough to satisfy the request.
		
		Header* hd_buffer = this->NewHeader ();
		Header* hd_chunk  = this->NewHeader ();
		BytePtr ptr       = 0;
		Card32  allocated = 0;
		Card32  mem_id    = 0;
		Card32  buf_size  = (MIN_BUF_SIZE < size ? size : MIN_BUF_SIZE);
		
		if ( (hd_buffer == 0)
		  || (hd_chunk == 0)
		  || (SysMemAlloc (this->sys_zone, buf_size, ptr, allocated, mem_id) == FALSE) ) {
			
			this->DeleteHeader (hd_buffer);
			this->DeleteHeader (hd_chunk);
			this->DeleteHeader (header);
			
			throw OPExcMemory (OPExcMemory::FULL, 0, line, file);
		}
		
		the_buf   = & hd_buffer->buffer;
		the_chunk = & hd_chunk->chunk;
		the_size  = allocated;
		
		//	Initialise the buffer header and insert it into the buffer list.
		
		the_buf->next = this->buf_head;
		the_buf->prev = 0;
		
		(the_buf->prev ? the_buf->prev->next : this->buf_head) = the_buf;
		(the_buf->next ? the_buf->next->prev : this->buf_tail) = the_buf;
		
		the_buf->head = the_chunk;
		the_buf->tail = the_chunk;
		
		the_buf->size   = the_size;
		the_buf->buffer = ptr;
		the_buf->mem_id = mem_id;
		
		the_buf->free_hole = the_size;
		
		//	Now, initialise the only chunk which will contain the whole free
		//	memory.
		
		the_chunk->next     = 0;
		the_chunk->prev     = 0;
		the_chunk->size     = the_size;
		the_chunk->ptr      = the_buf->buffer;
		the_chunk->mem_type = OP_MTYPE_FREE;
		
		//	And now for the statistics :
		
		this->buf_allocated += allocated;
		this->buf_remaining += allocated;
	}
	
	//	Extract a chunk to satisfy the request. It will have to be split if it
	//	is too large.
	
	if (size < the_size) {
		
		//	Insert the new chunk header in the list.
		
		ChunkHeader* new_chunk = & header->chunk;
		
		new_chunk->next = the_chunk;
		new_chunk->prev = the_chunk->prev;
		
		(new_chunk->prev ? new_chunk->prev->next : the_buf->head) = new_chunk;
		(new_chunk->next ? new_chunk->next->prev : the_buf->tail) = new_chunk;
		
		affected_size = the_size;
			
		if (the_chunk->next == 0) {
			
			//	If we have to split the last chunk, rather use the top part than
			//	the bottom part, so that probability of a future merging increases.
			
			new_chunk->size = the_size - size;
			the_chunk->size = size;
			new_chunk->ptr  = the_chunk->ptr;
			the_chunk->ptr += new_chunk->size;
			
			//	The inserted chunk will be the free one. The other chunk will
			//	become used.
			
			new_chunk->mem_type = OP_MTYPE_FREE;
			
		} else {
			
			//	Usually, the bottom part will be used and the top part will be
			//	free... The newly inserted chunk points to the bottom part.
			
			new_chunk->size = size;
			the_chunk->size = the_size - size;
			new_chunk->ptr  = the_chunk->ptr;
			the_chunk->ptr += new_chunk->size;
			
			the_chunk = new_chunk;
		}
		
	} else {
		
		//	Easy : the chunk has just the correct size. It can just be used as
		//	is... Remove the previously allocated header.
		
		this->DeleteHeader (header);
	}
	
	//	Ok. Let's initialise the chunk description and remember the chunk to
	//	be used (no longer free).
	
	the_chunk->source_line = line;
	the_chunk->source_file = file;
	the_chunk->source_type = type;
	the_chunk->mem_type    = mem_type;
	the_chunk->call_id     = OPaC_CallFrame::ReturnCallerFrame ()->ReturnCallID ();
	
	//	If the programmer asked for a special initialisation of the memory on
	//	allocation, then do it. This is useful to find the uninitialised fields
	//	of a structure !
	
	if (this->fill_on_new) {
		Card32* data  = (Card32*)(the_chunk->ptr);
		Card32  count = size / 4;
		while (count--) *data++ = this->data_on_new;
	}
	
	//	And now for the statistics :
	
	this->buf_remaining -= size;
	
	if (affected_size == the_buf->free_hole) {
		
		//	Since we hacked into the biggest free hole, we will have to look
		//	for another biggest free hole.
		
		this->UpdateFreeHole (the_buf);
	}
	
#ifdef _DEBUG
	void** pptr = (void**)(the_chunk->ptr);
	*pptr++ = (void*)(the_chunk);
	void* data = (void*)(pptr);
#else /* _DEBUG */
	void* data = the_chunk->ptr;
#endif /* else _DEBUG */
	
	this->latest_allocation = the_chunk;
	
	if (mem_type == OP_MTYPE_OBJECT) {
		((OPObject*)(data))->mem_zone = this;
	}
	
	return data;
}



/*
 *	Free memory referenced by a pointer. The memory must
 *	have been allocated with this MemZone's Alloc method.
 *	
 *	An attempt to free memory associated to a null or an
 *	invalid pointer results in an exception.
 *	
 *	Throws :
 *
 *	  - OPExcMemory (NULL_PTR) if the pointer is NULL.
 *	
 *	  - OPExcMemory (BAD_PTR) if the pointer does not refer
 *		memory contained in this memory zone.
 *	
 *	  - OPExcMemory (FREE) if the pointer refers to memory
 *		which has already been freed.
 */

void
OPMemZone::FreeVector (void* ptr, Count line, const LatinChar* file)
{
	OP_METHOD (FreeVector, 3);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	ptr = (void*)((Address)(ptr) - vector_offset);
	
	this->Free (ptr, line, file);
}

void
OPMemZone::Free (void* ptr, Count line, const LatinChar* file)
{
	OP_METHOD (Free, 3);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	if (ptr == 0) {
		throw OPExcMemory (OPExcMemory::NULL_PTR, 0, line, file);
	}
	
#ifdef _DEBUG
	ptr = (void*)(((Address)(ptr)) - sizeof (void*));
#endif
	
	BufferHeader* the_buf   = this->buf_head;
	ChunkHeader*  the_chunk = 0;
	Address       address   = (Address)(ptr);
	
	while (the_buf) {
		
		if ( ((Address)(the_buf->buffer) <= address)
		  && ((Address)(the_buf->buffer + the_buf->size) > address) ) {
			
			//	The pointer belongs to this buffer. Now find the matching chunk
			//	in the buffer. The pointer must match the chunk's start exactly.
			
			the_chunk = the_buf->head;
			
			while (the_chunk) {
				if (the_chunk->ptr == (BytePtr)(ptr)) {
					if (the_chunk->mem_type == OP_MTYPE_FREE) {
						throw OPExcMemory (OPExcMemory::FREE, ptr, line, file);
					}
					
					//	Ok, pointer is valid and chunk will now be freed and maybe
					//	recycled and merged with the previous and the next one.
					
					if (this->fill_on_delete) {
						Card32* data  = (Card32*)(ptr);
						Card32  count = the_chunk->size / 4;
						while (count--) *data++ = this->data_on_delete;
					}
					
					this->buf_remaining += the_chunk->size;
					
					the_chunk->mem_type = OP_MTYPE_FREE;
					
					if ( (the_chunk->prev)
					  && (the_chunk->prev->mem_type == OP_MTYPE_FREE) ) {
						the_chunk = the_chunk->prev;
						this->MergeChunks (the_chunk, the_buf);
					}
					
					if ( (the_chunk->next)
					  && (the_chunk->next->mem_type == OP_MTYPE_FREE) ) {
						this->MergeChunks (the_chunk, the_buf);
					}
					
					if (the_chunk->size > the_buf->free_hole) {
						the_buf->free_hole = the_chunk->size;
					}
					
					return;
				}
				
				the_chunk = the_chunk->next;
			}
			
			//	The pointer belongs to the buffer, yet we cannot find any matching
			//	chunk : the user has certainly already freed this pointer before.
			
			throw OPExcMemory (OPExcMemory::FREE, ptr, line, file);
		}
		
		the_buf = the_buf->next;
	}
	
	throw OPExcMemory (OPExcMemory::BAD_PTR, ptr, line, file);
}



/*
 *	Enable or disable the debug mode for operator new.
 */

void
OPMemZone::SetAllocDebugMode (Bool on, Card32 fill)
{
	OP_METHOD (SetAllocDebugMode, 2);
	OP_ARG (1, Bool, on);
	OP_ARG (2, Card32, fill);
	
	this->fill_on_new = on;
	this->data_on_new = fill;
}



/*
 *	Enable or disable the debug mode for operator delete.
 */

void
OPMemZone::SetFreeDebugMode (Bool on, Card32 fill)
{
	OP_METHOD (SetFreeDebugMode, 2);
	OP_ARG (1, Bool, on);
	OP_ARG (2, Card32, fill);
	
	this->fill_on_delete = on;
	this->data_on_delete = fill;
}



/*
 *	Return the base pointer, using the specified pointer
 *	into one of the memory chunks of this zone.
 */

const void*
OPMemZone::FindBasePointer (const void* ptr)
{
	OP_METHOD (FindBasePointer, 1);
	OP_ARG (1, const void*, ptr);
	
	BufferHeader* the_buf   = this->buf_head;
	ChunkHeader*  the_chunk = 0;
	Address       address   = (Address)(ptr);
	
	while (the_buf) {
		
		if ( ((Address)(the_buf->buffer) <= address)
		  && ((Address)(the_buf->buffer + the_buf->size) > address) ) {
			
			the_chunk = the_buf->head;
			
			while (the_chunk) {
				if ( (the_chunk->mem_type != OP_MTYPE_FREE)
				  && ((Address)(the_chunk->ptr) <= address)
				  && ((Address)(the_chunk->ptr + the_chunk->size) > address) ) {
					return (const void*)(the_chunk->ptr);
				}
				the_chunk = the_chunk->next;
			}
			
			return 0;
		}
		
		the_buf = the_buf->next;
	}
	
	return 0;
}



/*
 *	Return informations about a memory chunk base pointer.
 */

Bool
OPMemZone::AboutBasePointer (const void* ptr,
							 OPStringOut type,
							 CountOut line,
							 OPStringOut file)
{
	OP_METHOD (AboutBasePointer, 4);
	OP_ARG (1, const void*, ptr);
	OP_ARG (2, OPStringOut, type);
	OP_ARG (3, CountOut, line);
	OP_ARG (4, OPStringOut, file);
	
	BufferHeader* the_buf   = this->buf_head;
	ChunkHeader*  the_chunk = 0;
	Address       address   = (Address)(ptr);
	
	while (the_buf) {
		
		if ( ((Address)(the_buf->buffer) <= address)
		  && ((Address)(the_buf->buffer + the_buf->size) > address) ) {
			
			//	The pointer belongs to this buffer. Now find the matching chunk
			//	in the buffer. The pointer must match the chunk's start exactly.
			
			the_chunk = the_buf->head;
			
			while (the_chunk) {
				if ( (the_chunk->mem_type != OP_MTYPE_FREE)
				  && ((Address)(the_chunk->ptr) == address) ) {
					
					type = the_chunk->source_type;
					line = the_chunk->source_line;
					file = the_chunk->source_file;
					
					return TRUE;
				}
				
				the_chunk = the_chunk->next;
			}
			
			//	The pointer belongs to the buffer, yet we cannot find any matching
			//	chunk : the user has certainly already freed this pointer before.
			
			return FALSE;
		}
		
		the_buf = the_buf->next;
	}
	
	return FALSE;
}



/*
 *	Look for the zone which contains the memory pointed to
 *	by the given pointer.
 */

Bool
OPMemZone::FindZone (const void* ptr, OPMemZone*& zone)
{
	OP_MEMBER (FindZone, 2);
	OP_ARG (1, const void*, ptr);
	OP_ARG (2, OPMemZone*&, zone);
	
	OPMemZone* iter = OPMemZone::zone_head;
	Address    addr = (Address)(ptr);
	
	while (iter) {
		
		BufferHeader* buffer = iter->buf_head;
		
		//	Scan every buffer of the current zone, until we find a matching
		//	address.
		
		while (buffer) {
			if ( ((Address)(buffer->buffer) <= addr)
			  && ((Address)(buffer->buffer + buffer->size) > addr) ) {
				zone = iter;
				return TRUE;
			}
			
			buffer = buffer->next;
		}
		
		//	Not found in this zone, look for the next one...
		
		iter = iter->next;
	}
	
	return FALSE;
}



/*
 *	Check to see if the pointer is a valid pointer, one
 *	which points into one of the OPaC zones.
 */

Bool
OPMemZone::CheckPointer (const void* ptr)
{
	OP_MEMBER (CheckPointer, 1);
	OP_ARG (1, const void*, ptr);
	
	OPMemZone* zone = 0;
	
	return OPMemZone::FindZone (ptr, zone);
}



/*
 *	Tried to allocate from some NULL zone. This produces
 *	a fatal error.
 */

void*
OPMemZone::NullAllocZone (const LatinChar* zone, Count line, const LatinChar* file)
{
	OP_MEMBER (NullAllocZone, 3);
	OP_ARG (1, const LatinChar*, zone);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	//	Concatenate "Alloc Null Zone : " and "zone_name" to produce the final fatal
	//	error message.
	
	const LatinChar* src = "Alloc Null Zone : ";
	LatinChar buffer[200];
	int index = 0;
	
	for (; (*src) && (index < 199); index++) buffer[index] = *src++;
	for (; (*src) && (index < 199); index++) buffer[index] = *zone++;
	buffer[index] = 0;
	
	OPaC_ErrorAbort (buffer, (int)(line), file);
	return 0;
}


/*
 *	Tried to allocate from some NULL zone. This produces
 *	a fatal error.
 */

void
OPMemZone::NullFreeZone (const LatinChar* zone, Count line, const LatinChar* file)
{
	OP_MEMBER (FreeAllocZone, 3);
	OP_ARG (1, const LatinChar*, zone);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	//	Concatenate "Free Null Zone : " and "zone_name" to produce the final fatal
	//	error message.
	
	const LatinChar* src = "Free Null Zone : ";
	LatinChar buffer[200];
	int index = 0;
	
	for (; (*src) && (index < 199); index++) buffer[index] = *src++;
	for (; (*src) && (index < 199); index++) buffer[index] = *zone++;
	buffer[index] = 0;
	
	OPaC_ErrorAbort (buffer, (int)(line), file);
}



/*
 *	Find out which zone this pointer belongs to and then
 *	simply free the chunk.
 *
 *	Throws :
 *
 *	  - OPExcMemory (NULL_PTR) if the pointer is NULL.
 *	
 *	  - OPExcMemory (BAD_PTR) if the pointer does not refer
 *		memory contained in this memory zone.
 *	
 *	  - OPExcMemory (FREE) if the pointer refers to memory
 *		which has already been freed.
 */

void
OPMemZone::FreeAnyZone (void* ptr, Count line, const LatinChar* file)
{
	OP_MEMBER (FreeAnyZone, 3);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	if (ptr == 0) {
		throw OPExcMemory (OPExcMemory::NULL_PTR, 0, line, file);
	}
	
	OPMemZone* zone = 0;
	
	if (OPMemZone::FindZone (ptr, zone)) {
		zone->Free (ptr, line, file);
	} else {
		throw OPExcMemory (OPExcMemory::BAD_PTR, ptr, line, file);
	}
}

void
OPMemZone::FreeVectorAnyZone (void* ptr, Count line, const LatinChar* file)
{
	OP_MEMBER (FreeVectorAnyZone, 3);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Count, line);
	OP_ARG (3, const LatinChar*, file);
	
	if (ptr == 0) {
		throw OPExcMemory (OPExcMemory::NULL_PTR, 0, line, file);
	}
	
	OPMemZone* zone = 0;
	
	if (OPMemZone::FindZone (ptr, zone)) {
		zone->FreeVector (ptr, line, file);
	} else {
		throw OPExcMemory (OPExcMemory::BAD_PTR, ptr, line, file);
	}
}



/*
 *	Initialise the memory zone handler. This will simply
 *	set up the public memory zones.
 */

void
OPMemZone::Initialise ()
{
	OP_MEMBER (Initialise, 0);
	
	OP_ASSERT (OPMemZone::zone_head == 0);
	
	//	The first memory zone cannot be allocated using the MemZone mechanisms,
	//	since there are no MemZones available at that point !
	
	static OPMemZone sys_zone;
	sys_zone.Initialise ("SysZone");
	OPaC_sys_zone   = & sys_zone;
	
	//	Now just initialise the miscellanous memory zones :
	
	OPaC_data_zone  = NEW_FROM_ZONE (OPaC_sys_zone) OPMemZone;
	OPaC_app_zone   = NEW_FROM_ZONE (OPaC_sys_zone) OPMemZone;
	OPaC_temp_zone  = NEW_FROM_ZONE (OPaC_sys_zone) OPMemZone;
	OPaC_iface_zone = NEW_FROM_ZONE (OPaC_sys_zone) OPMemZone;
	
	OPaC_data_zone->Initialise ("DataZone");
	OPaC_app_zone->Initialise ("ApplicationZone");
	OPaC_temp_zone->Initialise ("TempZone");
	OPaC_iface_zone->Initialise ("InterfaceZone");
}



/*
 *	Kill the memory zone handler. This will simply delete
 *	the public memory zones.
 */

void
OPMemZone::Kill ()
{
	OP_MEMBER (Kill, 0);
	
	OP_ASSERT (OPaC_sys_zone);
	
#ifdef _DEBUG
	OPObjectRef::g_tail = 0;
	OPObjectRef::g_head = 0;
#endif
	
	OPMemZone* iter = OPMemZone::zone_tail;
	OPMemZone* prev = 0;
	
	//	Remove the memory zones allocated by other classes, but which have not
	//	been deleted automatically.
	
	while (iter != OPaC_iface_zone) {
		
		prev = iter->prev;
		
		OP_ASSERT (prev);
		DELETE (iter);
		
		iter = prev;
	}
	
	//	Now delete our own zone.
	
	DELETE_FROM_ZONE (OPaC_sys_zone, OPaC_data_zone);	OPaC_data_zone  = 0;
	DELETE_FROM_ZONE (OPaC_sys_zone, OPaC_app_zone);	OPaC_app_zone   = 0;
	DELETE_FROM_ZONE (OPaC_sys_zone, OPaC_temp_zone);	OPaC_temp_zone  = 0;
	DELETE_FROM_ZONE (OPaC_sys_zone, OPaC_iface_zone);	OPaC_iface_zone = 0;
	
	OPaC_sys_zone->FreeAndDestroy ();
	OPaC_sys_zone = 0;
	
	//	And let the automatic destruction of Init's static variable sys_zone
	//	do the rest of the deallocation...
}



void
OPMemZone::KillRefFromZone ()
{
	OP_METHOD (KillRefFromZone, 0);
	
	BufferHeader* buffer = this->buf_head;
	
	while (buffer) {
		this->KillRefFromBuffer (buffer);
		buffer = buffer->next;
	}
}

void
OPMemZone::KillRefToZone ()
{
	OP_METHOD (KillRefFromZone, 0);
	
	BufferHeader* buffer = this->buf_head;
	
	while (buffer) {
		this->KillRefToBuffer (buffer);
		buffer = buffer->next;
	}
}


/*
 *	Remove any reference originating from the specified buffer. In
 *	order to find these references, we have to step through every
 *	memory zone/buffer/chunk and identify objects which might be
 *	referred to from the specified buffer.
 */

void
OPMemZone::KillRefFromBuffer (BufferHeader* from)
{
	OP_METHOD (KillRefFromBuffer, 1);
	OP_ARG (1, BufferHeader*, from);
	
	OPMemZone* zone = OPMemZone::zone_head;
	
	Address addr_lo = (Address)(from->buffer);
	Address addr_hi = (Address)(from->buffer) + from->size - 1;
	
	while (zone) {

		BufferHeader* buffer = zone->buf_head;
		while (buffer) {
			ChunkHeader* chunk = buffer->head;
			while (chunk) {
				if (chunk->mem_type == OP_MTYPE_OBJECT) {
					void* ptr = chunk->ptr;
#ifdef _DEBUG
					ptr = (void*)(((Address)(ptr)) + sizeof (void*));
#endif
					OPObject*    object = (OPObject*)(ptr);
					OPObjectRef* ref    = object->ref_head;
					
					//	Is the reference to this object stored in the specified
					//	buffer ? If so, reduce the count to the object and remove
					//	the link.
					
					while (ref) {
						
						Address      addr = (Address)(ref);
						OPObjectRef* next = ref->next;
						
						if ( (addr >= addr_lo)
						  && (addr <= addr_hi) ) {
							
							(ref->next ? ref->next->prev : object->ref_tail) = ref->prev;
							(ref->prev ? ref->prev->next : object->ref_head) = ref->next;
							
							ref->object = 0;
							ref->next   = 0;
							ref->prev   = 0;
							ref->c_name = 0;
#ifdef _DEBUG
							ref->g_next = 0;
							ref->g_prev = 0;
#endif
							
							object->ref_count--;
						}
						ref = next;
					}
				}
				chunk = chunk->next;
			}
			buffer = buffer->next;
		}
		zone = zone->next;
	}
}


/*
 *	Remove any reference pointing to the specified buffer. We
 *	just have to walk through each object stored in the buffer
 *	and reset the references to it.
 */

void
OPMemZone::KillRefToBuffer (BufferHeader* buffer)
{
	OP_METHOD (KillRefToBuffer, 1);
	OP_ARG (1, BufferHeader*, buffer);
	
	ChunkHeader* chunk = buffer->head;
	while (chunk) {
		if (chunk->mem_type == OP_MTYPE_OBJECT) {
			void* ptr = chunk->ptr;
#ifdef _DEBUG
			ptr = (void*)(((Address)(ptr)) + sizeof (void*));
#endif
			OPObject*    object = (OPObject*)(ptr);
			OPObjectRef* ref    = object->ref_head;
			
			//	Set each reference to this object to zero.
			
			while (ref) {
				
				OPObjectRef* next = ref->next;
				
				(ref->next ? ref->next->prev : object->ref_tail) = ref->prev;
				(ref->prev ? ref->prev->next : object->ref_head) = ref->next;
				
				ref->object = 0;
				ref->next   = 0;
				ref->prev   = 0;
				ref->c_name = 0;
#ifdef _DEBUG
				ref->g_next = 0;
				ref->g_prev = 0;
#endif
				object->ref_count--;
				ref = next;
			}
		}
		chunk = chunk->next;
	}
}

