/*
 *	objectprovider.cxx
 *
 *	The Object Provider is a simple implementation of a provider which
 *	returns icons for well-known objects.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/container.h>
#include <OPaC/objectpane.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPObjectProvider);

/****************************************************************/

OPObjectProvider::OPObjectProvider ()
{
	OP_METHOD (OPObjectProvider, 0);
}

OPObjectProvider::~OPObjectProvider ()
{
}

void
OPObjectProvider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPObjectProvider>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << OPStorageSync::end;
}

void
OPObjectProvider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::end;
}

void
OPObjectProvider::Initialise (OPGenericContainerIn container)
{
	this->container = container;
}


Bool
OPObjectProvider::GetContainer (OPGenericContainerOut container) const
{
	OP_METHOD (GetContainer, 1);
	OP_ARG (1, OPGenericContainerOut, container);
	
	container = this->container;
	
	return container.IsValid ();
}

Count
OPObjectProvider::ReturnNumObjects ()
{
	return this->container.IsInvalid () ? 0 : this->container->TotalCount ();
}


Bool
OPObjectProvider::GetObject (Count index, OPObjectOut object)
{
	OP_METHOD (GetObject, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	return this->container.IsValid ()
		&& this->container->Get (index, object);
}


Bool
OPObjectProvider::GetLargeIcon (Count index, OPLookIconOut icon)
{
	OPObjectRef object;
	
	if (this->GetObject (index, object)) {
		if (object.IsValid ()) {
			if (object->IsKindOf (OP_RUNTIME (OPApp))) {
				icon = OP_ICON_OBJ_APP;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPMenu))) {
				icon = OP_ICON_OBJ_MENU;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPWindow))) {
				icon = OP_ICON_OBJ_WINDOW;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPDataSlot))) {
				icon = OP_ICON_OBJ_SLOT;
				return TRUE;
			}
			icon = OP_ICON_CUBE_NORMAL;
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPObjectProvider::GetSmallIcon (Count index, OPLookIconOut icon)
{
	OPObjectRef object;
	
	if (this->GetObject (index, object)) {
	}
	
	return FALSE;
}


Bool
OPObjectProvider::GetText (Count index, OPStringOut text)
{
	OP_METHOD (GetText, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPStringOut, text);
	
	OPObjectRef object;
	
	if (this->GetObject (index, object)) {
		if (object.IsValid ()) {
			
			static OPAtomRef user_friendly_name = "UserFriendlyName";
			OPTextRef name;
			
			if ( (object->GetAttribute (user_friendly_name, name))
			  && (name->IsKindOf (OP_RUNTIME (OPText))) ) {
				text = name->ReturnString ();
				return TRUE;
			}
			
			if (object->IsKindOf (OP_RUNTIME (OPApp))) {
				text = "Application";
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPMenu))) {
				text = "Menu";
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPWindow))) {
				text = "Window";
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPDataSlot))) {
				text = "DataSlot";
				return TRUE;
			}
			
			text = "Object";
			return TRUE;
		}
	}
	
	return FALSE;
}


Bool
OPObjectProvider::GetMarkIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetMarkIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	return FALSE;
}

