/*
 *	shadeview.cxx
 *
 *	The shade view draws a rectangular shade behind a specified
 *	view.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/dragview.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/shadeview.h>

#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPShadeView);

/****************************************************************/

OP_PUBLISH_SLOT (OPShadeView, color_ext, ReturnColorExt, SetColorExt, "External color", OPColorPixel);
OP_PUBLISH_SLOT (OPShadeView, color_int, ReturnColorInt, SetColorInt, "Internal color", OPColorPixel);
OP_PUBLISH_SLOT (OPShadeView, dx_offset, ReturnDxOffset, SetDxOffset, "Horizontal offset", Int32);
OP_PUBLISH_SLOT (OPShadeView, dy_offset, ReturnDyOffset, SetDyOffset, "Vertical offset", Int32);
OP_PUBLISH_SLOT (OPShadeView, grow, ReturnGrow, SetGrow, "Relative growing", Int32);
OP_PUBLISH_SLOT (OPShadeView, transition, ReturnTransition, SetTransition, "Transition steps", Card32);

/****************************************************************/

#define	OP_IMPLEMENT_SLOT(C,T,n,Set,Return)		\
												\
Bool											\
C::Set (T ## In n)								\
{												\
	OP_METHOD (Set, 1);							\
	OP_ARG (1, T ## In, n);						\
	this->n = n;								\
	return TRUE;								\
}												\
												\
T												\
C::Return () const								\
{												\
	OP_METHOD (Return, 0);						\
	return this->n;								\
}

OP_IMPLEMENT_SLOT (OPShadeView, OPColorPixel, color_ext, SetColorExt, ReturnColorExt)
OP_IMPLEMENT_SLOT (OPShadeView, OPColorPixel, color_int, SetColorInt, ReturnColorInt)
OP_IMPLEMENT_SLOT (OPShadeView, Int32, dx_offset, SetDxOffset, ReturnDxOffset)
OP_IMPLEMENT_SLOT (OPShadeView, Int32, dy_offset, SetDyOffset, ReturnDyOffset)
OP_IMPLEMENT_SLOT (OPShadeView, Int32, grow, SetGrow, ReturnGrow)
OP_IMPLEMENT_SLOT (OPShadeView, Card32, transition, SetTransition, ReturnTransition)

/****************************************************************/

OPShadeView::OPShadeView ()
{
	this->color_ext  = OPColorPixel (0.0, 0.0, 0.0, 1.0);
	this->color_int  = OPColorPixel (0.0, 0.0, 0.0, 0.5);
	this->dx_offset  = 0;
	this->dy_offset  = 0;
	this->grow       = 0;
	this->transition = 4;
}

OPShadeView::~OPShadeView ()
{
}

void
OPShadeView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPShadeView view color_ext color_int dx_offset dy_offset"
							   " grow transition>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->view;
	storage << this->color_ext << this->color_int;
	storage << this->dx_offset << this->dy_offset;
	storage << this->grow << this->transition;
	storage << OPStorageSync::end;
}

void
OPShadeView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->view;
	storage >> this->color_ext >> this->color_int;
	storage >> this->dx_offset >> this->dy_offset;
	storage >> this->grow >> this->transition;
	storage >> OPStorageSync::end;
}

void
OPShadeView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
	
	this->view = OPViewNull;
}

Bool
OPShadeView::SetView (OPViewIn view)
{
	OP_METHOD (SetView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->view = view;
	
	return TRUE;
}

const OPViewRef&
OPShadeView::ReturnView () const
{
	OP_METHOD (ReturnView, 0);
	return this->view;
}

Bool
OPShadeView::SetOffset (Int32 dx, Int32 dy)
{
	OP_METHOD (SetOffset, 2);
	OP_ARG (1, Int32, dx);
	OP_ARG (2, Int32, dy);
	
	this->dx_offset = dx;
	this->dy_offset = dy;
	
	return TRUE;
}

static OPColorPixel
mix_colors (OPColorPixelIn a, OPColorPixelIn b, Real ratio)
{
	Real ra = 1.0 - ratio;
	Real rb = ratio;
	return OPColorPixel (a.RealRed () * ra + b.RealRed () * rb,
						 a.RealGreen () * ra + b.RealGreen () * rb,
						 a.RealBlue () * ra + b.RealBlue () * rb,
						 a.RealTransparency () * ra + b.RealTransparency () * rb);
}

void
OPShadeView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	if (what & OP_DRAW_BACKGROUND) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		const OPLookManRef& lookman = OPLookMan::Current ();
		
		OPRect rect = this->view.IsValid () ? this->view->ReturnBox () : this->ReturnBox ();
		
		rect.MoveOrigin (this->dx_offset - this->grow, this->dy_offset - this->grow);
		rect.ChangeSize (this->grow, this->grow);
		
		for (Real step = 0.0; step < this->transition; step++) {
			port->SetForeColor (mix_colors (this->color_ext, this->color_int, step / this->transition));
			port->DrawRect (rect, step);
		}
		
		port->SetForeColor (this->color_int);
		port->FillRect (rect, this->transition);
		
		what &= ~OP_DRAW_BACKGROUND;
	}
	
	super::Draw (port, rect, what);
}
