/*
 *	springband.cxx
 *
 *	A spring band is a collection of views connected by springs.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/new.h>
#include <OPaC/spring.h>
#include <OPaC/springboard.h>
#include <OPaC/view.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPSpringBand);

/****************************************************************/


/*
 *	Create an empty spring board.
 */

OPSpringBand::OPSpringBand ()
{
	OP_METHOD (OPSpringBand, 0);

	this->is_vertical   = FALSE;
	this->is_meta_ready = FALSE;

	this->meta_spring = OP_NEW (OPSpring);
	this->view        = OP_NEW (OPArray);
}


/*
 *	Destroy.
 */

OPSpringBand::~OPSpringBand ()
{
	OP_METHOD (~OPSpringBand, 0);
}

void
OPSpringBand::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	static OPStorageSync info ("<OPSpringBand is_vertical view spring>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->is_vertical;
	storage << this->view;
	storage << this->spring;
	storage << OPStorageSync::end;
}

/*
 *	
 */

void
OPSpringBand::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->is_vertical;
	storage >> this->view;
	storage >> this->spring;
	storage >> OPStorageSync::end;
	
	this->is_meta_ready = FALSE;
}

/*
 *	
 */

void
OPSpringBand::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	this->view->SetSize (0);
	this->spring = 0;
	
	this->is_meta_ready = FALSE;
}

/*
 *	
 */

void
OPSpringBand::Delete ()
{
	OP_METHOD (Delete, 0);

	super::Delete ();
	this->RemoveAll ();
}

/*
 *	
 */

void
OPSpringBand::BuildMetaSpring () const
{
	OP_METHOD (BuildMetaSpring, 0);

	Count index = 0;
	Count num   = this->spring->TotalCount ();
	
	if (num == 0) {
		return;
	}
	
	this->meta_spring->Reset (0);

	do {
		this->meta_spring->Merge (this->spring[index]);
	} while (++index < num);
	
	this->is_meta_ready = TRUE;
}

/* 
 *	Return the meta spring which models the band's overall
 *	stretching behaviour.
 */

const OPSpringRef&
OPSpringBand::ReturnMetaSpring () const
{
	OP_METHOD (ReturnMetaSpring, 0);
	
	if (this->is_meta_ready == FALSE) {
		this->BuildMetaSpring ();
	}
	
	return this->meta_spring;
}


/*
 *	
 */

void
OPSpringBand::SetVertical (Bool yes)
{
	OP_METHOD (SetVertical, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_vertical = yes;
}

/*
 *	
 */

Bool
OPSpringBand::IsVertical () const
{
	OP_METHOD (IsVertical, 0);

	return this->is_vertical;
}


void
OPSpringBand::SpringChanged ()
{
	OP_METHOD (SpringChanged, 0);
	this->is_meta_ready = FALSE;
}


/*
 *	
 */

void
OPSpringBand::Resize (OPRectIn rect)
{
	OP_METHOD (Resize, 1);
	OP_ARG (1, OPRectIn, rect);
	
	Count index = 0;
	Count num   = this->view->TotalCount ();
	
	if (num == 0) {
		return;
	}
	
	if (this->is_meta_ready == FALSE) {
		this->BuildMetaSpring ();
	}
	
	OPCoord size    = this->is_vertical ? rect.DY () : rect.DX ();
	OPCoord start   = this->is_vertical ? rect.Y1 () : rect.X1 ();
	OPCoord delta   = 0;
	Real    stretch = this->meta_spring->FindStretch (size);

	if (this->is_vertical) {
		
		OPRect    box;
		OPViewRef the_view;
		
		do {
			the_view = this->view[index];
			box      = the_view->ReturnBox ();
			
			start += delta;
			start += this->spring[index*2+0]->ComputeDefaultSize (stretch);
			delta  = this->spring[index*2+1]->ComputeDefaultSize (stretch);
			
			box.Y1 () = start;
			box.Y2 () = start + delta;

			this->view[index]->SetBox (box);

		} while (++index < num);

	} else {
		
		OPRect box;
		
		do {
			box = this->view[index]->ReturnBox ();
			
			start += delta;
			start += this->spring[index*2+0]->ComputeDefaultSize (stretch);
			delta  = this->spring[index*2+1]->ComputeDefaultSize (stretch);
			
			box.X1 () = start;
			box.X2 () = start + delta;

			this->view[index]->SetBox (box);

		} while (++index < num);
	}
}

/*
 *	
 */

Count
OPSpringBand::CountViews () const
{
	OP_METHOD (CountViews, 0);
	
	return this->view->TotalCount ();
}

/*
 *	
 */

Count
OPSpringBand::CountSprings () const
{
	OP_METHOD (CountSprings, 0);
	
	return this->spring->TotalCount ();
}

/*
 *	
 */

Bool
OPSpringBand::InsertView (Count index, OPViewIn view)
{
	OP_METHOD (InsertView, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPViewIn, view);

	if (view.IsValid () == FALSE) {
		return FALSE;
	}

	if (this->view->TotalCount () == 0) {
		this->spring = OP_NEW (OPArray);
		this->spring->SetSize (1);
		this->spring[0] = OP_NEW (OPSpring);
	}
	
	this->view->Insert (index);
	this->spring->Insert (index*2+1, 2);
	
	this->view[index]       = view;
	this->spring[index*2+1] = OP_NEW (OPSpring);
	this->spring[index*2+2] = OP_NEW (OPSpring);
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	view->GetSize (dx, dy);
	
	OPCoord delta_1 = this->spring[index*2+0]->FindSize (OP_SPRING_ONE);
	OPCoord delta_2 = this->is_vertical ? dy : dx;
	
	//	Adjust the different spring sizes so that the view fits into the
	//	provided space. If the view is too large to fit into the gap, don't
	//	shrink it; the space around it will then be set to 10% of the view's
	//	size.
	
	if (delta_2 < delta_1) {
		delta_1 -= delta_2;
	} else {
		delta_1  = delta_2 / 10;
	}
	
	this->spring[index*2+0]->Reset (delta_1 / 2);
	this->spring[index*2+1]->Reset (delta_2);
	this->spring[index*2+2]->Reset (delta_1 / 2);
	
	this->is_meta_ready = FALSE;
	
	return TRUE;
}


/*
 *	Remove the specified view, if it exists in the band. If not,
 *	simply return FALSE.
 */

Bool
OPSpringBand::RemoveView (OPViewIn view)
{
	OP_METHOD (RemoveView, 1);
	OP_ARG (1, OPViewIn, view);
	
	Count index = 0;
	
	if (this->FindView (view, index)) {
		return this->RemoveView (index);
	}
	
	return FALSE;
}


/*
 *	
 */

Bool
OPSpringBand::RemoveView (Count index)
{
	OP_METHOD (RemoveView, 1);
	OP_ARG (1, Count, index);

	if (index >= this->view->TotalCount ()) {
		return FALSE;
	}
	
	//	Adjust the size of the remaining spring, so that it will have
	//	exactly the same width as the previous three.
	
	OPCoord delta_1 = this->spring[index*2+0]->FindSize (OP_SPRING_ONE);
	OPCoord delta_2 = this->spring[index*2+1]->FindSize (OP_SPRING_ONE);
	OPCoord delta_3 = this->spring[index*2+2]->FindSize (OP_SPRING_ONE);
	
	this->spring[index*2+0]->Reset (delta_1 + delta_2 + delta_3);

	this->view->Remove (index);
	this->spring->Remove (index*2+1, 2);
	
	if (this->view->TotalCount () == 0) {
		this->spring->SetSize (0);
	}
	
	this->is_meta_ready = FALSE;
	
	return TRUE;
}


/*
 *	Find the associated window (if any). This can be inferred
 *	from the inserted views, which all belong to the same window.
 */

Bool
OPSpringBand::GetAssociatedWindow (OPWindowOut window) const
{
	OP_METHOD (GetAssociatedWindow, 1);
	OP_ARG (1, OPWindowOut, window);
	
	OPViewRef view;
	
	if (this->GetNthView (0, view)) {
		if (view.IsValid ()) {
			window = view->ReturnWindow ();
			return window.IsValid ();
		}
	}
	
	return FALSE;
}


Bool
OPSpringBand::ResetFromModel ()
{
	OP_METHOD (ResetFromModel, 0);
	
	Count max = this->CountViews ();
	if (max == 0) return FALSE;
	
	OPViewRef parent = this->view[0]->ReturnParent ();
	
	OPCoord ox, oy;
	OPCoord dx, dy;
	
	parent->GetOrigin (ox, oy);
	parent->GetSize (dx, dy);
	
	if (parent->IsKindOf (OP_RUNTIME (OPWindow))) {
		ox = 0;
		oy = 0;
	}
	
	OPCoord lox = ox;
	OPCoord loy = oy;
	
	for (Count index = 0; index < max; index++) {
		
		OPCoord vox, voy;
		OPCoord vdx, vdy;
		OPCoord mdx, mdy;
		
		this->view[index]->GetOrigin (vox, voy);
		this->view[index]->GetSize (vdx, vdy);
		this->view[index]->GetMinSize (mdx, mdy);
		
		OPCoord min = 0.0;
		OPCoord len = (this->is_vertical) ? voy - loy : vox - lox;
		
		if (this->spring[index*2+0]->FindSlope (OP_SPRING_ONE) == 0.0) {
			
			//	Not stretchable spring.
			
			this->spring[index*2+0]->Reset (len, 0.0);
			
		} else {
			
			//	Stretchable spring.
			
			this->spring[index*2+0]->Reset (min, len-min);
		}
		
		min = (this->is_vertical) ? mdy : mdx;
		len = (this->is_vertical) ? vdy : vdx;
		
		if (this->spring[index*2+1]->FindSlope (OP_SPRING_ONE) == 0.0) {
			
			//	Not stretchable spring.
			
			this->spring[index*2+1]->Reset (len, 0.0);
			
		} else {
			
			//	Stretchable spring.
			
			this->spring[index*2+1]->Reset (min, len-min);
		}
		
		lox = vox + vdx - 1;
		loy = voy + vdy - 1;
		
		if (index != max-1) continue;
		
		min = 0.0;
		len = (this->is_vertical) ? oy+dy-loy : ox+dx-lox;
		
		if (this->spring[index*2+2]->FindSlope (OP_SPRING_ONE) == 0.0) {
			
			//	Not stretchable spring.
			
			this->spring[index*2+2]->Reset (len, 0.0);
			
		} else {
			
			//	Stretchable spring.
			
			this->spring[index*2+2]->Reset (min, len-min);
		}
	}
	
	this->is_meta_ready = FALSE;
	
	return TRUE;
}

/*
 *	
 */

Bool
OPSpringBand::GetNthView (Count index, OPViewOut view) const
{
	OP_METHOD (GetNthView, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPViewOut, spring);

	if (index >= this->view->TotalCount ()) {
		view = 0;
		return FALSE;
	}
	
	view = this->view[index];
	return TRUE;
}

/*
 *	
 */

Bool
OPSpringBand::GetNthSpring (Count index, OPSpringOut spring) const
{
	OP_METHOD (GetNthSpring, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringOut, spring);

	if (index >= this->spring->TotalCount ()) {
		spring = 0;
		return FALSE;
	}
	
	spring = this->spring[index];
	return TRUE;
}



/*
 *	Find if the view is contained in the band. Return TRUE
 *	and the index if it is the case.
 */

Bool
OPSpringBand::FindView (OPViewIn view) const
{
	OP_METHOD (FindView, 1);
	OP_ARG (1, OPViewIn, view);
	
	Count index = 0;
	return this->FindView (view, index);
}


Bool
OPSpringBand::FindView (OPViewIn view, CountOut index) const
{
	OP_METHOD (FindView, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, CountOut, index);

	Count max = this->view->TotalCount ();
	index = 0;

	while (index < max) {
		if (this->view[index] == view) {
			return TRUE;
		}
		index++;
	}

	index = 0;
	return FALSE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringBeforeView (Count index, OPSpringOut spring) const
{
	OP_METHOD (FindSpringBeforeView, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringOut, spring);

	if (index >= this->view->TotalCount ()) {
		spring = 0;
		return FALSE;
	}

	spring = this->spring[index*2+0];
	return TRUE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringOfView (Count index, OPSpringOut spring) const
{
	OP_METHOD (FindSpringOfView, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringOut, spring);

	if (index >= this->view->TotalCount ()) {
		spring = 0;
		return FALSE;
	}

	spring = this->spring[index*2+1];
	return TRUE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringAfterView (Count index, OPSpringOut spring) const
{
	OP_METHOD (FindSpringAfterView, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPSpringOut, spring);

	if (index >= this->view->TotalCount ()) {
		spring = 0;
		return FALSE;
	}

	spring = this->spring[index*2+2];
	return TRUE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringBeforeView (OPViewIn view, OPSpringOut spring) const
{
	OP_METHOD (FindSpringBeforeView, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, OPSpringOut, spring);

	Count index = 0;

	return (this->FindView (view, index))
		? this->FindSpringBeforeView (index, spring)
		: FALSE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringOfView (OPViewIn view, OPSpringOut spring) const
{
	OP_METHOD (FindSpringOfView, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, OPSpringOut, spring);

	Count index = 0;

	return (this->FindView (view, index))
		? this->FindSpringOfView (index, spring)
		: FALSE;
}

/*
 *	
 */

Bool
OPSpringBand::FindSpringAfterView (OPViewIn view, OPSpringOut spring) const
{
	OP_METHOD (FindSpringAfterView, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, OPSpringOut, spring);

	Count index = 0;

	return (this->FindView (view, index))
		? this->FindSpringAfterView (index, spring)
		: FALSE;
}



Bool
OPSpringBand::ReplaceView (OPViewIn old_view, OPViewIn new_view)
{
	OP_METHOD (ReplaceView, 2);
	OP_ARG (1, OPViewIn, old_view);
	OP_ARG (2, OPViewIn, new_view);
	
	Count total = this->view->TotalCount ();
	Count index = 0;

	while (index < total) {
		if (this->view[index] == old_view) {
			this->view[index] = new_view;
		}
		index++;
	}

	return TRUE;
}

