/*
 *	target.cxx
 *	
 *	The target definition class is very useful in building dynamic message
 *	forwarding systems, since the receiving member functions may be linked
 *	at run-time.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>
#include <OPaC/opacirt.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPTarget);

OP_IMPLEMENT_TYPEDEF_VAL (Card32, OPTargetType, OP_TYPE_CARD, Card32);

/****************************************************************/


/*
 *	Creates an empty target.
 */
	
OPTarget::OPTarget ()
{
	OP_METHOD (OPTarget, 0);
	
	this->name   = OPAtomNull;
	this->object = (OPObject*)(0);
	this->target = 0;
	this->type   = OP_TARGET_IS_INVALID;
}


OPTarget::~OPTarget ()
{
	OP_METHOD (~OPTarget, 0);
}



/*
 *	Store the target definition.
 */

void
OPTarget::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPTarget name object type>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->name << this->object << this->type;
	storage << OPStorageSync::end;
}



/*
 *	Restore the target definition.
 */

void
OPTarget::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);

	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->name >> this->object >> this->type;
	storage >> OPStorageSync::end;
	
	this->Attach ();
}



/*
 *	Attach to the real target. This will translate the name/object
 *	pair into some OPaC_Target pointer, if possible.
 */

Bool
OPTarget::Attach ()
{
	OP_METHOD (Attach, 0);
	
	this->target = 0;

	if (this->object.IsValid ()) {
		
		const OPRuntimeRef& rt = this->object->ReturnRuntime ();
		
		switch (this->type) {
			case OP_TARGET_IS_OBSERVER:
				this->target = rt->FindObserver (this->name);
				break;
			
			case OP_TARGET_IS_METHOD:
				this->target = rt->FindMethod (this->name);
				break;
			
			case OP_TARGET_IS_ACTOR:
				this->target = rt->FindActor (this->name);
				break;
			
			default:
				this->target = 0;
				break;
		}
	}
	
	return (this->target) ? TRUE : FALSE;
}



/*
 *	Prepare the target for the deletion.
 */	

void
OPTarget::Delete ()
{
	OP_METHOD (Delete, 0);

	this->name   = OPAtomNull;
	this->object = (OPObject*)(0);
	
	super::Delete ();
}



/*
 *	Attach to the real target. This will translate the name/object
 *	pair into some OPaC_Target pointer, if possible.
 */

Bool
OPTarget::AttachToObject (OPObjectIn object)
{
	OP_METHOD (AttachToObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	//	If the new object is the same as the previous one, nothing has to
	//	be changed...
	
	if (this->object == object) {
		return (this->target) ? TRUE : FALSE;
	}
	
	//	Attaching to an invalid object is the same as detaching it.

	if (object.IsValid () == FALSE) {
		this->object = object;
		this->target = 0;
		return FALSE;
	}
	
	//	Attaching to the same runtime class as before won't need any new
	//	research...

	if ( (this->object.IsValid ())
	  && (this->object->ReturnRuntime () == object->ReturnRuntime ()) ) {
		return (this->target) ? TRUE : FALSE;
	}

	//	Otherwise, just try to attach again.

	this->object = object;
	
	return this->Attach ();
}


	
/*
 *	Setup a target and return TRUE if it has been successfully
 *	attached to.
 */

Bool
OPTarget::SetTarget (OPAtomIn name, OPObjectIn object, OPTargetTypeIn type)
{
	OP_METHOD (SetTarget, 3);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectIn, object);
	OP_ARG (3, OPTargetTypeIn, type);
	
	this->name   = name;
	this->object = object;
	this->type   = type;
	
	return this->Attach ();
}



/*
 *	Return a full target description.
 */

void
OPTarget::GetTarget (OPAtomOut name, OPObjectOut object, OPTargetTypeOut type) const
{
	OP_METHOD (GetTarget, 3);
	OP_ARG (1, OPAtomOut, name);
	OP_ARG (2, OPObjectOut, object);
	OP_ARG (3, OPTargetTypeOut, type);
	
	name   = this->name;
	object = this->object;
	type   = this->type;
}



/*
 *	Return the target's name, object and type. This is just a
 *	conveniance call, since GetTarget returns all the three.
 */

const OPAtomRef&
OPTarget::GetName () const
{
	OP_METHOD (GetName, 0);
	return this->name;
}

const OPObjectRef&
OPTarget::GetObject () const
{
	OP_METHOD (GetObject, 0);
	return this->object;
}

OPTargetType
OPTarget::GetType () const
{
	OP_METHOD (GetType, 0);
	return this->type;
}



/*
 *	Return TRUE if the current target is valid, that is, it could
 *	be used to send messages to.
 */

Bool
OPTarget::IsValid () const
{
	OP_METHOD (IsValid, 0);
	
	return (this->target) ? TRUE : FALSE;
}



/*
 *	Return TRUE if the target could report the specified data.
 *	If the target type isn't OP_TARGET_IS_OBSERVER, this will
 *	return FALSE.
 */

Bool
OPTarget::DoReport (OPObjectIn object, Card32 value) const
{
	OP_METHOD (DoReport, 2);
	OP_ARG (1, OPObjectIn, object);

	if ( (this->target)
	  && (this->type == OP_TARGET_IS_OBSERVER) ) {
		OPObject* ptr = this->object.CheckAndResolve ();
		(this->target->target_method->o_trampo) (ptr, object, value);
		return TRUE;
	}

	return FALSE;
}



/*
 *	Return TRUE if the target could execute its action. If
 *	the target type isn't OP_TARGET_IS_ACTOR, this will
 *	return FALSE.
 */

Bool
OPTarget::DoAction () const
{
	OP_METHOD (DoAction, 0);

	if ( (this->target)
	  && (this->type == OP_TARGET_IS_ACTOR) ) {
		OPObject* ptr = this->object.CheckAndResolve ();
		(this->target->target_method->a_trampo) (ptr);
		return TRUE;
	}

	return FALSE;
}



/*
 *	Return TRUE if the target could handle the message. If
 *	the target type isn't OP_TARGET_IS_METHOD, this will
 *	return FALSE.
 */

Bool
OPTarget::DoMessage (OPMessageInOut message) const
{
	OP_METHOD (DoMessage, 1);
	OP_ARG (1, OPMessageInOut, message);

	if ( (this->target)
	  && (this->type == OP_TARGET_IS_METHOD) ) {
		OPObject* ptr = this->object.CheckAndResolve ();
		(this->target->target_method->m_trampo) (message, ptr, 0);
		return TRUE;
	}

	return FALSE;
}

