/*
 *	tooltip.cxx
 *
 *	The Tool Tip class is some kind of very simple help balloon.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/texttag.h>
#include <OPaC/tooltip.h>

#include <OPaC/targetmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPToolTip);

/****************************************************************/

OP_TARGET_ACTOR (OPToolTip, Dismiss);

/****************************************************************/

/*
 *	Construct the tool tip.
 */

OPToolTip::OPToolTip ()
{
	OP_METHOD (OPToolTip, 0);
	
	this->style = OP_WIN_NAKED | OP_WIN_TOP_LEVEL | OP_WIN_FLOAT_TOOL;
}



/*
 *	Destroy the tool tip.
 */

OPToolTip::~OPToolTip ()
{
	OP_METHOD (~OPToolTip, 0);
}



/*
 *	Save the tool tip to the storage.
 */

void
OPToolTip::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPToolTip tip>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->tip;
	storage << OPStorageSync::end;
}



/*
 *	Restore the tool tip from the storage.
 */

void
OPToolTip::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->tip;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the tool tip for destruction. Remove the
 *	tip and the rest...
 */

void
OPToolTip::Delete ()
{
	OP_METHOD (Delete, 0);

	this->Close ();
	
	if (this->tip.IsValid ()) {
		this->tip->Delete ();
		this->tip = 0;
	}
	
	super::Delete ();
}


/*
 *	Dismiss the tool tip. This will close the window
 *	and delete it. It will no longer be needed.
 */

void
OPToolTip::Dismiss ()
{
	OP_METHOD (Dismiss, 0);
	
	OPToolTipRef self = this;
	self->Delete ();
}


/*
 *	Define the title of the tool tip (this will be
 *	visible as its contents).
 */

Bool
OPToolTip::SetTitle (OPStringIn name)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, name);
	
	return this->SetTitle (name, 0, 0, FALSE);
}
	

/*
 *	Same as above; if show is set to TRUE, make the tool
 *	tip visible at the specified co-ordinates.
 */

Bool
OPToolTip::SetTitle (OPStringIn name, OPCoord x, OPCoord y, Bool show)
{
	OP_METHOD (SetTitle, 4);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, OPCoord, x);
	OP_ARG (3, OPCoord, y);
	OP_ARG (4, Bool, show);
	
	super::SetTitle (name);
	
	if (this->tip.IsValid ()) {
		this->tip->Delete ();
	}
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	this->tip = OP_NEW (OPTextTag);
	this->tip->Initialise (0, 0, this->box.DX (), this->box.DY (), name, this);
	this->tip->GetMinSize (dx, dy);
	
	dx += 4;
	dy += 2;
	
	Count  display = 0;
	OPRect box;
	
	//	Make sure that the tool tip window is completely visible.
	
	OPGrailWindow::GetDisplayFromPoint (x, y, display);
	OPGrailWindow::GetDisplayBox (box, display);
	
	if ((x + dx) >= box.X2 ()) x = box.X2 () - dx - 1;
	if ((y + dy) >= box.Y2 ()) y = box.Y2 () - dy - 1;
	
	this->tip->SetSize (dx, dy);
	this->SetOrigin (x, y);
	this->SetSize (dx, dy);
	
	if (show) {
		this->ShowNormal ();
	}
	
	return TRUE;
}



/*
 *	Draw the tool tip.
 */

void
OPToolTip::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp op)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, op);
	
	if (port.IsValid ()) {
		
		OPRect box = this->box;
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		box.SetOrigin (0, 0);
		
		if (op & OP_DRAW_BACKGROUND) {
			adorner->FillRect (box, OP_COLOR_TOOLTIP_BACK);
		}
		if (op & OP_DRAW_OBJECT) {
			adorner->DrawRect (box, OP_COLOR_TOOLTIP_FRAME);
			if (this->tip.IsValid ()) {
				this->tip->SetColor (OP_COLOR_TOOLTIP_TEXT);
			}
		}
		
		super::Draw (port, rect, op & ~OP_DRAW_BACKGROUND);
	}
}

