/*
 *	window.cxx
 *
 *	Base OPaC Window class implementation. This is the higher level definition
 *	of a window and uses the Grail Window class for any drawings.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/buildmisc.h>
#include <OPaC/except.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/selview.h>
#include <OPaC/window.h>
#include <OPaC/grail.h>
#include <OPaC/event.h>

#include <OPaC/targetmacro.h>
#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPWindow);

/****************************************************************/

OPAtomRef OPWindow::atom_title;

static const OPaC_BitsetDataDef local_style_info_1[] = {
	{ OP_WIN_MODE_MASK,		"Window mode",			OP_BITSET_DATA_MASK },
	{ OP_WIN_NAKED,			"naked, no frame",		OP_BITSET_DATA_VALUE },
	{ OP_WIN_NORMAL,		"normal, no title",		OP_BITSET_DATA_VALUE },
	{ OP_WIN_TITLE,			"normal, with title",	OP_BITSET_DATA_VALUE }
};

static const OPaC_BitsetDataDef local_style_info_2[] = {
	{ OP_WIN_ATTR_MASK,		"Window attributes",	OP_BITSET_DATA_MASK },
	{ OP_WIN_MOVE_OK,		"move",					OP_BITSET_DATA_BIT },
	{ OP_WIN_SIZE_OK,		"resize",				OP_BITSET_DATA_BIT },
	{ OP_WIN_REDUCE_OK,		"reduce/minimise",		OP_BITSET_DATA_BIT },
	{ OP_WIN_CLOSE_OK,		"close",				OP_BITSET_DATA_BIT }
};

static const OPaC_BitsetDataDef local_style_info_3[] = {
	{ OP_WIN_EXTEND_MASK,	"Extended attributes",	OP_BITSET_DATA_MASK },
	{ OP_WIN_TRANSPARENT,	"transparent",			OP_BITSET_DATA_BIT },
	{ OP_WIN_FLOAT_TOOL,	"floating tool window",	OP_BITSET_DATA_BIT },
	{ OP_WIN_TOP_LEVEL,		"always on top",		OP_BITSET_DATA_BIT },
	{ OP_WIN_PARTITION,		"partition window",		OP_BITSET_DATA_BIT }
};

/****************************************************************/

OP_PUBLISH_SLOT (OPWindow, title, ReturnTitle, SetTitle, "Window title", OPString);
OP_PUBLISH_SLOT (OPWindow, reduced_view, ReturnReducedView, SetReducedView, "Reduced view", OPView);
OP_PUBLISH_SLOT (OPWindow, style, ReturnStyle, SetStyle, "Window style", OPWindowStyle);

OP_PUBLISH_BITSET_INFO (OPWindow, style, local_style_info_1);
OP_PUBLISH_BITSET_INFO (OPWindow, style, local_style_info_2);
OP_PUBLISH_BITSET_INFO (OPWindow, style, local_style_info_3);

/****************************************************************/

OP_TARGET_ACTOR (OPWindow, ShowNormal);
OP_TARGET_ACTOR (OPWindow, ShowReduced);
OP_TARGET_ACTOR (OPWindow, Hide);
OP_TARGET_ACTOR (OPWindow, Close);
OP_TARGET_ACTOR (OPWindow, Kill);
OP_TARGET_ACTOR (OPWindow, UpdatePartitions);

OP_NOTIFY_EVENT (OPWindow, WindowOpened);
OP_NOTIFY_EVENT (OPWindow, WindowClosed);
OP_NOTIFY_EVENT (OPWindow, WindowActivated);
OP_NOTIFY_EVENT (OPWindow, WindowDeactivated);
OP_NOTIFY_EVENT (OPWindow, WindowShown);
OP_NOTIFY_EVENT (OPWindow, WindowHidden);
OP_NOTIFY_EVENT (OPWindow, WindowReduced);
OP_NOTIFY_EVENT (OPWindow, WindowMaximised);

/****************************************************************/


/*
 *	Initialise a window to the zero state. It occupies a
 *	minimum of memory.
 */

OPWindow::OPWindow ()
{
	OP_METHOD (OPWindow, 0);
	
	this->grail = 0;
	
	this->is_open    = FALSE;
	this->is_visible = FALSE;
	this->is_reduced = FALSE;
	this->has_updated_partitions = FALSE;
	
	this->style   = OP_WIN_NAKED;				//	window without adornments
	this->title   = "";							//	no title
	this->reduced = 0;							//	no iconized view
	
	this->default_active = 0;					//	no default active view
	this->current_active = 0;					//	no current active view
	this->current_active_ok = TRUE;
	
	this->view_collection    = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPHashTable);
	this->view_overlays      = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPHashTable);
	this->key_alt_collection = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPHashTable);
	
	this->cursor_view       = 0;				//	no view containing cursor
	this->drag_initial_view = 0;				//	no view where dragging began
	
	//	Register the window into the application's window data base. The window
	//	will be its own window.
	
	this->window = this;
	this->app    = OPApp::CurrentApp ();
	
	OPWindow::atom_title = "Title";
	this->app->InsertWindow (this->window);
}



/*
 *	Destroy a window.
 */

OPWindow::~OPWindow ()
{
	OP_METHOD (~OPWindow, 0);
}



/*
 *	Store the window's definition, but not its buffer.
 *	The information is sufficient to restore it later
 *	on.
 */

void
OPWindow::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	//	We do not store the view collection, since it will be automatically
	//	rebuilt by the window's Awake methods.
	
	static OPStorageSync info ("<OPWindow style title reduced>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->style << this->title << this->reduced;
	storage << OPStorageSync::end;
}



/*
 *	Restore the window's definition and open & show it if
 *	needed.
 */

void
OPWindow::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	//	Restore the saved information. This will be used to set up the
	//	window in its original state, using its methods so this can be
	//	easily overridden.
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->style >> this->title >> this->reduced;
	storage >> OPStorageSync::end;
}



/*
 *	This casting function is needed by the OPaC extension
 *	class OPxTitleWidget.
 */

OPView*
OPWindow::ReturnObjectAsView ()
{
	return this;
}


/*
 *	Called whenever the window's geometry changes. This
 *	should/will resize/move the window...
 */

void
OPWindow::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	if (this->grail.IsValid ()) {
		this->grail->SetOrigin (box.X1 (), box.Y1 ());
		this->grail->SetSize (box.DX (), box.DY ());
	}
	
	super::ChangeGeometry (box);
	
	this->has_updated_partitions = FALSE;
	
	if (this->grail.IsValid ()) {
		this->Display ();
		this->grail->Flush ();
	}
}



/*
 *	Awake a window. Do the hard job...
 */

void
OPWindow::AwakeMyself ()
{
	OP_METHOD (AwakeMyself, 0);
	
	//	Tell every view and subview who their parent window is. This cannot be
	//	done sooner, since the objects are only guaranteed to be loaded at this
	//	point.
	
	OPWindowRef self = this;
	this->SetWindow (self);
	
	//	Now restore the GRAIL window and open/reduce it if needed.
	
	if (this->is_open) {
		
		this->grail = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPGrailWindow);
		this->grail->Initialise (this->window, this->style, this->title, this->box);
		this->has_updated_partitions = FALSE;
		
		if (this->is_visible) {
			if (this->is_reduced) {
				this->grail->Reduce ();
			} else {
				this->grail->Open ();
			}
		}
	}
}



/*
 *	Prepare a window for deletion. This will hide the
 *	window if still visible and then close it. It does
 *	also remove the window from the application's list.
 */

void
OPWindow::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->is_visible) {
		this->Hide ();
	}
	
	if (this->is_open) {
		this->Close ();
	}
	
	super::Delete ();
	
	if (this->window.IsValid ()) {
		if (this->app.IsValid ()) {
			this->app->RemoveWindow (this->window);
		}
	}
	
	//	The references have to be cleared, so that a direct or indirect self-
	//	reference can be avoided.

	this->default_active    = 0;
	this->current_active    = 0;
	this->current_active_ok = FALSE;
	this->cursor_view       = 0;
	this->drag_initial_view = 0;

	this->window = 0;
}


const OPAppRef&
OPWindow::ReturnApp () const
{
	OP_METHOD (ReturnApp, 0);
	return this->app;
}



/*
 *	When awaking a window, this will automatically tell
 *	every child view who's their window.
 *
 *	This will be called anyway, even if the initial callee
 *	is AwakeFromXXX... Therefore, don't put "AwakeMyself"
 *	in these methods.
 */

void
OPWindow::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
	
	this->AwakeMyself ();
}



/*
 *	When awaking a window, this will automatically tell
 *	every child view who's their window.
 */

void
OPWindow::AwakeFromApp (OPAppIn app)
{
	OP_METHOD (AwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);
	
	this->app = app;
	
	super::AwakeFromApp (app);
}



/*
 *	When awaking a window, this will automatically tell
 *	every child view who's their window.
 */

void
OPWindow::AwakeFromObject (OPObjectIn object)
{
	OP_METHOD (AwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	super::AwakeFromObject (object);
}



/*
 *	The style of the window has to be defined before
 *	it is opened.
 */

Bool
OPWindow::SetStyle (OPWindowStyle style)
{
	OP_METHOD (SetStyle, 1);
	OP_ARG (1, OPWindowStyle, style);
	
	this->style = style;
	
	if (this->is_open) {
		
		Bool is_visible = this->is_visible;
		Bool is_reduced = this->is_reduced;
		
		this->Close ();
		this->Open ();
		this->Display ();
		
		if (is_visible) {
			if (is_reduced) {
				this->ShowReduced ();
			} else {
				this->ShowNormal ();
			}
		} else {
			this->Hide ();
		}
	}
	
	return TRUE;
}



/*
 *	Return the current window style.
 */

OPWindowStyle
OPWindow::ReturnStyle () const
{
	OP_METHOD (ReturnStyle, 0);
	return this->style;
}



/*
 *	Set the window's title. This can be done at any
 *	time. If the window has been displayed, its title
 *	bar will be updated.
 */

Bool
OPWindow::SetTitle (OPStringIn title)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, title);
	
	this->title = title;
	
	if (this->grail.IsValid ()) {
		this->grail->ChangeName (title);
	}
	
	return TRUE;
}



/*
 *	Return the window's title.
 */

Bool
OPWindow::GetTitle (OPStringOut title) const
{
	OP_METHOD (GetTitle, 1);
	OP_ARG (1, OPStringOut, title);
	
	title = this->title;
	
	return title.IsValid ();
}

const OPStringRef&
OPWindow::ReturnTitle () const
{
	OP_METHOD (ReturnTitle, 0);
	return this->title;
}



/*
 *	Set the reduced view associated to the window. It
 *	will be used to draw into the window's icon.
 */

Bool
OPWindow::SetReducedView (OPViewIn view)
{
	OP_METHOD (SetReducedView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->reduced = view;

	return TRUE;
}



/*
 *	Return the reduced view associated to the window.
 */

Bool
OPWindow::GetReducedView (OPViewOut view) const
{
	OP_METHOD (GetReducedView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->reduced;
	
	return view.IsValid ();
}

const OPViewRef&
OPWindow::ReturnReducedView () const
{
	OP_METHOD (ReturnReducedView, 0);
	return this->reduced;
}


/*
 *	Open the window : this will allocate a buffer for
 *	drawing operations and set up the system structures.
 *	Nothing will be visible on screen till ShowNormal or
 *	ShowReduced get called.
 */

void
OPWindow::Open ()
{
	OP_METHOD (Open, 0);
	
	if (this->grail.IsInvalid ()) {
		this->grail = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPGrailWindow);
		this->grail->Initialise (this->window, this->style, this->title, this->box);
		this->has_updated_partitions = FALSE;
	}
	
	this->is_open = TRUE;
}



/*
 *	Close a window. This will hide it automatically, if
 *	it hasn't been done yet. Free the buffer : any drawing
 *	are prohibited now.
 */

void
OPWindow::Close ()
{
	OP_METHOD (Close, 0);
	
	if (this->is_visible) {
		this->Hide ();
	}
	
	if (this->grail.IsValid ()) {
		this->grail->Delete ();
		this->grail = 0;
	}
	
	this->is_open    = FALSE;
	this->is_visible = FALSE;
	this->is_reduced = FALSE;
}


static void
kill_from_here (OPViewIn view)
{
	OPViewRef child;
	
	while ( view->GetFirstChild (child)
		 && child.IsValid () ) {
		kill_from_here (child);
		child->SetParent (OPViewNull);
		child->Delete ();
	}
}

/*
 *	Kill a window. This will try to free the resources allocated
 *	by the window and any of its subviews.
 */

void
OPWindow::Kill ()
{
	OP_METHOD (Kill, 0);
	
	if (this->is_open) {
		this->Close ();
	}
	
	kill_from_here (this);
	OPViewRef self = this;
	
	self->Delete ();
}


/*
 *	Show a window in its normal size. If required, the window
 *	will be opened.
 */

void
OPWindow::ShowNormal ()
{
	OP_METHOD (ShowNormal, 0);
	
	if (!this->is_open) {
		this->Open ();
		this->Display ();
	}
	
	if (this->grail.IsValid ()) {
		this->grail->Open ();
		this->is_visible = TRUE;
		this->is_reduced = FALSE;
	}
}


/*
 *	Center the window and make it visible (in its normal
 *	size).
 */

void
OPWindow::ShowCentered (OPWindowIn window)
{
	OP_METHOD (ShowCentered, 1);
	OP_ARG (1, OPWindowIn, window);
	
	OPRect box;
	
	if (window.IsValid ()) {
		window->GetBox (box);
	} else {
		OPGrailWindow::GetDisplayBox (box);
	}
	
	OPCoord x = box.X1 () + box.X2 () - this->box.DX ();
	OPCoord y = box.Y1 () + box.Y2 () - this->box.DY ();
	
	this->SetOrigin (x/2, y/2);
	this->ShowNormal ();
}


/*
 *	Show the window in a reduced size. This will
 *	usually show only an icon on screen. It is only
 *	possible if the window has been opened.
 */

void
OPWindow::ShowReduced ()
{
	OP_METHOD (ShowReduced, 0);
	
	if (!this->is_open) {
		this->Open ();
		this->Display ();
	}
	
	if (this->grail.IsValid ()) {
		this->grail->Reduce ();
		this->is_visible = TRUE;
		this->is_reduced = TRUE;
	}
}



/*
 *	Hide the window. This will make it disappear from
 *	screen, but won't close it.
 */

void
OPWindow::Hide ()
{
	OP_METHOD (Hide, 0);
	
	if ( (this->is_open)
	  && (this->grail.IsValid ())
	  && (this->is_visible == TRUE) ) {
		this->grail->Close ();
		this->is_visible = FALSE;
		this->is_reduced = FALSE;
	}
}


/*
 *	Activate the window and bring it to the front. It the
 *	window is in the reduced state, this won't change it to
 *	the normal mode.
 */

void
OPWindow::ActivateAsFrontWindow (Bool front)
{
	OP_METHOD (ActivateAsFrontWindow, 1);
	OP_ARG (1, Bool, front);
	
	if (this->grail.IsValid ()) {
		this->grail->ActivateAsFrontWindow (front);
	}
}

void
OPWindow::SysChangedActive (Bool yes)
{
	OP_METHOD (SysChangedActive, 1);
	OP_ARG (1, Bool, yes);
	
	if (yes) {
		
		OPWindowRef modal;
		
		this->Report (notify_WindowActivated);
		
		if (this->GetModalWindow (modal)) {
			this->app->NotifyActiveWindow (modal);
			modal->ShowNormal ();
			modal->ActivateAsFrontWindow (TRUE);
		} else {
			this->app->NotifyActiveWindow (this);
			if (this->modal_parent.IsValid ()) {
				this->modal_parent->NotifyActiveModalWindow (this);
			}
		}
		
	} else {
		
		this->Report (notify_WindowDeactivated);
		this->app->NotifyActiveWindow (OPWindowNull);
	}
}

void
OPWindow::SysChangedOpen (Bool yes)
{
	OP_METHOD (SysChangedOpen, 1);
	OP_ARG (1, Bool, yes);
	
	if (yes) {
		OP_ASSERT (this->is_open == yes);
		this->Report (notify_WindowOpened);
	} else {
		this->Close ();
		this->Report (notify_WindowClosed);
	}
}
 
void
OPWindow::SysChangedVisible (Bool yes)
{
	OP_METHOD (SysChangedVisible, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_visible = yes;
	
	if (yes) {
		this->Report (notify_WindowShown);
	} else {
		this->Report (notify_WindowHidden);
	}
}
 
void
OPWindow::SysChangedReduced (Bool yes)
{
	OP_METHOD (SysChangedReduced, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_reduced = yes;
	
	if (yes) {
		this->Report (notify_WindowReduced);
	} else {
		this->Report (notify_WindowMaximised);
	}
}


Bool
OPWindow::IsVisible () const
{
	OP_METHOD (IsVisible, 0);
	return this->is_visible;
}



Bool
OPWindow::IsReduced () const
{
	OP_METHOD (IsReduced, 0);
	return this->is_reduced;
}

Bool
OPWindow::IsActive () const
{
	OP_METHOD (IsActive, 0);
	return (this->grail.IsValid ()) ? this->grail->IsActive () : FALSE;
}

Bool
OPWindow::IsModal () const
{
	OP_METHOD (IsModal, 0);
	return this->modal_parent.IsValid ();
}


/*
 *	A view has been inserted in the window. It gets added
 *	to the view collection.
 */

void
OPWindow::NotifyInsertView (OPViewIn view)
{
	OP_METHOD (NotifyInsertView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->view_collection->Insert (view);
}



/*
 *	A view will be removed from the window. It gets
 *	removed from the view collection.
 */

void
OPWindow::NotifyRemoveView (OPViewIn view)
{
	OP_METHOD (NotifyRemoveView, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view == this->cursor_view) this->cursor_view = 0;
	if (view == this->drag_initial_view) this->drag_initial_view = 0;
	
	this->view_collection->Remove (view);
}



/*
 *	A view has been reparented. This event might be needed
 *	in the future...
 */

void
OPWindow::NotifyReparentView (OPViewIn view)
{
	OP_METHOD (NotifyReparentView, 1);
	OP_ARG (1, OPViewIn, view);
}



/*
 *	A view will be removed from the active list of this
 *	window. Look if it was not the current or default one.
 */

void
OPWindow::NoLongerActive (OPViewIn view)
{
	OP_METHOD (NoLongerActive, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsInvalid ()) {
		return;
	}
	
	OPViewRef other = view->ReturnNextActive ();
	
	//	Look for replacement for the specified view. Either pick up the next
	//	active view, the previous one or finally pick up the parent.
	
	if (other.IsInvalid ()) other = view->ReturnPrevActive ();
	if (other.IsInvalid ()) other = view->ReturnParent ();
	
	if ( (other.IsValid ())
	  && (other->ReturnWindow () != this) ) {
		other = OPViewNull;
	}
	
	if (view == this->default_active) {
		this->default_active = other;
	}
	
	if (view == this->current_active) {
		
		//	NIY: Notification of the change ??? No display allowed !!!
		
		this->current_active->ResignMain ();
		this->current_active = other;
		
		if (other.IsValid ()) {
			other->BecomeMain (FALSE);
		}
	}
}


void
OPWindow::NotifyInsertKeyAlt (OPControlIn ctrl)
{
	OP_METHOD (NotifyInsertKeyAlt, 1);
	OP_ARG (1, OPControlIn, ctrl);

	this->key_alt_collection->Insert (ctrl);
}

void
OPWindow::NotifyRemoveKeyAlt (OPControlIn ctrl)
{
	OP_METHOD (NotifyRemoveKeyAlt, 1);
	OP_ARG (1, OPControlIn, ctrl);

	this->key_alt_collection->Remove (ctrl);
}


/*
 *	View overlays get the opportunity to draw something after the
 *	full window has been drawn...
 */

void
OPWindow::NotifyInsertOverlay (OPViewIn view)
{
	OP_METHOD (NotifyInsertOverlay, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->view_overlays->Insert (view);
}

void
OPWindow::NotifyRemoveOverlay (OPViewIn view)
{
	OP_METHOD (NotifyRemoveOverlay, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->view_overlays->Remove (view);
}

void
OPWindow::NotifyInsertModalWindow (OPWindowIn window)
{
	OP_METHOD (NotifyInsertModalWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	if (this->modal_windows.IsInvalid ()) {
		this->modal_windows = OP_NEW (OPList);
		this->modal_windows->SetInsertMode (FALSE);
	}
	
	this->modal_windows->Insert (window);
}

void
OPWindow::NotifyRemoveModalWindow (OPWindowIn window)
{
	OP_METHOD (NotifyRemoveMOdalWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	if (this->modal_windows.IsValid ()) {
		
		this->modal_windows->Remove (window);
		
		if (this->modal_windows->TotalCount () == 0) {
			this->modal_windows->Delete ();
			this->modal_windows = OPListNull;
		}
	}
}

void
OPWindow::NotifyActiveModalWindow (OPWindowIn window)
{
	OP_METHOD (NotifyActiveModalWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	if (this->modal_windows.IsValid ()) {
		this->modal_windows->MoveToHead (window);
	}
}

Bool
OPWindow::GetModalWindow (OPWindowOut window, Count nth) const
{
	OP_METHOD (GetModalWindow, 2);
	OP_ARG (1, OPWindowOut, window);
	OP_ARG (1, Count, nth);
	
	if ( (this->modal_windows.IsValid ())
	  && (this->modal_windows->Get (nth, window)) ) {
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPWindow::GetModalWindowList (OPListOut list) const
{
	OP_METHOD (GetModalWindowList, 1);
	OP_ARG (1, OPListOut, list);
	
	list = this->modal_windows;
	
	return list.IsValid ();
}

Bool
OPWindow::SetModalWindowParent (OPWindowIn window)
{
	OP_METHOD (SetModalWindowParent, 1);
	OP_ARG (1, OPWindowIn, window);
	
	if (this->modal_parent == window) {
		return TRUE;
	}
	
	if (this->modal_parent.IsValid ()) {
		this->modal_parent->NotifyRemoveModalWindow (this);
	}
	
	this->modal_parent = window;
	
	if (this->modal_parent.IsValid ()) {
		this->modal_parent->NotifyInsertModalWindow (this);
	}
	
	return TRUE;
}

Bool
OPWindow::GetModalWindowParent (OPWindowOut window) const
{
	OP_METHOD (GetModalWindowParent, 1);
	OP_ARG (1, OPWindowOut, window);
	
	window = this->modal_parent;
	
	return window.IsValid ();
}


/*
 *	Draw the overlays (if any).
 */

void
OPWindow::DrawOverlays (OPGrailPortIn port, OPRectIn rect)
{
	OP_METHOD (DrawOverlays, 2);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	
	Count num = this->view_overlays->TotalCount ();
	
	for (Count index = 0; index < num; index++) {
		OPViewRef view;
		this->view_overlays->Get (index, view);
		view->DrawOverlay (port, rect);
	}
}


/*
 *	Redraw the specified rectangle. This will repaint
 *	every intersecting view.
 */

void
OPWindow::Redraw (OPRectIn rect)
{
	OP_METHOD (Redraw, 1);
	OP_ARG (1, OPRectIn, rect);
	
	if (this->grail.IsValid ()) {
		this->grail->BeginDraw (rect);
		this->Draw (this->grail, rect);
		this->grail->EndDraw ();
	}
}



/*
 *	The user either resized or moved the window, or both,
 *	so we have to take care of it.
 */

void
OPWindow::UserChangedGeometry (OPRectIn box)
{
	OP_METHOD (UserChangedGeometry, 1);
	OP_ARG (1, OPRectIn, box);

	if ( (this->box.DX () != box.DX ())
	  || (this->box.DY () != box.DY ()) ) {
		this->box.X1 () = box.X1 ();
		this->box.Y1 () = box.Y1 ();
		this->SetSize (box.DX (), box.DY ());
	} else {
		this->box.SetOrigin (box.X1 (), box.Y1 ());
	}

//	this->Display ();
}



/*
 *	Return the offset of the inner rectangle with respect to
 *	the real position.
 */

void
OPWindow::GetFrameOffset (OPCoordOut dx, OPCoordOut dy)
{
	OP_METHOD (GetFrameOffset, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord xl = 0;
	OPCoord xr = 0;
	OPCoord yt = 0;
	OPCoord yb = 0;

	OPGrailWindow::GetFrameSize (this->style, xl, yt, xr, yb);
	
	dx = xl;
	dy = yt;
}



/*
 *	Return the overhead (width and height) which is taken by
 *	the window's boarder.
 */

void
OPWindow::GetFrameOverhead (OPCoordOut dx, OPCoordOut dy)
{
	OP_METHOD (GetFrameOverhead, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);

	OPCoord xl = 0;
	OPCoord xr = 0;
	OPCoord yt = 0;
	OPCoord yb = 0;

	OPGrailWindow::GetFrameSize (this->style, xl, yt, xr, yb);
	
	dx = xl+xr;
	dy = yt+yb;
}



/*
 *	Return the current window GRAIL port. This might be
 *	invalid if the window has not been opened.
 */

const OPGrailPortRef&
OPWindow::ReturnGrailPort () const
{
	OP_METHOD (ReturnGrailPort, 0);
	return this->grail;
}



/*
 *	Find the view in the window, which matches the specified
 *	coordinates.
 */

Bool
OPWindow::Detect (OPCoord x, OPCoord y, OPViewOut view)
{
	OP_METHOD (Detect, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPViewOut, view);
	
	view = 0;
	
	if (this->cursor_view.IsValid ()) {
		
		//	The cursor may still be in the same view as previously. But
		//	beware, the superview may imply some additionnal clipping !
		
		OPViewRef sview = this->cursor_view;
		Bool      ok    = FALSE;
		
		while (sview.IsValid ()) {
			
			//	Look at every parent's clipping rectangle to see if it
			//	accepts the current mouse position...
			
			ok = sview->ReturnBox ().IsInside (x, y);
			
			if (ok == FALSE) {
				break;
			}
			
			sview->GetParent (sview);
		}
		
		if (ok) {
			this->cursor_view->Detect (x, y, view);
		}
	}
	
	if (view.IsInvalid ()) {
		super::Detect (x, y, view);
	}
	
	if (view == this->window) {
		view = 0;
	}
	
	return view.IsValid ();
}



/*
 *	Draw the window. This will fill the background with the
 *	predefined gray and then draw every subview.
 */

void
OPWindow::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	if (port.IsValid ()) {
		
		if ( (this->has_updated_partitions == FALSE)
		  && (this->style & OP_WIN_PARTITION) ) {
			this->UpdatePartitions ();
		}
		
		OPAdornerRef adorner = port->ReturnAdorner ();
		OPRect       box     = this->box;
		
		box.SetOrigin (0, 0);
		
		if ( (what & OP_DRAW_BACKGROUND)
		  && ((this->style & OP_WIN_TRANSPARENT) == 0) ) {
			
			OPLookColor color = (this->state & OP_VIEW_IS_DROPPING)
							  ? OP_COLOR_VIEW_DROPPING
							  : OP_COLOR_BACKGROUND;
			
			adorner->FillRect (box, color);
		}
		
		super::Draw (port, rect, what);
		
		//	If there are any overlay views, draw them now...
		
		this->DrawOverlays (port, rect);
		
		if (this->state & OP_VIEW_IS_CONNECT) {
			adorner->DrawRect (box, OP_COLOR_VIEW_CONNECT, 0);
			adorner->DrawRect (box, OP_COLOR_VIEW_CONNECT, 1);
		}
	}
}


/*
 *	An event has been received by the application and will
 *	have to be routed to the proper responder.
 */

void
OPWindow::DispatchEvent (OPEventIn event)
{
	OP_METHOD (DispatchEvent, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->drag_initial_view.IsValid ()) {
		
		//	If dragging in progress, only the view where dragging started
		//	will be notified of the events... until dragging ends.
		
		this->drag_initial_view->DispatchEvent (event);
		
		//	Check if dragging will continue or if we have to stop (if so,
		//	this would mean the BUTTON RELEASE event hasn't been handled
		//	properly be the dragging view).
		
		if (event->ReturnButtons () == 0) {
			this->drag_initial_view = 0;
		}
		
		return;
	}
	
	//	The user pressed some key. Look if it is a hot key before
	//	forwarding it...
	
	if (event->ReturnType () == OP_EVENT_KEY_DOWN) {
		
		Count total = this->key_alt_collection->TotalCount ();
		Count index = 0;
		Bool  ok    = FALSE;

		while (total--) {
			OPControlRef ctrl;
			this->key_alt_collection->Get (index++, ctrl);
			ok |= ctrl->HandleKeyAlt (event);
		}
		
		if (ok) return;
	}
	
	//	The user CONTROL-SHIFT-clicked in the window. This will select
	//	or deselect a view for edition.
	
	if ( (event->ReturnType () == OP_EVENT_PRESS)
	  && (event->ReturnButtons () == OP_EV_BUTTON_LEFT)
	  && (event->ReturnKey ().modifiers & OP_EV_SHIFT)
	  && (event->ReturnKey ().modifiers & OP_EV_CONTROL)
	  && (this->app->IsBuildEnabled ()) ) {
		
		OPViewRef view;
		OPViewRef parent;
		OPRect    box;
		
		this->Detect (event->ReturnX (), event->ReturnY (), view);
		
		//	Find which view will have to be dynamically edited.
		
		for (;;) {
			
			if ( (view.IsInvalid ())
			  || (view == this) ) {
				return;
			}
			
			view->GetBox (box);
			view->GetParent (parent);
			
			//	If the current view cannot be edited, work the way up to the
			//	window...
			
			if (view->DragAndDrop (box, this, OP_DDP_UI_EDIT_CHECK)) {
				break;
			}
			
			view = parent;
		}
		
		this->RemoveMainResponder ();
		
		//	Toggle between the selected view and the deselected one.
		
		if (view->IsKindOf (OP_RUNTIME (OPSelView))) {
			
			OPSelViewRef sel = view;
			
			this->cursor_view = sel->ReturnSelectedView ();
			
			sel->EndSelection ();
			sel->Delete ();
			
		} else {
			
			OPSelViewRef sel = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPSelView);
			
			if (sel->BeginSelection (view) == FALSE) {
				
				sel->Delete ();
				
			} else {
				
				this->cursor_view       = sel;
				this->drag_initial_view = 0;
			}
		}
		
		this->SetMainResponder (this->cursor_view, TRUE);
		return;
	}
	
	
	//	Control-clicking on an object will start a connection... if the builder
	//	is enabled.
	
	if ( (event->ReturnType () == OP_EVENT_PRESS)
	  && (event->ReturnButtons () == OP_EV_BUTTON_LEFT)
	  && ((event->ReturnKey ().modifiers & OP_EV_MOD_MASK) == OP_EV_CONTROL)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT)
	  && (this->app->IsBuildEnabled ()) ) {
		
		OPViewRef view;
		this->Detect (event->ReturnX (), event->ReturnY (), view);
		
		if (view.IsInvalid ()) {
			view = this;
		}
		
		if ( (view.IsValid ())
		  && (view->IsKindOf (OP_RUNTIME (OPSelView)) == FALSE) ) {
			
			OPObjectRef builder = this->app->ReturnBuilder ();
			
			builder->DefineData (OP_BUILDER_OBJECT, view);
			builder->DefineData (OP_BUILDER_SELVIEW, OPSelViewNull);
			builder->Action (OP_BUILDER_CONNECT);
			
			return;
		}
	}
	
	
	//	Shift-clicking on an object will show the edition menu... if the builder
	//	is enabled.
	
	if ( (event->ReturnType () == OP_EVENT_PRESS)
	  && (event->ReturnButtons () == OP_EV_BUTTON_LEFT)
	  && ((event->ReturnKey ().modifiers & OP_EV_MOD_MASK) == OP_EV_SHIFT)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT)
	  && (this->app->IsBuildEnabled ()) ) {
		
		OPViewRef view;
		this->Detect (event->ReturnX (), event->ReturnY (), view);
		
		if (view.IsInvalid ()) {
			view = this;
		}
		
		if ( (view.IsValid ())
		  && (view->IsKindOf (OP_RUNTIME (OPSelView)) == FALSE) ) {
			
			OPObjectRef builder = this->app->ReturnBuilder ();
			
			builder->DefineData (OP_BUILDER_OBJECT, view);
			builder->DefineData (OP_BUILDER_SELVIEW, OPSelViewNull);
			builder->Action (OP_BUILDER_MENU);
			
			return;
		}
	}
	
	if ( (event->ReturnType () == OP_EVENT_MOTION)
	  || (event->ReturnType () == OP_EVENT_PRESS)
	  || (event->ReturnType () == OP_EVENT_RELEASE) ) {
		
		OPViewRef view;
		
		this->Detect (event->ReturnX (), event->ReturnY (), view);
		
		if (view != this->cursor_view) {
			
			if (this->cursor_view.IsValid ()) {
				this->cursor_view->LeaveNotify (event);
			}
			
			if (view.IsValid ()) {
				view->EnterNotify (event);
			}
			
			this->cursor_view = view;
		}
		
		(view.IsValid ()) ? view->DispatchEvent (event)
						  : super::DispatchEvent (event);
		
		return;
	}
	
	if ( (this->current_active.IsValid ())
	  && (this->current_active_ok) ) {
		
		//	Inhibit temporarily the deletion of the object to be called. It might
		//	otherwise vanish before it returns here, which could be catastrophic.
		
		OPViewRef view = this->current_active;

		if (view == this) {
			return;
		}

		view->DispatchEvent (event);
		
		return;
	}
	
	super::DispatchEvent (event);
}



/*
 *	Define the new Main Responder. If another main responder
 *	had been defined previously, it will have to resign...
 *	Returns TRUE if "responder" became the Main Responder.
 */

Bool
OPWindow::SetMainResponder (OPResponderIn responder, Bool clicked)
{
	OP_METHOD (SetMainResponder, 2);
	OP_ARG (1, OPResponderIn, responder);
	OP_ARG (2, Bool, clicked);
	
	OPResponderRef resp = responder;
	
	for (;;) {
		if (resp.IsInvalid ()) return FALSE;
		if (resp->WouldAcceptMain () == TRUE) break;

		OPViewRef view = resp;
		
		//	If the responder does not accept to become the main responder,
		//	look for its children (if any).
		
		for (;;) {
			
			view = view->ReturnFirstActive ();
			
			if (view.IsInvalid ()) break;
			
			if (view->WouldAcceptMain ()) {
				resp = view;
				goto ok;
			}
		}

		resp->GetNextResponder (resp);
	}

ok:	if (this->current_active == resp) {
		
		//	Activating the currently active view won't do anything
		//	special...
		
		return TRUE;
	}
	
	if (this->current_active.IsValid ()) {
		this->current_active->ResignMain ();
	}
	
	this->current_active    = resp;
	this->current_active_ok = TRUE;
	
	this->current_active->BecomeMain (clicked);
	
	return TRUE;
}



/*
 *	Return current main responder.
 */

Bool
OPWindow::GetMainResponder (OPResponderOut responder) const
{
	OP_METHOD (GetMainResponder, 1);
	OP_ARG (1, OPResponderOut, responder);
	
	responder = this->current_active;
	return responder.IsValid ();
}

Bool
OPWindow::GetActiveView (OPViewOut view) const
{
	OP_METHOD (GetActiveView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->current_active;
	return view.IsValid ();
}

Bool
OPWindow::GetCursorView (OPViewOut view) const
{
	OP_METHOD (GetCursorView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->cursor_view;
	return view.IsValid ();
}



/*
 *	Disable the main responder : this is a temporary
 *	change which can be undone using "EnableMainResponder".
 */

void
OPWindow::DisableMainResponder ()
{
	OP_METHOD (DisableMainResponder, 0);
	this->current_active_ok = FALSE;
}



/*
 *	Enable the main responder.
 */

void
OPWindow::EnableMainResponder ()
{
	OP_METHOD (EnableMainResponder, 0);
	this->current_active_ok = TRUE;
}



/*
 *	Remove the main responder. This method has to be used
 *	when a responder wants to resign and cannot offer an
 *	other responder as replacement.
 */

void
OPWindow::RemoveMainResponder ()
{
	OP_METHOD (RemoveMainResponder, 0);
	
	if (this->current_active.IsValid ()) {
		this->current_active->ResignMain ();
		this->current_active = 0;
	}
}



/*
 *	Return TRUE if the current main responder is valid
 *	and enabled.
 */

Bool
OPWindow::IsMainResponderOk () const
{
	OP_METHOD (IsMainResponderOk, 0);
	
	return (this->current_active_ok) && (this->current_active.IsValid ())
		   ? TRUE : FALSE;
}



/*
 *	Specify which view will get the dragging events and this, until
 *	the mouse button is released. When dragging stops, the view has
 *	to call SetDraggingView with a null reference.
 */

Bool
OPWindow::SetDraggingView (OPViewIn view)
{
	OP_METHOD (SetDraggingView, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsValid ()) {
		if (this->drag_initial_view.IsValid ()) {
			this->SetDraggingView (OPViewNull);
		}
	}
	
	if (view == this->drag_initial_view) {
		return TRUE;
	}
	
	this->drag_initial_view = view;
	
	OPEvent::GrabWindowEvents (this, view.IsValid ());
	
	return TRUE;
}



/*
 *	Return current dragging view.
 */

Bool
OPWindow::GetDraggingView (OPViewOut view) const
{
	OP_METHOD (GetDraggingView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->drag_initial_view;
	return view.IsValid ();
}


/*
 *	Update the partitions (and create them if needed).
 */

Bool
OPWindow::UpdatePartitions ()
{
	OP_METHOD (UpdatePartitions, 0);
	
	if (this->grail.IsValid ()) {
		
		OPViewRef child;
		OPRect    box;
		
		this->grail->RemovePartitions ();
		this->GetFirstChild (child);
		
		while (child.IsValid ()) {
			
			child->GetBox (box);
			
			if (child->ReturnState () & OP_VIEW_OK_DRAW) {
				this->grail->AddPartition (box);
			}
			
			child->GetNextSibling (child);
		}
		
		this->has_updated_partitions = TRUE;
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	The data slot is an object which is attached to the window
 *	when it is being created. It is a place holder used to
 *	exchange data between user interface and application.
 */

Bool
OPWindow::SetSlot (OPDataSlotIn slot)
{
	OP_METHOD (SetSlot, 1);
	OP_ARG (1, OPDataSlotIn, slot);
	
	this->slot = slot;
	return TRUE;
}

Bool
OPWindow::GetSlot (OPDataSlotOut slot)
{
	OP_METHOD (GetSlot, 1);
	OP_ARG (1, OPDataSlotOut, slot);
	
	slot = this->slot;
	return slot.IsValid ();
}

const OPDataSlotRef&
OPWindow::ReturnSlot ()
{
	OP_METHOD (ReturnSlot, 0);
	return this->slot;
}





void
OPWindow::KeyDown (OPEventIn event)
{
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::KeyDown (event);
}

void
OPWindow::KeyChange (OPEventIn event)
{
	OP_METHOD (KeyChange, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::KeyChange (event);
}

void
OPWindow::KeyUp (OPEventIn event)
{
	OP_METHOD (KeyUp, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::KeyUp (event);
}

void
OPWindow::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::ButtonPress (event);
}

void
OPWindow::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::ButtonRelease (event);
}

void
OPWindow::MotionNotify (OPEventIn event)
{
	OP_METHOD (MotionNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::MotionNotify (event);
}

void
OPWindow::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::DragNotify (event);
}

void
OPWindow::Miscellanous (OPEventIn event)
{
	OP_METHOD (Miscellanous, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::Miscellanous (event);
}

void
OPWindow::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	super::EnterNotify (event);
}

void
OPWindow::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->cursor_view.IsValid ()) {
		
		OPViewRef view = this->cursor_view;
		this->cursor_view = OPViewNull;
		
		if (view == this) {
			//	Don't notify ourselves...
		} else {
			if (view != this->drag_initial_view) {
				view->LeaveNotify (event);
			}
		}
		
	}
	
	super::LeaveNotify (event);
}
