/*
 *	gfxdrawing.cxx
 *
 *	Basic drawing methods supporting transparency and an alpha
 *	channel. This is based on the GfxSurface class (libgrafix.cxx
 *	implements the core of GfxSurface).
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

#include "gfxtypes.h"
#include "gfxsurface.h"
#include "gfxhelpers.h"

/*****************************************************************************/

#define	MAX_LINE_LENGTH		10000

/*****************************************************************************/

/*
 *	Compute the Y co-ordinates of a line. This requires that :
 *	
 *	- dx  >  0,   the line is drawn from left to right
 *	- dx >= |dy|, the angle lies between -45 and +45 degrees.
 *
 *	The dots array will be filled with dx+1 values. The errors will
 *	be in the range of 0 to 255, with 0 specifying an exact match.
 *	This is used for anti-aliasing.
 *
 *	The error parameter can be used to specify an initial vertical
 *	error (useful only in the range 0 to 0.996).
 */

void
GfxSurface::ComputeLineDotsDX (Card32 x, Card32 y, Card32 dx, Int32 dy, Real error)
{
	if (error < 0) {
		Card32 num = (Card32)(-error);
		error += (Real)(num+1);
		y     -= num;
	} else if (error >= 1.0) {
		Card32 num = (Card32)(error);
		error -= (Real)(num);
		y     += num;
	}
	
	Real m = (Real)(dy) / (Real)(dx);
	Real e = error;
	
	if (this->dots == 0) this->AllocDotsBuffer ();
	LineDot* dots = this->dots;
	
	if (m >= 0) {
		
		while (dx--) {
			
			dots->error = (Card8)(e * 255.99);
			dots->pos   = y;
			
			dots += 1;
			e    += m;
			
			if (e >= 0.9999) {
				e -= 1.0;
				y++;
			}
		}
		
		dots->error = (Card8)(e * 255.99);
		dots->pos   = y;
		
	} else {
		
		while (dx--) {
			
			dots->error = (Card8)(e * 255.99);
			dots->pos   = y;
			
			dots += 1;
			e    += m;
			
			if (e < 0.0) {
				e += 1.0;
				y--;
			}
		}
		
		dots->error = (Card8)(e * 255.99);
		dots->pos   = y;
	}
}

/*
 *	Compute the X co-ordinates of a line. This requires that :
 *	
 *	- dy  >  0,   the line is drawn from top to bottom
 *  - dy >= |dx|, the angle lies between +45 and +135 degrees.
 *
 *	Same remarks as for ComputeLineDXDots.
 */

void
GfxSurface::ComputeLineDotsDY (Card32 x, Card32 y, Int32 dx, Card32 dy, Real error)
{
	if (error < 0) {
		Card32 num = (Card32)(-error);
		error += (Real)(num+1);
		x     -= num;
	} else if (error >= 1.0) {
		Card32 num = (Card32)(error);
		error -= (Real)(num);
		x     += num;
	}
	
	Real m = (Real)(dx) / (Real)(dy);
	Real e = error;
	
	if (this->dots == 0) this->AllocDotsBuffer ();
	LineDot* dots = this->dots;
	
	if (m >= 0) {
		
		while (dy--) {
			
			dots->error = (Card8)(e * 255.99);
			dots->pos   = x;
			
			dots += 1;
			e    += m;
			
			if (e >= 0.9999) {
				e -= 1.0;
				x++;
			}
		}
		
		dots->error = (Card8)(e * 255.99);
		dots->pos   = x;
		
	} else {
		
		while (dy--) {
			
			dots->error = (Card8)(e * 255.99);
			dots->pos   = x;
			
			dots += 1;
			e    += m;
			
			if (e < 0.0) {
				e += 1.0;
				x--;
			}
		}
		
		dots->error = (Card8)(e * 255.99);
		dots->pos   = x;
	}
}

/*****************************************************************************/

/*
 *	Draw a solid line including both the starting point and the end.
 *	The pixel value is used as is. No anti-aliasing is done.
 */

void
GfxSurface::DrawLine (Int32 x1, Int32 y1, Int32 x2, Int32 y2, Card32 pixel)
{
	Card32 dx = (x2>x1) ? x2-x1 : x1-x2;
	Card32 dy = (y2>y1) ? y2-y1 : y1-y2;
	
	if (dx >= dy) {
		
		//	Mostly horizontal line... Always drom from left to right.
		
		if (x1 > x2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32   x  = x1;
		Int32   y  = y1;
		Card32  dx = (Card32)(x2-x1);
		Int32   dy = y2-y1;
		
		if (dx == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDX (x, y, dx, dy, 0.5);
		LineDot* dots = this->dots;
		
		do {
			this->SetPixelRGBBlend (x++, dots->pos, pixel);
			dots++;
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line... Always draw from top to bottom.
		
		if (y1 > y2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Int32  dx = x2-x1;
		Card32 dy = (Card32)(y2-y1);
		
		if (dy == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDY (x, y, dx, dy, 0.5);
		LineDot* dots = this->dots;
		
		do {
			this->SetPixelRGBBlend (dots->pos, y++, pixel);
			dots++;
		} while (dy--);
	}
}


/*
 *	Draw a dashed line including both the starting point and the end.
 *	The pixel value is used as is. No anti-aliasing is done.
 */

void
GfxSurface::DrawLine (Int32 x1, Int32 y1, Int32 x2, Int32 y2,
					  Card32 pixel, const Card16* dash, int num)
{
	if ( (dash == 0) || (num == 0) || (dash[0] == 0) ) {
		this->DrawLine (x1, y1, x2, y2, pixel);
		return;
	}
	
	Card32 dx = (x2>x1) ? x2-x1 : x1-x2;
	Card32 dy = (y2>y1) ? y2-y1 : y1-y2;
	
	if (dx >= dy) {
		
		//	Mostly horizontal line... Always drom from left to right.
		
		if (x1 > x2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Card32 dx = (Card32)(x2-x1);
		Int32  dy = y2-y1;
		
		if (dx == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDX (x, y, dx, dy, 0.5);
		LineDot* dots  = this->dots;
		int      index = 0;
		Card16   count = dash[index++];
		Bool     draw  = TRUE;
		
		do {
			while (count == 0) {
				if (index >= num) index = 0;
				count = dash[index++];
				draw  = !draw;
			}
			if (draw) {
				this->SetPixelRGBBlend (x, dots->pos, pixel);
			}
			x++;
			dots++;
			count--;
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line... Always draw from top to bottom.
		
		if (y1 > y2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Card32 x  = x1;
		Card32 y  = y1;
		Int32  dx = x2-x1;
		Card32 dy = (Card32)(y2-y1);
		
		if (dy == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDY (x, y, dx, dy, 0.5);
		LineDot* dots = this->dots;
		int      index = 0;
		Card16   count = dash[index++];
		Bool     draw  = TRUE;
		
		do {
			while (count == 0) {
				if (index >= num) index = 0;
				count = dash[index++];
				draw  = !draw;
			}
			if (draw) {
				this->SetPixelRGBBlend (dots->pos, y, pixel);
			}
			y++;
			dots++;
			count--;
		} while (dy--);
	}
}


/*
 *	Draw a solid anti-aliased line including both the starting point
 *	and the end. The pixel value is combined with an internal alpha
 *	value (depending on the anti-aliasing value).
 */

void
GfxSurface::DrawAALine (Int32 x1, Int32 y1, Int32 x2, Int32 y2, Card32 pixel)
{
	Card32 dx    = (x2>x1) ? x2-x1 : x1-x2;
	Card32 dy    = (y2>y1) ? y2-y1 : y1-y2;
	Card32 alpha = pixel >> 24;
	Card32 trans = 0;
	
	LineDot* dots  = this->dots;
	
	if (alpha > 127) alpha++;
	
	pixel &= 0x00FFFFFF;
	trans  = 256 - alpha;
	
	//	If trans == 0 => fully transparent pixel, 256 => fully opaque pixel. This
	//	gets multiplied by the transparency of each drawn pixel.
	
	if (dx >= dy) {
		
		//	Mostly horizontal line...
		
		if (x1 > x2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Card32 dx = (Card32)(x2-x1);
		Int32  dy = y2-y1;
		
		if (dx == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDX (x, y, dx, dy);
		
		do {
			Card32 error   = 255-dots->error;			//	error==0 => transparent
			Card32 combine = (trans*error >> 8) << 24;	//	comination of both transparencies
			this->SetPixelRGBBlend (x, dots->pos, pixel | combine ^ 0xFF000000);
			this->SetPixelRGBBlend (x, dots->pos+1, pixel | combine);
			x++;
			dots++;
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line...
		
		if (y1 > y2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Int32  dx = x2-x1;
		Card32 dy = (Card32)(y2-y1);
		
		if (dy == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDY (x, y, dx, dy);
		
		do {
			Card32 error   = 255-dots->error;			//	error==0 => transparent
			Card32 combine = (trans*error >> 8) << 24;	//	comination of both transparencies
			this->SetPixelRGBBlend (dots->pos, y, pixel | combine ^ 0xFF000000);
			this->SetPixelRGBBlend (dots->pos+1, y, pixel | combine);
			y++;
			dots++;
		} while (dy--);
	}
}


/*
 *	Draw a dashed anti-aliased line including both the starting point
 *	and the end. The pixel value is combined with an internal alpha
 *	value (depending on the anti-aliasing value).
 */

void
GfxSurface::DrawAALine (Int32 x1, Int32 y1, Int32 x2, Int32 y2,
						Card32 pixel, const Card16* dash, int num)
{
	if ( (dash == 0) || (num == 0) || (dash[0] == 0) ) {
		this->DrawAALine (x1, y1, x2, y2, pixel);
		return;
	}
	
	Card32 dx    = (x2>x1) ? x2-x1 : x1-x2;
	Card32 dy    = (y2>y1) ? y2-y1 : y1-y2;
	Card32 alpha = pixel >> 24;
	Card32 trans = 0;
	
	LineDot* dots  = this->dots;
	
	if (alpha > 127) alpha++;
	
	pixel &= 0x00FFFFFF;
	trans  = 256 - alpha;
	
	//	If trans == 0 => fully transparent pixel, 256 => fully opaque pixel. This
	//	gets multiplied by the transparency of each drawn pixel.
	
	if (dx >= dy) {
		
		//	Mostly horizontal line...
		
		if (x1 > x2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Card32 dx = (Card32)(x2-x1);
		Int32  dy = y2-y1;
		
		if (dx == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDX (x, y, dx, dy);
		
		int      index = 0;
		Card16   count = dash[index++];
		Bool     draw  = TRUE;
		
		do {
			while (count == 0) {
				if (index >= num) index = 0;
				count = dash[index++];
				draw  = !draw;
			}
			
			Card32 error   = 255-dots->error;			//	error==0 => transparent
			Card32 combine = (trans*error >> 8) << 24;	//	comination of both transparencies
			
			if (draw) {
				this->SetPixelRGBBlend (x, dots->pos, pixel | combine ^ 0xFF000000);
				this->SetPixelRGBBlend (x, dots->pos+1, pixel | combine);
			}
			
			x++;
			dots++;
			count--;
			
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line...
		
		if (y1 > y2) {
			Int32 xt = x2;
			Int32 yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Int32  x  = x1;
		Int32  y  = y1;
		Int32  dx = x2-x1;
		Card32 dy = (Card32)(y2-y1);
		
		if (dy == 0) {
			this->SetPixelRGBBlend (x, y, pixel);
			return;
		}
		
		GFX_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDotsDY (x, y, dx, dy);
		
		int      index = 0;
		Card16   count = dash[index++];
		Bool     draw  = TRUE;
		
		do {
			while (count == 0) {
				if (index >= num) index = 0;
				count = dash[index++];
				draw  = !draw;
			}
			
			Card32 error   = 255-dots->error;			//	error==0 => transparent
			Card32 combine = (trans*error >> 8) << 24;	//	comination of both transparencies
			
			if (draw) {
				this->SetPixelRGBBlend (dots->pos, y, pixel | combine ^ 0xFF000000);
				this->SetPixelRGBBlend (dots->pos+1, y, pixel | combine);
			}
			
			y++;
			dots++;
			count--;
			
		} while (dy--);
	}
}




/*
 *	Allocate the buffer for the dot co-ordinates (used for the line
 *	draing algorithms).
 */

void
GfxSurface::AllocDotsBuffer ()
{
	this->dots = NEW LineDot[MAX_LINE_LENGTH];
}
