/*
 *	springbox.cxx
 *
 *	Spring box classes used to represent and edit the spring constants
 *	attached to a view.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "springbox.h"

#include <OPaC/app.h>
#include <OPaC/button.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/image.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/spring.h>
#include <OPaC/springboard.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
APP_IMPLEMENT_CLASS (OPSpringBox);

OP_PUBLISH_SLOT (OPSpringBox, band, ReturnSpringBand, SetSpringBand, "Spring band", OPSpringBand);
OP_PUBLISH_SLOT (OPSpringBox, view, ReturnSpringView, SetSpringView, "Spring view", OPView);

OP_TARGET_ACTOR (OPSpringBox, SpringChanged);

OP_NOTIFY_EVENT (OPSpringBox, SpringBeforePress);
OP_NOTIFY_EVENT (OPSpringBox, SpringMiddlePress);
OP_NOTIFY_EVENT (OPSpringBox, SpringAfterPress);

/****************************************************************/

OPSpringBox::OPSpringBox ()
{
	OP_METHOD (OPSpringBox, 0);
	
	this->is_uptodate = FALSE;
	this->is_vertical = FALSE;
	this->SetAttribute ("OPaC.volatile.observer");
}

OPSpringBox::~OPSpringBox ()
{
	OP_METHOD (~OPSpringBox, 0);
}

void
OPSpringBox::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->band = OPSpringBandNull;
	this->view = OPViewNull;
	
	super::Delete ();
}

Bool
OPSpringBox::SetSpringBand (OPSpringBandIn band)
{
	OP_METHOD (SetSpringBand, 1);
	OP_ARG (1, OPSpringBandIn, band);
	
	this->band = band;
	this->is_uptodate = FALSE;
	
	return TRUE;
}

Bool
OPSpringBox::SetSpringView (OPViewIn view)
{
	OP_METHOD (SetSpringView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->view = view;
	this->is_uptodate = FALSE;
	
	return TRUE;
}



const OPSpringBandRef&
OPSpringBox::ReturnSpringBand () const
{
	OP_METHOD (ReturnSpringBand, 0);
	return this->band;
}

const OPViewRef&
OPSpringBox::ReturnSpringView () const
{
	OP_METHOD (ReturnSpringView, 0);
	return this->view;
}


void
OPSpringBox::UpdateSprings ()
{
	OP_METHOD (UpdateSprings, 0);
	
	if (this->is_uptodate) return;
	
	this->is_uptodate = TRUE;
	this->is_valid    = FALSE;
	
	this->s_before = OPSpringNull;
	this->s_middle = OPSpringNull;
	this->s_after  = OPSpringNull;
	
	if (this->band.IsInvalid ()) return;
	if (this->band->FindView (this->view) == FALSE) return;
	
	this->is_vertical = this->band->IsVertical ();
	this->is_valid    = TRUE;
	
	this->band->FindSpringBeforeView (this->view, this->s_before);
	this->band->FindSpringOfView (this->view, this->s_middle);
	this->band->FindSpringAfterView (this->view, this->s_after);
}

static Bool
is_stretch_spring (OPSpringIn spring)
{
	if (spring.IsInvalid ()) return FALSE;
	if (spring->CountSamples () < 2) return FALSE;
	
	if (spring->FindSlope (1.0)) {
		return TRUE;
	}
	
	return FALSE;
}

static void
draw_v_spring (OPGrailPortIn port, OPCoord x, OPCoord y, OPCoord d, OPSpringIn spring)
{
	if (is_stretch_spring (spring)) {
		port->DrawLineDY (x, y, 3);
		port->DrawLineDY (x, d-3, 3);
		port->DrawRect (x-3, y+3, x+3, d-3);
	} else {
		port->DrawLineToY (x, y, d);
	}
}

static void
draw_h_spring (OPGrailPortIn port, OPCoord x, OPCoord y, OPCoord d, OPSpringIn spring)
{
	if (is_stretch_spring (spring)) {
		port->DrawLineDX (x, y, 3);
		port->DrawLineDX (d-3, y, 3);
		port->DrawRect (x+3, y-3, d-3, y+3);
	} else {
		port->DrawLineToX (x, y, d);
	}
}


void
OPSpringBox::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	this->UpdateSprings ();
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	if (what & OP_DRAW_BACKGROUND) {
	}
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		OPCoord ox = this->box.X1 ();
		OPCoord oy = this->box.Y1 ();
		OPCoord dx = this->box.DX ();
		OPCoord dy = this->box.DY ();
		OPCoord d  = lookman->Return3DEffectWidth ();
		
		OPRect inside (ox + dx/3 - d, oy + dy/3 - d, ox + 2*dx/3 + d, oy + 2*dy/3 + d);
		
		if (this->is_valid == FALSE) {
			
			port->SetForeColor (OP_COLOR_DGRAY);
			port->DrawRect (this->box);
			
		} else {
			
			port->SetForeColor (OP_COLOR_BLACK);
			
			if (this->is_vertical) {
				
				OPCoord x = ox + dx/2;
				
				Int32 y0 = (Int32)(oy+0*dy/3);
				Int32 y1 = (Int32)(oy+1*dy/3);
				Int32 y2 = (Int32)(oy+2*dy/3);
				Int32 y3 = (Int32)(oy+3*dy/3);
				
				draw_v_spring (port, x, y0+1, y1-2, this->s_before);
				draw_v_spring (port, x, y1+0, y2-0, this->s_middle);
				draw_v_spring (port, x, y2+2, y3-2, this->s_after);
				
			} else {
				
				OPCoord y = oy + dy/2;
				
				Int32 x0 = (Int32)(ox+0*dx/3);
				Int32 x1 = (Int32)(ox+1*dx/3);
				Int32 x2 = (Int32)(ox+2*dx/3);
				Int32 x3 = (Int32)(ox+3*dx/3);
				
				draw_h_spring (port, x0+1, y, x1-2, this->s_before);
				draw_h_spring (port, x1+0, y, x2-0, this->s_middle);
				draw_h_spring (port, x2+2, y, x3-2, this->s_after);
			}
			
			adorner->DrawRectHigh (inside);
			adorner->DrawRectLow (this->box);
		}
	}
}


void
OPSpringBox::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->is_valid) {
		
		Real x = (event->ReturnX () - this->box.X1 ()) / this->box.DX ();
		Real y = (event->ReturnY () - this->box.Y1 ()) / this->box.DY ();
		
		OPSpringRef spring = OPSpringNull;
		OPViewRef   parent = this->view->ReturnParent ();
		OPAtomRef   which  = OPAtomNull;
		
		if (this->is_vertical) {
			
			if ((x > 0.4) && (x < 0.6)) {
				if (y < 0.33) {
					spring = this->s_before;
					which  = notify_SpringBeforePress;
				} else if (y > 0.66) {
					spring = this->s_after;
					which  = notify_SpringAfterPress;
				} else {
					spring = this->s_middle;
					which  = notify_SpringMiddlePress;
				}
			}
			
		} else {
			
			if ((y > 0.4) && (y < 0.6)) {
				if (x < 0.33) {
					spring = this->s_before;
					which  = notify_SpringBeforePress;
				} else if (x > 0.66) {
					spring = this->s_after;
					which  = notify_SpringAfterPress;
				} else {
					spring = this->s_middle;
					which  = notify_SpringMiddlePress;
				}
			}
		}
		
		if (spring.IsValid ()) {
			
			if (is_stretch_spring (spring)) {
				
				OPCoord size = spring->ReturnDefaultSize ();
				
				spring->RemoveAll ();
				spring->Reset (size, 0.0);
				
				this->band->ResetFromModel ();
				
			} else {
				
				OPCoord size = spring->ReturnDefaultSize ();
				
				spring->RemoveAll ();
				spring->Reset (0.0, 1.0);
				
				this->band->ResetFromModel ();
			}
			
			this->Report (which);
			this->Display ();
		}
	}
}


/*
 *	The spring band has changed, we will have to update the
 *	internally cached springs.
 */

Bool
OPSpringBox::SpringChanged ()
{
	OP_METHOD (SpringChanged, 0);
	
	this->is_uptodate = FALSE;
	
	return TRUE;
}
