/*
 *	springedit.cxx
 *
 *	The Spring Edit class is used to edit a view's springs.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "springedit.h"

#include <OPaC/app.h>
#include <OPaC/array.h>
#include <OPaC/boxview.h>
#include <OPaC/button.h>
#include <OPaC/grail.h>
#include <OPaC/image.h>
#include <OPaC/lookman.h>
#include <OPaC/multiview.h>
#include <OPaC/new.h>
#include <OPaC/spring.h>
#include <OPaC/springboard.h>
#include <OPaC/switch.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
APP_IMPLEMENT_CLASS (OPSpringEdit);

OP_TARGET_ACTOR (OPSpringEdit, DoBandNew);
OP_TARGET_ACTOR (OPSpringEdit, DoBandSelect);
OP_TARGET_ACTOR (OPSpringEdit, DoBandInsert);
OP_TARGET_ACTOR (OPSpringEdit, DoBandRemove);

OP_TARGET_ACTOR (OPSpringEdit, DoShowHorizontal);
OP_TARGET_ACTOR (OPSpringEdit, DoShowVertical);

OP_TARGET_OBSERVER (OPSpringEdit, ActiveViewChanged);

/****************************************************************/

OPSpringEdit::OPSpringEdit ()
{
	OP_METHOD (OPSpringEdit, 0);
	
	this->buttons  = OP_NEW (OPArray);
	this->switches = OP_NEW (OPArray);
	
	this->band_num    = 0;
	this->is_vertical = FALSE;
	this->SetAttribute ("OPaC.volatile.observer");
}

OPSpringEdit::~OPSpringEdit ()
{
	OP_METHOD (~OPSpringEdit, 0);
}

void
OPSpringEdit::Delete ()
{
	OP_METHOD (Delete, 0);
	
	Count num   = 0;
	Count index = 0;
	
	num = this->buttons->TotalCount ();
	
	for (index = 0; index < num; index++) {
		this->buttons[index]->Delete ();
	}
	
	num = this->switches->TotalCount ();
	
	for (index = 0; index < num; index++) {
		this->switches[index]->Delete ();
	}
	
	this->buttons->SetSize (0);
	this->switches->SetSize (0);
	
	if (this->spring_view.IsValid ()) {
		this->spring_view->Delete ();
		this->spring_view = OPSpringViewNull;
	}
	
	if (this->view.IsValid ()) {
		this->view->Delete ();
	}
	
	if (this->edit.IsValid ()) {
		this->edit->ReturnWindow ()->Display ();
	}
	
	this->edit = OPObjectNull;
	this->view = OPViewNull;
	
	super::Delete ();
}

void
build_interface (OPSpringEditIn editor,
				 OPViewIn view, OPSpringBoxOut spring_box,
				 OPButtonArrayInOut buttons,
				 OPSwitchArrayInOut switches)
{
	OPCoord dx;
	OPCoord dy;
	
	buttons->SetSize (4);
	switches->SetSize (2);
	
	view->GetSize (dx, dy);
	
	dx -= 120;
	dy -= 20;
	
	OPBoxViewRef box = OP_NEW (OPBoxView);
	OPImageRef image = OP_NEW (OPImage);
	
	image->Initialise ("springs.gif");
	image->SetOrigin (10, 10);
	image->SetSize (dx, dy);
	image->SetParent (view);
	
	box->Initialise (10, 10, dx, dy, "", image);
	box->SetEnableTitle (FALSE);
		
	switches[0] = OP_NEW (OPSwitch);
	switches[0]->Initialise (dx+20, 20, 85, 20, "horizontal", view);
	switches[0]->InitAsRadioButton ();
	switches[0]->AttachObserver (editor, "OnClick", "DoShowHorizontal");
	
	switches[1] = OP_NEW (OPSwitch);
	switches[1]->Initialise (dx+20, 40, 70, 20, "vertical", view);
	switches[1]->InitAsRadioButton ();
	switches[1]->AttachObserver (editor, "OnClick", "DoShowVertical");
	
	buttons[0] = OP_NEW (OPButton);
	buttons[0]->Initialise (20, 20, 80, 24, "New", box);
	buttons[0]->AttachObserver (editor, "OnClick", "DoBandNew");
	
	buttons[1] = OP_NEW (OPButton);
	buttons[1]->Initialise (20, 45, 80, 24, "Select", box);
	buttons[1]->AttachObserver (editor, "OnClick", "DoBandSelect");
	
	buttons[2] = OP_NEW (OPButton);
	buttons[2]->Initialise (20, 70, 80, 24, "Insert", box);
	buttons[2]->AttachObserver (editor, "OnClick", "DoBandInsert");
	
	buttons[3] = OP_NEW (OPButton);
	buttons[3]->Initialise (20, 95, 80, 24, "Remove", box);
	buttons[3]->AttachObserver (editor, "OnClick", "DoBandRemove");
	
	spring_box = OP_NEW (OPSpringBox);
	spring_box->SetOrigin (120, 20);
	spring_box->SetSize (100, 100);
	spring_box->SetParent (box);
}


Bool
OPSpringEdit::Initialise (OPObjectIn edit, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
		
	if ( (edit.IsValid ())
	  && (edit->IsKindOf (OP_RUNTIME (OPView))) ) {
		
		this->edit = edit;
		this->view = OP_NEW (OPView);
		
		this->board    = OPSpringBoardNull;
		this->band     = OPSpringBandNull;
		this->band_num = 0;
		
		this->is_vertical = FALSE;
		
		this->view->SetOrigin (0, 0);
		this->view->SetSize (dx, dy);
		
		build_interface (this, this->view, this->spring_box,
						 this->buttons, this->switches);
		
		this->spring_view = OP_NEW (OPSpringView);
		this->switches[0]->TurnOn ();
		
		OPViewRef parent = this->edit->ReturnParent ();
		
		if (parent.IsValid ()) {
			this->board = parent->ReturnSpringBoard ();
		}
		
		this->DoShowHorizontal ();
		
	} else {
		this->view = OP_NEW (OPView);
		this->view->SetOrigin (0, 0);
		this->view->SetSize (0, 0);
	}
	
	return TRUE;
}

Bool
OPSpringEdit::Update (Bool find_view)
{
	OP_METHOD (Update, 1);
	OP_ARG (1, Bool, find_view);
	
	Count index = 0;
	
	OPSpringBandRef band;
	
	if ( (find_view)
	  && (this->board.IsValid ())
	  && (this->board->FindView (this->edit, index)) ) {
		this->band_num = index;
	}
	
	if (this->board.IsValid ()) {
		this->board->GetNthBand (this->band_num, band);
	}
	
	if (band.IsValid ()) {
		if (band->IsVertical () != this->is_vertical) {
			band = OPSpringBandNull;
		}
	}
	
	//	Update the selected items.
	
	this->board = board;
	this->band  = band;
	
	this->spring_box->SetSpringView (this->edit);
	this->spring_box->SetSpringBand (this->band);
	
	if (this->spring_view.IsValid ()) {
		this->spring_view->SetHiliteView (this->edit);
		this->spring_view->SetSpringBand (this->band);
	}

	this->UpdateButtonEnable ();
	
	this->view->Display ();
	this->edit->ReturnWindow ()->Display ();
	
	return TRUE;
}

//	New Select Insert Remove

Bool
OPSpringEdit::UpdateButtonEnable ()
{
	OP_METHOD (UpdateButtonEnable, 0);
	
	Count index = 0;
	
	Bool is_validband  = this->band.IsValid ();
	Bool is_validboard = this->board.IsValid ();
	
	Bool is_boardview = is_validboard && this->board->FindView (this->edit, index); index = 0;
	Bool is_bandview  = is_validband  && this->band->FindView (this->edit, index);  index = 0;
	
	Bool is_new_ok    = !is_bandview;
	Bool is_select_ok = is_validboard;
	Bool is_insert_ok = is_validband && !is_bandview;
	Bool is_remove_ok = is_validband &&  is_bandview;
	
	this->buttons[0]->SetEnable (is_new_ok);
	this->buttons[1]->SetEnable (is_select_ok);
	this->buttons[2]->SetEnable (is_insert_ok);
	this->buttons[3]->SetEnable (is_remove_ok);
	
	return TRUE;
}


const OPViewRef&
OPSpringEdit::ReturnView () const
{
	OP_METHOD (ReturnView, 0);
	return this->view;
}


Bool
OPSpringEdit::DoBandNew ()
{
	OP_METHOD (DoBandNew, 0);
	
	if (this->board.IsInvalid ()) {
		this->board = OP_NEW (OPSpringBoard);
		this->edit->ReturnParent ()->SetSpringBoard (this->board);
	}
	
	this->band_num = this->board->CountBands ();
	this->band = OP_NEW (OPSpringBand);
	this->band->SetVertical (this->is_vertical);
	this->board->InsertBand (this->band_num, this->band);
	this->DoBandInsert ();
	
	return TRUE;
}

Bool
OPSpringEdit::DoBandSelect (int increment)
{
	OP_METHOD (DoBandSelect, 1);
	OP_ARG (1, int, increment);
	
	if (board.IsInvalid ()) {
		this->Update ();
		return TRUE;
	}
	
	Count max = this->board->CountBands ();
	Count num = max+1;
	
	if (increment == 0) {
		if (max) {
			for (Count index = 0; index < max; index++) {
				
				OPSpringBandRef band;
				
				if ( (this->board->GetNthBand (index, band))
				  && (band->IsVertical () == this->is_vertical)
				  && (band->FindView (this->edit)) ) {
					this->band_num = index;
					this->Update ();
					return TRUE;
				}
			}
		}
	}
	
	if (max) {
		
		Count index = this->band_num+increment;
		OPSpringBandRef band;
		
		while (num--) {
			if (index >= max) index = 0;
			this->board->GetNthBand (index, band);
			if (band.IsValid ()) {
				if (band->IsVertical () == this->is_vertical) {
					this->band_num = index;
					this->Update ();
					return TRUE;
				}
			}
			index++;
		}
	}
	
	this->band_num = 0;
	this->Update ();
	
	return TRUE;
}

Bool
OPSpringEdit::DoBandInsert ()
{
	OP_METHOD (DoBandInsert, 0);
	
	Count     max = this->band->CountViews ();
	OPViewRef view;
	OPCoord   ox;
	OPCoord   oy;
	
	this->edit->GetOrigin (ox, oy);
	
	for (Count index = 0; index < max; index++) {
		if (this->band->GetNthView (index, view)) {
			const OPRect& box = view->ReturnBox ();
			if (this->is_vertical) {
				if (oy < box.Y1 ()) {
					break;
				}
			} else {
				if (ox < box.X1 ()) {
					break;
				}
			}
		}
	}
	
	this->band->InsertView (index, this->edit);
	this->band->ResetFromModel ();
	this->DoBandSelect (0);
	
	return TRUE;
}

Bool
OPSpringEdit::DoBandRemove ()
{
	OP_METHOD (DoBandRemove, 0);
	
	if (this->band.IsValid ()) {
		
		OPSpringBandRef band = this->band;
		
		this->band->RemoveView (this->edit);
		this->band->ResetFromModel ();
		this->band_num = 0;
		this->DoBandSelect (0);
		
		if (band->CountViews () == 0) {
			
			Count index = 0;
			
			if (this->board->FindBand (band, index)) {
				
				this->board->RemoveBand (index);
				
				if (this->board->CountBands () == 0) {
					this->board->Delete ();
					this->board = OPSpringBoardNull;
					this->edit->ReturnParent ()->SetSpringBoard (this->board);
				}
			}
		}
	}
	
	return TRUE;
}

Bool
OPSpringEdit::DoShowHorizontal ()
{
	OP_METHOD (DoShowHorizontal, 0);
	
	this->is_vertical = FALSE;
	this->band_num = 0;
	this->DoBandSelect (0);
	
	return TRUE;
}

Bool
OPSpringEdit::DoShowVertical ()
{
	OP_METHOD (DoShowVertical, 0);
	
	this->is_vertical = TRUE;
	this->band_num = 0;
	this->DoBandSelect (0);
	
	return TRUE;
}

Bool
OPSpringEdit::ActiveViewChanged (OPObjectIn tab, Card32 id)
{
	OP_METHOD (ActiveViewChanged, 2);
	OP_ARG (1, OPObjectIn, tab);
	OP_ARG (2, Card32, id);
	
	if (this->edit.IsValid ()) {
		
		OPViewRef      active;
		OPMultiViewRef multi = tab;
		
		multi->GetActiveView (active);
		
		if (active == this->view) {
			this->spring_view->SetHiliteView (this->edit);
			this->spring_view->SetSpringBand (this->band);
		} else {
			this->spring_view->SetHiliteView (OPViewNull);
			this->spring_view->SetSpringBand (OPSpringBandNull);
		}
		
		this->edit->ReturnWindow ()->Display ();
	}
	
	return TRUE;
}
